#include "BiasedResidualsProcessor.h"

//C++
#include <iostream>
using namespace std;

//LCIO
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include "IntersectionCalculator.h"
#include "UTIL/LCRelationNavigator.h"

// Aida
//#ifdef MARLIN_USE_AIDA // the whole processor does not make sense without aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
// endif

// Marlin 
#include <marlin/Global.h>

// GEAR
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>

#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

BiasedResidualsProcessor aBiasedResidualsProcessor;
    
BiasedResidualsProcessor::BiasedResidualsProcessor(): Processor("BiasedResidualsProcessor")
{
    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(    LCIO::TRACK,
				"InputTracks" , 
				"Name of the input tracks collection"  ,
				_inputColName ,
				std::string("TPCSeedTracks") ) ;
    // modify processor description
    _description = "BiasedResidualsProcessor: Fill biased residuals of hits in tracks into AIDA histgram" ;

  registerProcessorParameter( "HistogramMinValueXY" , 
			      "Minimum value in xy histogram (default: -0.3 mm)"  ,
			      _histogramMinValueXY ,
			      float(-0.3) ) ;

  registerOptionalParameter( "ReferenceTrackRelationsColName",
			     "Name of the collection with relations to reference tracks (MC truth or hodoscope)",
			     _relationsColName,
			     std::string("ReferenceTracksRelations") );

  registerProcessorParameter( "HistogramMaxValueXY" , 
			      "Maximal value in xy histogram (default:  0.3 mm)"  ,
			      _histogramMaxValueXY ,
			      float(0.3) ) ;

  registerProcessorParameter( "HistogramMinValueZ" , 
			      "Minimum value in z histogram (default: -1.0 mm)"  ,
			      _histogramMinValueZ ,
			      float(-1.0) ) ;

  registerProcessorParameter( "HistogramMaxValueZ" , 
			      "Maximal value in z histogram (default:  1.0 mm)"  ,
			      _histogramMaxValueZ ,
			      float(1.0) ) ;

  registerProcessorParameter("ApplyCorrectionFactor" ,
			     "Multiply residual with sqrt{n / (n-dof)} (default: true)" ,
			     _applyCorrectionFactor ,
			     bool (true) ) ;

  registerProcessorParameter("DoFXY" ,
			     "Degrees of freedom in the XY plane (2 for straight line, 3 for helix) (default: 3)",
			     _dofXY,
			     int (3) ) ;

  registerOptionalParameter("XRMin",
			    "Optional: overwrite x or r minimum which is usually taken from GEAR",
			    _xrMin,
			    float(0) );

  registerOptionalParameter("XRMax",
			    "Optional: overwrite x or r maximum which is usually taken from GEAR",
			    _xrMax,
			    float(0) );

  registerOptionalParameter("YPhiMin",
			    "Optional: overwrite y or phi minimum which is usually taken from GEAR",
			    _yphiMin,
			    float(0) );

  registerOptionalParameter("YPhiMax",
			    "Optional: overwrite y or phi maximum which is usually taken from GEAR",
			    _yphiMax,
			    float(0) );

  registerProcessorParameter("ForceCartesian",
		    "force the histograms to cartesian values in case of polar TPC coordinates, default: false",
		    _forceCartesian,
		    false);
  
  
}

BiasedResidualsProcessor::~BiasedResidualsProcessor()
{
}

void BiasedResidualsProcessor::init()
{ 
	// usually a good idea to
	printParameters() ;

	// get histograms from the AIDAProcessor
	_residualsXYHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "residualsXYHisto", 
			     "residualsXYHisto", 
			     200, _histogramMinValueXY, _histogramMaxValueXY);
	_residualsZHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "residualsZHisto", 
			     "residualsZHisto", 
			     4000, _histogramMinValueZ, _histogramMaxValueZ);

	// the plane extend is taken from GEAR
	const  std::vector< double > &planeExtent = Global::GEAR->getTPCParameters().getPlaneExtent();

	// convert to cartesian in polar case
	// FIXME: Currently it's always a full circle which is much easyer to calculate
	float xrMin, xrMax, yphiMin, yphiMax;
	switch (Global::GEAR->getTPCParameters().getCoordinateType () )
	{
	  case gear::PadRowLayout2D::CARTESIAN:
	    xrMin   =  planeExtent[0];
	    xrMax   =  planeExtent[1];
	    yphiMin =  planeExtent[2];
	    yphiMax =  planeExtent[3];
	    break;

	  case gear::PadRowLayout2D::POLAR:
	    if ( _forceCartesian ) // Take the outer boundaries of the maximal circle
	      // to avoid complicated calculations.
	      // The user can set different values in the steering file
	    {
	      xrMin   =  -planeExtent[1];// the outer radius
	      xrMax   =   planeExtent[1];
	      yphiMin =  -planeExtent[1];
	      yphiMax =   planeExtent[1];
	    }
	    else // take polar coordinates from gear
	    {
	      xrMin   =  planeExtent[0];
	      xrMax   =  planeExtent[1];
	      yphiMin =  planeExtent[2];
	      yphiMax =  planeExtent[3];
	    }
	    break;

	  default:
	    throw gear::Exception("Unkown coordinate type");
	}

	//overwrite the gear parameter if the user has specified it
	if ( parameterSet("XRMin") )   { xrMin   = _xrMin; }
	if ( parameterSet("XRMax") )   { xrMax   = _xrMax; }
	if ( parameterSet("YPhiMin") ) { yphiMin = _yphiMin; }
	if ( parameterSet("YPhiMax") ) { yphiMax = _yphiMax; }

	// fixme: get all anode positions for correct range
	double zMax = Global::GEAR->getTPCParameters().getMaxDriftLength();

	_residualsXY_vs_xr =   AIDAProcessor::histogramFactory(this)->
	    createHistogram2D( "residualsXY_vs_xr", 
			       "residualsXY_vs_xr", 
			       200, _histogramMinValueXY, _histogramMaxValueXY,
		               200, xrMin, xrMax);

 	_residualsXY_vs_yphi =   AIDAProcessor::histogramFactory(this)->
	    createHistogram2D( "residualsXY_vs_yphi", 
			       "residualsXY_vs_yphi", 
			       200, _histogramMinValueXY, _histogramMaxValueXY,
		               200, yphiMin, yphiMax);

	_residualsZ_vs_xr =   AIDAProcessor::histogramFactory(this)->
	    createHistogram2D( "residualsZ_vs_xr", 
			       "residualsZ_vs_xr", 
			       200, _histogramMinValueZ, _histogramMaxValueZ,
		               200, xrMin, xrMax);

 	_residualsZ_vs_yphi =   AIDAProcessor::histogramFactory(this)->
	    createHistogram2D( "residualsZ_vs_yphi", 
			       "residualsZ_vs_yphi", 
			       200, _histogramMinValueZ, _histogramMaxValueZ,
		               200, yphiMin, yphiMax);

 	_residualsZ_vs_z =   AIDAProcessor::histogramFactory(this)->
	    createHistogram2D( "residualsZ_vs_z", 
			       "residualsZ_vs_z", 
			       200, _histogramMinValueZ, _histogramMaxValueZ,
		               200, 0, zMax );

 	_residualsXY_vs_z =   AIDAProcessor::histogramFactory(this)->
	    createHistogram2D( "residualsXY_vs_z", 
			       "residualsXY_vs_z", 
			       200, _histogramMinValueXY, _histogramMaxValueXY,
		               200, 0, zMax );
		       
   }

void BiasedResidualsProcessor::processRunHeader( EVENT::LCRunHeader* run )
{
}
    
void BiasedResidualsProcessor::processEvent( LCEvent * evt ) 
{ 

    // we need a LCCollectionVec to use iterators, as LCCollectio is not a std::vector
    LCCollectionVec* inputCollection=0;
    try{
	inputCollection = dynamic_cast<LCCollectionVec*>(evt->getCollection( _inputColName )) ;
    }
    catch(DataNotAvailableException &e)
    {
	// no input data
	// just continue ...
	return;
    }

    if( inputCollection != 0 )
    {
        // Try to find a collection with LCRelations to the MC truth.
        // This is an optional parameter. Only use it if it is set
        LCCollection* referenceTrackRelationsCol = 0;
	LCRelationNavigator* relationNavigator = 0;
        if (parameterSet( "ReferenceTrackRelationsColName" ))
        {
	  try {
	    referenceTrackRelationsCol = evt->getCollection( _relationsColName ) ;
	  }
	  catch(DataNotAvailableException){
	    streamlog_out(WARNING) << "reference track collection requested, "
				   << "but no collection with name \"" << _relationsColName
				   << "\" found in event! Skipping this event" << std::endl;
	    return;
	  }

	  relationNavigator = new LCRelationNavigator( referenceTrackRelationsCol );
        }

	int trackNr=0;

        LCParameters const & trackCollectionParameters = inputCollection->getParameters();

	for(LCCollectionVec::iterator it = inputCollection->begin() ; 
	    it < inputCollection->end(); it ++ )
	{
	    
	    streamlog_out(DEBUG4) <<"\t TrackNumber "<<trackNr++<<std::endl;

	    Track * track = dynamic_cast<Track *>(*it) ;

	    Track * referenceTrack = 0; // possible reference track for MCTruth
	    
	    if ( relationNavigator ) // there is MCTruth info
	    {
	      LCObjectVec referenceTracks = relationNavigator->getRelatedToObjects(track);
	      if ( referenceTracks.size() == 0 )
	      {
		streamlog_out(MESSAGE) << "No mc truth for track " << trackNr++
				       << " in event " << evt->getEventNumber() 
				       << ". Skipping this track!" << std::endl;
		continue; // with next track
	      }
	      
	      // always use the first entry... this should be the one with 
	      /// most overlapping hits
	      referenceTrack = dynamic_cast<Track *>( *referenceTracks.begin() );
	    }
	    else // no external reference
	    {
	      // use the track itself to calculate the residuals (they will be/ might be biased)
	      referenceTrack = track;
	    }

	    // first try to get a fitter, according to the fitter type defined in the track
	    unsigned char fitterType = TrackFitterBase::getFitterTypeFromTrack( track->getType() );

	    if (fitterType == 0) // no fitter has been set
	    {
	        fitterType = TrackFitterBase::FitterTypes::SIMPLEMINIMIZER;
	    }
	    
	    streamlog_out(DEBUG4) << "Using fitter type "<<static_cast<unsigned short>(fitterType)<< " to calculate residuals"<<std::endl;

	    TrackFitterBase *trackFitter = TrackFitterFactory::getTrackFitter( fitterType ,
									       & trackCollectionParameters);
	    
	    TrackerHitVec hitVector = track->getTrackerHits();
	    
//	    cout << "nr of hits " << hitVector.size()<< endl;

//	    int hitNr=0;

//	    cout << "\t\t hit\t x \t y \t z \t residual_x/r*phi" << endl;
	    for (TrackerHitVec::iterator hit = hitVector.begin();
		 hit <  hitVector.end(); hit ++)
	    {
		const double *position = (*hit)->getPosition();

		try
		{
		    EVENT::DoubleVec residuals = trackFitter->calculateResiduals(*hit, referenceTrack);
		    _residualsXYHisto->fill( residuals[0] * (_applyCorrectionFactor?
							     sqrt(hitVector.size() / ( hitVector.size() - _dofXY) ):
							     1) );
		    _residualsZHisto->fill(  residuals[1]  * (_applyCorrectionFactor?
							      sqrt(hitVector.size() / (hitVector.size()-2) ):
							      1) );

		    if ( (Global::GEAR->getTPCParameters().getCoordinateType() 
			   == gear::PadRowLayout2D::POLAR ) &&
			 ( ! _forceCartesian ) )
		    { // convert the cartesian lcio coordinates to polar
		      double r =  sqrt(position[0]* position[0] +  position[1]* position[1]);
		      double phi = atan2( position[1] ,  position[0] );
		      // wrap phi to 0..2pi
		      if (phi< 0) phi+=2*M_PI;

 		      _residualsXY_vs_xr->fill( residuals[0], r );
		      _residualsXY_vs_yphi->fill( residuals[0], phi );

		      _residualsZ_vs_xr->fill( residuals[1], r );
		      _residualsZ_vs_yphi->fill( residuals[1], phi );
		    }
		    else
		    { // jist use the cartesian lcio coordinates
 		      _residualsXY_vs_xr->fill( residuals[0], position[0] );
		      _residualsXY_vs_yphi->fill( residuals[0], position[1] );

		      _residualsZ_vs_xr->fill( residuals[1], position[0] );
		      _residualsZ_vs_yphi->fill( residuals[1], position[1] );

		    }

		    _residualsXY_vs_z->fill( residuals[0], position[2] );
		    _residualsZ_vs_z->fill( residuals[1], position[2] );

//		    cout << residuals[0] << "\t" << residuals[1]  << std::endl;
		}
		catch ( IntersectionCalculator::NoIntersectionException &e )
		{
		    cout << e.what() << endl;
		}
	
	    }
        }

    }
}


}// namespace 
