#include "LinearGeometricMeanResolutionProcessor.h"
#include "ZBinTemplateProcessor.cc"

// C++
#include <iostream>
#include <sstream>
#include <cmath>
#include <vector>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include "UTIL/LCTOOLS.h"
#include "IntersectionCalculator.h"

// Aida
#ifdef MARLIN_USE_AIDA
//AIDA
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
#endif

#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"
#include "LinearRegressionProcessor.h"
#include "LinearTrackRegression.h"

using namespace lcio ;
using namespace marlin ;

namespace marlintpc{

// define the indices for the template parameter
// they are for convenience and better code readability
    const unsigned int LinearGeometricMeanResolutionProcessor::_histoGeoNIndex   = 0;
    const unsigned int LinearGeometricMeanResolutionProcessor::_histoGeoNminus1Index   = 1;
    const unsigned int LinearGeometricMeanResolutionProcessor::_histoGeoZNIndex   = 2;
    const unsigned int LinearGeometricMeanResolutionProcessor::_histoGeoZNminus1Index   = 3;


    LinearGeometricMeanResolutionProcessor aLinearGeometricMeanResolutionProcessor ;
    

    LinearGeometricMeanResolutionProcessor::LinearGeometricMeanResolutionProcessor() : 
	ZBinTemplateProcessor<N_1D_HISTOS_GEOMETRIC_MEAN_RESOLUTION_PROCESSOR ,
			      N_2D_HISTOS_GEOMETRIC_MEAN_RESOLUTION_PROCESSOR>("LinearGeometricMeanResolutionProcessor"),

	_histoGeoN(0), 
	_histoGeoNminus1(0),
	_histoGeoZN(0), 
	_histoGeoZNminus1(0)

    {
	

//     Processor("LinearGeometricMeanResolutionProcessor"), _fitterFactory(0) {
	/* the constructor call first the super constructor which the name 
	 * of the processor
	 * this must by the same as the class name.
	 */
	
	/* description of the processor which will be displayed in the
	 * steering file autometicly generated by 
	 * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	 */
	_description = "LinearGeometricMeanResolutionProcessor determines the resolution via the geometric mean method" ;
	
	
	/* register steering parameters: name, description, class-variable, default value
	 * the type will by definded by the type of the default value
	 * string, double, float and int are possible
	 * plaese note to get a float you must cast in C++: (float) 1.3 
	 */
	registerInputCollection( LCIO::TRACK,
				 "InputCollectionName" , 
				 "Name of the input collection"  ,
				 _inputColName ,
				 std::string("TimePixTracks") ) ;
	
	registerProcessorParameter( "HistoRange" , 
				    "Histograms will have range (-_histoRange, _histoRange), default _histoRange = 1mm"  ,
				    _histoRange ,
				    float(1.) ) ;

	registerProcessorParameter( "HistoBin" , 
				    "Histograms will be binned with _histoBin, default _histoBin = 100"  ,
				    _histoBin ,
				    int(100) ) ;

 	registerProcessorParameter( "ApplyZCut" , 
				    "Flag whether to cut in z , default 0",
				    _applyZCut ,
				    int(0) ) ;

	registerProcessorParameter( "ZMin",
				    "Minimal z value, default 0 mm",
				    _zMin ,
				    float(0.) ) ;

	registerProcessorParameter( "ZMax",
				    "Maximal z value, default 0 mm",
				    _zMax ,
				    float(100.) ) ;

	registerProcessorParameter( "HistoGeoNName",
				    "Name of the histogramm with all hits in the track fit ,"
				    " default: histoGeoN",
				    _histoGeoNName,
				    std::string("histoGeoN") );

	registerProcessorParameter( "HistoGeoNminus1Name",
				    "Name of the histogramm with n -1 hits in the track fit ,"
				    " default: histoGeoNminus1",
				    _histoGeoNminus1Name,
				    std::string("histoGeoNminus1") );
				    
	registerProcessorParameter( "HistoGeoZNName",
				    "Name of the histogramm with all hits in the track fit ,"
				    " default: histoGeoZN",
				    _histoGeoZNName,
				    std::string("histoGeoZN") );

	registerProcessorParameter( "HistoGeoZNminus1Name",
				    "Name of the histogramm with n -1 hits in the track fit ,"
				    " default: histoGeoZNminus1",
				    _histoGeoZNminus1Name,
				    std::string("histoGeoZNminus1") );

	// set correct names for the z binned histograms
	_histo1DBaseNames[_histoGeoNIndex]	   = "NFit";
	_histo1DTitles[_histoGeoNIndex] = "N Fit / mm";

	_histo1DBaseNames[_histoGeoNminus1Index]	   = "Nminus1Fit";
	_histo1DTitles[_histoGeoNminus1Index] = "N-1 Fit / mm";
	
	_histo1DBaseNames[_histoGeoZNIndex]	   = "ZNFit";
	_histo1DTitles[_histoGeoZNIndex] = "N Fit / mm";

	_histo1DBaseNames[_histoGeoZNminus1Index]	   = "ZNminus1Fit";
	_histo1DTitles[_histoGeoZNminus1Index] = "N-1 Fit / mm";


    }
    
  LinearGeometricMeanResolutionProcessor::~LinearGeometricMeanResolutionProcessor() {
  } 


 
    void LinearGeometricMeanResolutionProcessor::init() { 
	
	// usually a good idea to
	printParameters() ;

	std::string histoGeoNTitle = "N-fit for geometric mean resolution";
	std::string histoGeoNminus1Title = "N.1-fit for geometric mean resolution";
	std::string histoGeoZNTitle = "N-fit for geometric mean resolution";
	std::string histoGeoZNminus1Title = "N.1-fit for geometric mean resolution";

	if (_applyZCut)
	{
	    std::stringstream zRange;
	    zRange << " , "<< _zMin << " < z < " << _zMax ;
	    histoGeoNTitle += zRange.str();
	    histoGeoNminus1Title += zRange.str();
	    histoGeoZNTitle += zRange.str();
	    histoGeoZNminus1Title += zRange.str();
	    
	    
	}

	_histoGeoN = AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( _histoGeoNName, histoGeoNTitle, 
			     _histoBin, -_histoRange, _histoRange ); 

	_histoGeoNminus1 = AIDAProcessor::histogramFactory(this)->
	    createHistogram1D( _histoGeoNminus1Name, histoGeoNminus1Title,
			       _histoBin, -_histoRange, _histoRange ); 
			       
	_histoGeoZN = AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( _histoGeoZNName, histoGeoZNTitle, 
			     _histoBin, -_histoRange, _histoRange ); 

	_histoGeoZNminus1 = AIDAProcessor::histogramFactory(this)->
	    createHistogram1D( _histoGeoZNminus1Name, histoGeoZNminus1Title,
			       _histoBin, -_histoRange, _histoRange ); 

	// set histo range and one bin per int for templated histograms
	// according to Processor parameters
	_histoRanges1D[_histoGeoNIndex].min = -_histoRange;
	_histoRanges1D[_histoGeoNIndex].max = _histoRange;
	_nBins1D[_histoGeoNIndex] = _histoBin;

	_histoRanges1D[_histoGeoNminus1Index].min = -_histoRange;
	_histoRanges1D[_histoGeoNminus1Index].max = _histoRange;
	_nBins1D[_histoGeoNminus1Index] = _histoBin;
	
	_histoRanges1D[_histoGeoZNIndex].min = -_histoRange;
	_histoRanges1D[_histoGeoZNIndex].max = _histoRange;
	_nBins1D[_histoGeoZNIndex] = _histoBin;

	_histoRanges1D[_histoGeoZNminus1Index].min = -_histoRange;
	_histoRanges1D[_histoGeoZNminus1Index].max = _histoRange;
	_nBins1D[_histoGeoZNminus1Index] = _histoBin;


	// now call ZBinTemplateProcessor's init which initialises the histograms for us
	ZBinTemplateProcessor<N_1D_HISTOS_GEOMETRIC_MEAN_RESOLUTION_PROCESSOR,
	                      N_2D_HISTOS_GEOMETRIC_MEAN_RESOLUTION_PROCESSOR>::init();
	
	
    }
    
    void LinearGeometricMeanResolutionProcessor::processRunHeader( LCRunHeader* run) { 
	
      
      run->parameters().setValue( _processorName + "_revision", "$Rev: 2632 $");

      for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
	if ( ! i->second->isOptional() || i->second->valueSet() ) {
	  run->parameters().setValue( _processorName + "_" + i->second->name(), 
				      i->second->value() );
	}
      }

    } 
    
    void LinearGeometricMeanResolutionProcessor::processEvent( LCEvent * evt ) { 


	
      // get the input collection from the event
      LCCollection* inputCol = 0;

      try{

	inputCol = evt->getCollection( _inputColName ) ;

      } //end of try {

      catch(DataNotAvailableException){
	// if there is not input collection give a warning to the user
	streamlog_out(WARNING4)<<" Input collection "
		 <<_inputColName<<" cannot be found in event "
		 <<evt->getEventNumber()<<std::endl;

	//just return, no data to process
	return;
      }
      
      // use TrackFitterChiSqare to calculate the residuals (for the time being)
      //      LCParametersImpl trackCollectionParameters;
      //      trackCollectionParameters.setValue("TrackFitterType","ChiSquare");
      
     
      TrackFitterBase *trackFitter = TrackFitterFactory::getTrackFitter( TrackFitterBase::FitterTypes::SIMPLEMINIMIZER,
									 0);


      // get the number of elements in the input collection, 
      //and loop over it


      for(int iTrack=0; iTrack< inputCol->getNumberOfElements() ; iTrack++){
	
	/* getElement returns only a LCObject
	 * you have to cast it to the get the right type
	 * please use dynamic_cast which will return 0 if the object cannot be casted
	 */
	Track* track = 
	  dynamic_cast<Track*>( inputCol->getElementAt( iTrack ) ) ;
	
	if ( track != 0 ) {
	  //only proceed if the cast was successful

	  TrackerHitVec HitVec = track->getTrackerHits();
	  
//	  double phi =  track->getPhi();
//	  double d0 =  track->getD0();

	  
	  for ( TrackerHitVec::iterator hit = HitVec.begin(); hit < HitVec.end(); hit++ ) 
	  { 
	      // apply z cut, if not in range skip this hit
	      if (_applyZCut && ( ( (*hit)->getPosition()[2] < _zMin ) ||
				  ( (*hit)->getPosition()[2] >= _zMax ) ) )
		  continue;
	      
	      double residual1 = 0;
	      double residual2 = 0;
	      
	      //calculate the distance from each hit to the track and fill in histogram
	      try {
	      
		  EVENT::DoubleVec residuals = trackFitter->calculateResiduals(*hit, track);
		  _histoGeoN->fill( residuals[0] );
		  _histoGeoZN->fill( residuals[1] );
		  
		  //FIXME use z0 of the track as z position
		  //since this was wrong /  bad it's changed by now
		  fillHistogram( _histoGeoNIndex, 
				 residuals[0] , 
				 (*hit)->getPosition()[2] );
		  fillHistogram( _histoGeoZNIndex, 
				 residuals[1] , 
				 (*hit)->getPosition()[2] );	


		  residual1 = residuals[0];
		  
	      }
	      catch ( IntersectionCalculator::NoIntersectionException &e ){
		  std::cout << e.what() << std::endl;
	      }
	    
	      
	    //for each hit recalculate track
	      
	    // the actual track estimate
	      IMPL::TrackImpl* outputTrack = new IMPL::TrackImpl();
	      outputTrack->setTypeBit(LCIO::TRBIT_HITS);
	      
	      // fill all hits into the seed track:
	      for (EVENT::TrackerHitVec::iterator Hit = HitVec.begin();
		   Hit < HitVec.end(); Hit++) {
		  
		  if ( hit != Hit ) outputTrack->addHit(*Hit);
		  
	      }
	      
	      
	      //if track is parameterized by y ( x = ay + b )
	      if ( fabs( track->getPhi() ) > M_PI/4 || 
		   fabs( track->getPhi() ) == M_PI/4 ) {
		  
		  
		  LinearTrackRegression::StraightLine straightlineX = 
		      LinearTrackRegression::doRegressionX(outputTrack);
		  
		  streamlog_out(DEBUG4)<<"straigtline: a = "<<straightlineX.slope1
				       <<" b = "<<straightlineX.intersept1<<std::endl;
		  
		  // now set the estimated values in the seed track
		  outputTrack->setOmega( 0 );
		  
		  outputTrack->setPhi( M_PI/2 - atan(straightlineX.slope1) );
		  
		  outputTrack->setD0( - straightlineX.intersept1 * sin( outputTrack->getPhi())  ); 
		  
		  outputTrack->setTanLambda( tan( straightlineX.slope2 / fabs( straightlineX.slope2 ) * acos( sqrt( 				straightlineX.slope1*straightlineX.slope1 + 1 ) / 
				  sqrt( straightlineX.slope1*straightlineX.slope1 + 1 
				  + straightlineX.slope2*straightlineX.slope2) ) ) );
		  //z_0 = slope2 * y_0 + intersept2
		  //y_0 = sin( pi/2 - phi) * d_0
		  outputTrack->setZ0( - (straightlineX.slope1*straightlineX.intersept1*straightlineX.slope2)/
				  (straightlineX.slope1*straightlineX.slope1 + 1)
				   + straightlineX.intersept2);
		  
		  
	      }
	      
	      //if track is parameterized by x ( y = ax + b )
	      else { 
		  
		  LinearTrackRegression::StraightLine straightlineY = 
		      LinearTrackRegression::doRegressionY(outputTrack);
		  
	      
		  // now set the estimated values in the seed track
		  outputTrack->setOmega( 0 );
		  
		  outputTrack->setPhi( atan(straightlineY.slope1) );
		  
		  outputTrack->setD0( straightlineY.intersept1 * cos( outputTrack->getPhi())  );
		  
		  outputTrack->setTanLambda( tan( straightlineY.slope2 / fabs( straightlineY.slope2 ) * 
				  acos( sqrt( straightlineY.slope1*straightlineY.slope1 + 1 ) / 
				  sqrt( straightlineY.slope1*straightlineY.slope1 + 1 
				  + straightlineY.slope2*straightlineY.slope2) ) ) );
		  		  
		  //z_0 = slope2 * x_0 + intersept2
		  //x_0 = sin( phi ) * d_0
		  outputTrack->setZ0( - (straightlineY.slope1*straightlineY.intersept1*straightlineY.slope2)/
				  (straightlineY.slope1*straightlineY.slope1 + 1)
				  + straightlineY.intersept2);
		  		  
	      }
	      
	      
	      //calculate the distance from each hit to the track and fill in histogram
	      
	      
	      try {
		  
		  EVENT::DoubleVec residuals = trackFitter->calculateResiduals(*hit, 
									       outputTrack);
		  _histoGeoNminus1->fill( residuals[0] );
		  _histoGeoZNminus1->fill( residuals[1] );

		  //FIXME use z0 of the track as z position
		  //since this was wrong /  bad it's changed by now
		  fillHistogram( _histoGeoNminus1Index, 
				 residuals[0] , 
				 (*hit)->getPosition()[2] );
		  fillHistogram( _histoGeoZNminus1Index, 
				 residuals[1] , 
				 (*hit)->getPosition()[2] );

		  residual2 = residuals[0];
		  
	      }
	      catch ( IntersectionCalculator::NoIntersectionException &e ){
		  std::cout << e.what() << std::endl;
	      }

	      delete outputTrack;
	      
	      
	  } // for hit 
	  
	} //end of if ( track != 0 ) {
	
      } //end of for(int iTrack=0; iTrack< nTrack ; iTrack++){
      
      
      
    }
    
    
    void LinearGeometricMeanResolutionProcessor::check( LCEvent * evt ) { 



    }

    
    
    void LinearGeometricMeanResolutionProcessor::end(){ 
      /*	
      streamlog_out(MESSAGE4) << "LinearGeometricMeanResolutionProcessor::end()  " << name() 
		<< " processed " << evt->getEventNumber() 
		<< " events in " << _nRun << " runs "
		<< std::endl ;
      */	

    }
    
}
