/** Processor to amplify drifted electrons with GEM structure
 */

#include "GEMProcessor.h"

//MarlinTPC
#include "libgem.h"


//tpcconddata
#include "TPCConditions.h"

//lccd
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include <lccd/DataFileHandler.hh>



//C++
#include <iostream>
#include <cmath>
#include <complex>
#include <cstring>
#include <cstdlib>

//LCIO
#include <IO/LCWriter.h>
#include <EVENT/LCCollection.h>
#include <EVENT/MCParticle.h>
#include <IMPL/LCRunHeaderImpl.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/SimTrackerHitImpl.h>
#include <IMPL/LCGenericObjectImpl.h>

//GEAR
#include <gear/TPCParameters.h>

//Marlin
#include <marlin/Global.h>
#include <marlin/Exceptions.h>

//CLHEP
#include <CLHEP/Random/RandExponential.h> 
#include <CLHEP/Random/RandGauss.h> 
#include <CLHEP/Random/RandBinomial.h> 


#define GEMPITCH 0.14



using namespace lcio ;
using namespace marlin ;
using namespace std;
using namespace tpcconddata;



namespace marlintpc{

GEMProcessor aGEMProcessor ;


GEMProcessor::GEMProcessor() : Processor("GEMProcessor") 
{
  
  _description = "GEMProcessor determines amplification in GEM stack" ;
  
  
  // register steering parameters: name, description, class-variable, default value

  registerInputCollection(    LCIO::SIMTRACKERHIT,
                              "InputCollectionName" , 
			      "Name of the Input collection with drifted electrons (default: DriftedElectrons)"  ,
			      _inputCollectionName ,
			      string("DriftedElectrons") ) ;

  registerOutputCollection(   LCIO::SIMTRACKERHIT,
                              "OutputCollectionName" ,
                              "Name of the Output GEM collection (default: AmplifiedElectrons)" ,
                              _outputCollectionName ,
                              string("AmplifiedElectrons") ) ;

  registerProcessorParameter( "ParamFile" , 
			      "Name of the File containing Parametrisation of GEMs (default: paramerterfile)"  ,
			      paramfile ,
			      std::string("parameterfile") ) ;

 registerProcessorParameter( "UGEM1" , 
			      "Voltage on GEM 1 (default: 315)"  ,
			      UGEM1 ,
			      (float)315) ;

  registerProcessorParameter( "UGEM2" , 
			      "Voltage on GEM 2 (default: 315)"  ,
			      UGEM2 ,
			      (float)315) ;

  registerProcessorParameter( "UGEM3" , 
			      "Voltage on GEM 3 (default: 315)"  ,
			      UGEM3 ,
			      (float)315) ;

  registerProcessorParameter( "ETrans1" , 
			      "Value of Transfer Field 1 (default: 2500)"  ,
			      ETrans1 ,
			      (float)2500) ;

  registerProcessorParameter( "ETrans2" , 
			      "Value of Transfer Field 2 (default: 2500)"  ,
			      ETrans2 ,
			      (float)2500) ;

  registerProcessorParameter( "EInd" , 
			      "Value of Induction Field (default: 5000)"  ,
			      EInd ,
			      (float)5000) ;

  registerProcessorParameter( "GainFactor" , 
			      "Correction Factor for Gain (default: 1)"  ,
			      _gainFactor ,
			      (float)1) ;

  registerProcessorParameter( "Transient" , 
			      "switch off writing of output collection (derfault: 0)"  ,
			      _transient,
			      (int) 0) ;

  registerProcessorParameter( "DriftField" , 
			      "Drift field strength in V/cm (default: 240)"  ,
			      _driftField,
			      (float) 240) ;

  registerProcessorParameter( "Gas" ,
			      "Gas in the active Volume of the TPC (default: TDR)",
			      gas,
			      string("TDR") );

  registerOptionalParameter(  "TPCConditions" ,
			      "Name of collection with TPCConditions" ,
			      _tpcConditionsColName,
			      tpcconddata::TPCConditions::getDefaultColName() );

  registerOptionalParameter(  "AlternativTPCConditionsName",
			      "This is the name of newly created TPCConditions, if the original file specified by \"TPCConditions\" is not writeable (default: TPCConditionsGEM)" ,
			      _altTPCConditionsName,
			      string("TPCConditionsGEM") ) ;

  registerProcessorParameter( "IonisationEnergy" ,
			      "The energy in GeV needed to ionize a gas particle (default: 26.E-9 = Argon) ",
			      _ionisationEnergy,
			      26.E-9); //default value is for argon

  registerProcessorParameter( "FullCurlerBackwardCompatibility" ,
			      " Turn on for data from PrimaryIonisation < r1876 and FullCurler turned off (default: false)",
			      _fullCurlerBackwardCompatibility,
			      false);
}


void GEMProcessor::init() { 

  streamlog_out(MESSAGE2) << _description << endl;

  //Usually a good idea....
  printParameters() ;

  //Readout of the pad layout from gear
  _padLayout= &(Global::GEAR->getTPCParameters().getPadLayout());

  //Transform name of file with GEM parametrisation in c-string
  Gas=gas.c_str();

  //As the attachment in a GEM stack depends on the electric fields 
  //only it can be calculated once for the whole run
  A1 = CalculateAttachment( ETrans1 );
  A2 = CalculateAttachment( ETrans2 );
  A3 = CalculateAttachment( EInd );

  _nRun = 0;
  _nEvt = 0;

  //Load GEM settings from parametrisation file
  LoadGEMSettings();     

  //Estimate mean amplification of an single electron
  _meanAmp = Coll[0] * Gain[0] * Extr[0] * A1 
    * Coll[1] * Gain[1] * Extr[1] * A2 
    * Coll[2] * Gain[2] * Extr[2] * A3 ; 

  streamlog_out( MESSAGE4 ) << "Mean amplification:  " << _meanAmp << endl;

}

 
void GEMProcessor::processRunHeader( LCRunHeader* run) {

  //Revision logging
  run->parameters().setValue( _processorName + "_revision", "$Rev$" );

  //Parameter logging
  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {

    if ( ! i->second->isOptional() || i->second->valueSet() ) 
      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
  }  

  _nRun++;//LCCD
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include "lccd/DataFileHandler.hh"
} 


void GEMProcessor::processEvent( LCEvent * evt ) { 

  if( _nEvt == 0){

    LCCollectionVec *conditionsCol = 0;

    try{

      //First try to get the conditions collection with TPCConditions from 
      //the current event.
      conditionsCol = dynamic_cast<LCCollectionVec*>( evt->getCollection( _tpcConditionsColName ) );
      TPCConditions tpcCond( conditionsCol->getElementAt(0) );
    
      //Check whether object is writeable 
      try{
	//This thing throws an exception if the object is write protected
	//tpcCond.obj()->checkAccess();

	//Add amplification to TPCCondtions if collection is writeable
	tpcCond.setAmplification( _meanAmp );

      }catch(EVENT::ReadOnlyException &){

	//If collection is not writeable an new writeable collection has
	//to be created with a new also writeable conditions object. The
	//new conditions object needs to be a deep copy of the original 
	//conditions object 
	
	conditionsCol  = new LCCollectionVec( LCIO::LCGENERICOBJECT );
	TPCConditions* newtpcCond = new TPCConditions;
	newtpcCond->setTransDiffusionCoef( tpcCond.getTransDiffusionCoef() );
	newtpcCond->setLongDiffusionCoef(  tpcCond.getTransDiffusionCoef() );
	newtpcCond->setDriftVelocity(  tpcCond.getDriftVelocity() );
	
	//Add mean amplification to the new conditions object
	newtpcCond->setAmplification( _meanAmp );

	//Add conditions object to collection and collection to event
	conditionsCol->addElement( newtpcCond );
	evt->addCollection( conditionsCol, _altTPCConditionsName);

	//Create a lccd::DataFileHandler 
	lccd::LCConditionsMgr::instance()->registerHandler( _altTPCConditionsName , new lccd::DataFileHandler( _altTPCConditionsName ) );
  
	// call updateEvent only for the TPCConditions handler! Don't run it on LCConditionsMgr!!!
	lccd::LCConditionsMgr::instance()->getHandler( _altTPCConditionsName )->updateEvent(  evt ) ;
    
	// The updateEvent performs an LCEvent::takeCollection, which sets the collection to transient
	// In this special case we want to write it to the file, so we have to remove the transient flag
	conditionsCol->setTransient(false);
    }

    }catch( DataNotAvailableException &e ){

      //If there is no TPCConditions collection create a new one and register it 
      //with lccd.

      conditionsCol  = new LCCollectionVec( LCIO::LCGENERICOBJECT );

      TPCConditions *tpcCond = new TPCConditions;
      tpcCond->setAmplification( _meanAmp );

      conditionsCol->addElement( tpcCond );

      evt->addCollection( conditionsCol, _tpcConditionsColName );

      //Create a lccd::DataFileHandler 
      lccd::LCConditionsMgr::instance()->registerHandler( _tpcConditionsColName , new lccd::DataFileHandler( _tpcConditionsColName ) );

      // call updateEvent only for the TPCConditions handler! Don't run it on LCConditionsMgr!!!
      lccd::LCConditionsMgr::instance()->getHandler( _tpcConditionsColName )->updateEvent(  evt ) ;
      
   
      // The updateEvent performs an LCEvent::takeCollection, which sets the collection to transient
      // In this special case we want to write it to the file, so we have to remove the transient flag
      conditionsCol->setTransient(false);
    }

  }

  LCCollection* inputCol = 0; 

  int objects=1;

  try {
    inputCol = evt->getCollection( _inputCollectionName );
    objects=inputCol->getNumberOfElements();
    streamlog_out(DEBUG4) << objects << " objects in input collection" << endl;
  }
  catch(EventException &exception){
    streamlog_out(DEBUG4) << "Exception: " << exception.what() << endl;
    objects=0;
  }

  //Create output collection and set transient
  LCCollectionVec* outputColVec = new LCCollectionVec( LCIO::SIMTRACKERHIT )  ;
  if(_transient)
    outputColVec->setTransient( true );


  const double *parray = 0;  

  for( int i = 0; i < objects; i++ ){

    SimTrackerHit *hit_in=dynamic_cast<SimTrackerHit*>( inputCol->getElementAt(i) );
    MCParticle    *mcpart=dynamic_cast<MCParticle*>( hit_in->getMCParticle() );
 
    
    //Check wether object is curler
    bool isCurler=false;
    if( mcpart != 0 ){

      //Get momentum of the SimTrackerHits parent particle 
      parray=mcpart->getMomentum();

      if( sqrt( pow(parray[0],2) + pow(parray[1],2) ) < 1 && 
	  atan( sqrt( pow(parray[0] , 2) + pow(parray[1] , 2) )/(fabs(parray[2]))) * 180/M_PI > 88)
      {
	isCurler=true;
      }
    } 
     
    //Get position of SimTrackerHit
    const double *pos=hit_in->getPosition();

    //Only proceed for hits on readout plane with z>=0
    if(pos[2]>=0) {
    //amplifies hits in the active region of the readout plane

      double hitCoordOne = 0;
      double hitCoordTwo = 0;

      int padLayoutType = _padLayout->getPadLayoutType();
      try{
	switch( padLayoutType ){
	case gear::PadRowLayout2D::CARTESIAN:{  
          hitCoordOne = pos[0];
	  hitCoordTwo = pos[1];
	  break;
	}
	case gear::PadRowLayout2D::POLAR:{
	  complex<double> compHit(pos[0],pos[1]);
	  hitCoordOne = std::abs(compHit);
          hitCoordTwo = std::arg(compHit);
	  if( hitCoordTwo < 0)
	    hitCoordTwo = (2*M_PI)+hitCoordTwo;
	  break;
	}
	default:
	  throw gear::Exception("Unknown pad layout type");
	};
      }catch(Exception& e){
	streamlog_out(ERROR1) << e.what() << endl;
      }

      //Check whether hit is on pad plane
      if( hitCoordOne >= _padLayout->getPlaneExtent()[0] &&
          hitCoordOne <= _padLayout->getPlaneExtent()[1] &&
	  hitCoordTwo >= _padLayout->getPlaneExtent()[2] &&
	  hitCoordTwo <= _padLayout->getPlaneExtent()[3] ) { 

	int nPrimaryElectrons;
	if ( _fullCurlerBackwardCompatibility && isCurler )
	{
	  nPrimaryElectrons = 10;
	}
	else
	{
	  nPrimaryElectrons = static_cast<int>( round( hit_in->getdEdx() /  _ionisationEnergy ) );
	}

	if(nPrimaryElectrons==0) // it might be that some old versions don't set the charge at all
	{
	  nPrimaryElectrons=1;
	}

        int _numberSecEl = CalculateNrElectrons(A1, A2, A3, nPrimaryElectrons);

	//Only proceed if any electrons are left after the amplification process
        if(_numberSecEl>0) {

          //Map position of electron into GEM hole, pitch 140 um
          //pitch in y=sqrt(3/4)*pitch because of hexagonal structure
          double xe_gem=0.;
          double ye_gem = (int)(pos[1]/(sqrt(3./4.)*GEMPITCH)+(pos[1]<0 ? -0.5:0.5))*sqrt(3./4.)*GEMPITCH;
          int yint=(int)(pos[1]/(sqrt(3./4.)*GEMPITCH)+0.5);
          if(yint%2)
            xe_gem = static_cast<int>(pos[0]/GEMPITCH+(pos[0]<0 ? -0.5:0.5))*GEMPITCH;
          else
          //everysecond row is shifted with half a pitch
            xe_gem = (static_cast<int>(pos[0]/GEMPITCH)+(pos[0]<0 ? -0.5:0.5))*GEMPITCH;

          if(fabs(pos[0]-xe_gem)>0.08)
            streamlog_out(WARNING3)<<"x GEM pitch failed, difference: "<<fabs(pos[0]-xe_gem)<<"  "<<pos[0]<<endl;
          if(fabs(pos[1]-ye_gem)>0.08)
            streamlog_out(WARNING3)<<"y GEM pitch failed, difference: "<<fabs(pos[1]-ye_gem)<<"  "<<pos[1]<<endl;
	     
          double position[3];
          position[0]=xe_gem;
          position[1]=ye_gem;
          position[2]=0;

          SimTrackerHitImpl* gemhit = new SimTrackerHitImpl;
          gemhit->setTime( hit_in->getTime() );
          gemhit->setdEdx(_numberSecEl);
          gemhit->setCellID( hit_in->getCellID() );
          gemhit->setPosition( position );
          outputColVec->addElement( gemhit );
        }
      }
    }
  }
  
  //Add output collection to event 
  if(outputColVec->empty()) 
    delete outputColVec;   
  else
    evt->addCollection( outputColVec , _outputCollectionName ) ;
    
  _nEvt++;
}


void GEMProcessor::LoadGEMSettings(){

  if( ! readpar( paramfile.c_str() ) ) {   

    streamlog_out(ERROR4) << "Not all parameters could be read from parameter file " << paramfile << "!" << endl
			  << "Exiting." << endl;      
    exit(1);
  }
  else 
    streamlog_out(DEBUG4) << "Parameter file read successfully " << endl;    

  //set Parameter for GEM-Setting
  Extr[0]=eextr(ETrans1,UGEM1);
  Extr[1]=eextr(ETrans2,UGEM2);
  Extr[2]=eextr(EInd,UGEM3);
  Coll[0]=ecoll(_driftField , UGEM1);
  Coll[1]=ecoll(ETrans1,UGEM2);
  Coll[2]=ecoll(ETrans2,UGEM3);
  Gain[0]=gain(UGEM1 ,_driftField, ETrans1 );
  Gain[1]=gain(UGEM2 ,ETrans1, ETrans2 );
  Gain[2]=gain(UGEM3 ,ETrans2, EInd );
}


int GEMProcessor::GEM_Collection(int GEMNr, int enumber){

  //Collection

  if ( enumber == 0 ) 
    return 0;

  if ( enumber < 0 ) 
    throw std::invalid_argument( "GEMProcessor::GEM_Collection: enumber < 0" );
  
  // Valid input values for CLHEP::RandBinomial::shoot satisfy the relation
  // n*min(p,1-p) > 0. When invalid values are presented, the code silently
  // returns -1.0
  if ( enumber * min( Coll[GEMNr-1], 1 - Coll[GEMNr-1] ) > 0 )
    return static_cast<int>( CLHEP::RandBinomial::shoot( enumber, Coll[GEMNr-1] ) );
  else if ( Coll[GEMNr-1] > 1 || Coll[GEMNr-1] < 0 )
    throw std::out_of_range( "GEMProcessor::GEM_Collection: collection efficiency out of range [0,1]" );
  else if ( Coll[GEMNr-1] > 0.5 ) // p = 1 case (n = 0 case already treated above)
    return enumber;
  else // p = 0 case
    return 0;
}

double GEMProcessor::GEM_Gain(int GEMNr, int enumber){
  //Gain

  double gainmean=Gain[GEMNr-1]/_gainFactor;

  //factor for gain correction to fit data
  //exponential distribution for gain distribution of single electron

  double electrons=0.;

  if( enumber < 200 ){

    for(int i=0;i<enumber;i++)
      electrons += CLHEP::RandExponential::shoot( gainmean );
  }
  else //because than statistics is enough to form gaussian anyway
    electrons = CLHEP::RandGauss::shoot( enumber, sqrt( (double)enumber ) ) * gainmean;
   
  if( electrons >= 0 )
    return electrons;
  else
    return 0;
}


int GEMProcessor::GEM_Extraction(int GEMNr, int enumber){

  //Extraction

  if ( enumber == 0 ) 
    return 0;

  if ( enumber < 0 ) 
    throw std::invalid_argument( "GEMProcessor::GEM_Extraction: enumber <= 0" );
  
  // Valid input values for CLHEP::RandBinomial::shoot satisfy the relation
  // n*min(p,1-p) > 0. When invalid values are presented, the code silently
  // returns -1.0
  if ( enumber * min( Extr[GEMNr-1], 1 - Extr[GEMNr-1] ) > 0 )
    return static_cast<int>( CLHEP::RandBinomial::shoot( enumber, Extr[GEMNr-1] ) );
  else if ( Extr[GEMNr-1] > 1 || Extr[GEMNr-1] < 0 ) 
    throw std::out_of_range( "GEMProcessor::GEM_Extraction: extraction efficiency out of range [0,1]" );
  else if ( Extr[GEMNr-1] > 0.5 ) // p = 1 case
    return enumber;
  else // p = 0 case
    return 0;
}


int  GEMProcessor::CalculateNrElectrons(double attachment1, double attachment2 , double attachment3, int nPrimaryElectrons ){

  //GEM1
  int C1=GEM_Collection(1,nPrimaryElectrons);
  double G1=GEM_Gain(1,C1);
  int E1=GEM_Extraction(1,static_cast<int>(G1));
  int NE1 = static_cast<int>(attachment1*E1);
  
  //GEM2
  int C2=GEM_Collection(2,NE1);
  double G2=GEM_Gain(2,C2);
  int E2=GEM_Extraction(2,static_cast<int>(G2));
  int NE2 = static_cast<int>(attachment2*E2);
  
  //GEM3
  int C3=GEM_Collection(3,NE2);
  double G3=GEM_Gain(3,C3);
  int E3=GEM_Extraction(3,static_cast<int>(G3));
  int NE3 = static_cast<int>(attachment3*E3);

  return NE3;
}


double  GEMProcessor::CalculateAttachment(double E){

  double attachment=0.;
  if(E>1000 && E<10000){

    if(strcmp(Gas,"TDR")==0)
      attachment=AttachmentParam(E,-0.343716243708895,0.000310654263375315, -1.43700791338589e-08,
				 7.13759804300208e-12,-1.20510943273379e-15,4.88792048657487e-20);
	
    if(strcmp(Gas,"P5")==0)
      attachment=AttachmentParam(E,0.569805421256637,-0.00091365767373601,4.37189301033502e-07,
				 -7.06077170050197e-11,4.75904934032274e-15,-1.15317373111484e-19);
	
    if(strcmp(Gas,"P10")==0)
      attachment=AttachmentParam(E,0.383157426324749, -0.000468682252407133,1.45306944419336e-07,
				 -6.60161997388468e-12,-6.98791967319079e-16,4.27452327695256e-20);	
  }

  //0.2 because 2mm between GEMs
  return exp(-0.2*attachment);
}


double  GEMProcessor::AttachmentParam(double E,double a0,double a1,double a2,double a3, double a4,double a5){

  return a0 + a1*E + a2*pow(E,2) + a3*pow(E,3) + a4*pow(E,4) + a5*pow(E,5);
}


void GEMProcessor::end(){

}




}//end of namespace marlintpc
