#ifndef IonsInVoxelsProcessor_h
#define IonsInVoxelsProcessor_h 1

//MARLINTPC
#include "VoxelTPC.h"

//LCIO
#include <lcio.h>

//MARLIN
#include <marlin/Processor.h>

//C++
#include <string>

//GEAR
#include <gear/PadRowLayout2D.h>


namespace marlintpc{

class VoxelTPC;

/* The following comment is tranfered to doxygen documentation
 * which begins with a second asterix, or in case of on line comment an extra third slash 
 */

/** This processor takes all ions from the primary ionisation
 *  and stores there charges in voxels. The binning of the voxels 
 *  is given in the x,y-plane by the pads of the readout plane. The 
 *  length of the voxel in the z-direction can be chosen via the
 *  steering file. For storing and management of the voxels the 
 *  helper class VoxelTPC is used. Output collections contain
 *  the charges of all the ions which emerged to the given event or
 *  all ions which appeared up to the given events in the TPC.
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of SimTrackerHits with charge values set as dEdx
 *
 *  <h4>Output</h4> 
 *  Collection containing the charges of the primary ions in
 *  TPCVoxels. 
 *
 * @param InputCollectionName Name of input collection, SimTrackerHits
 * @param OutputCollectionName Name of output collection, TPCVoxel
 * @param Transient If not 0 the output collection is set transient
 * @param OutputType If not 0 all ions in the TPC are returned
 * @param BinLength Length of the bins in z direction
 *
 * @author Thorsten Krautscheid
 */


class IonsInVoxelsProcessor : public marlin::Processor {

 
 public:
  
  virtual Processor*  newProcessor() { return new IonsInVoxelsProcessor ; }  
  
  IonsInVoxelsProcessor() ;
  virtual ~IonsInVoxelsProcessor(){}
  
  /** Called at the begin of the job before anything is read.
   *  Use to initialize the processor, e.g. book histograms.
   */
  virtual void init() ;
  
  /** Called for every run.
   */
  virtual void processRunHeader( EVENT::LCRunHeader* run ) ;
  
  /** Called for every event - the working horse.
   */
  virtual void processEvent( EVENT::LCEvent * evt ) ; 
    

  /** Function to get the pad corresponding to a hit for 
   *  the different pad layouts
   */
  int getPad(gear::PadRowLayout2D const* padLayout, SimTrackerHit* hit);
  
  
protected:

  //input and output collections
  std::string _inputIonsColName;
  std::string _outputVoxelColName;

  //padLayout pointer
  gear::PadRowLayout2D const * _padLayout;

  //the pad layout type (cartesian or polar)
  int _padLayoutType;

  //pointer to the VoxelTPC which stores the ions
  marlintpc::VoxelTPC* _voxelTPC;

  //input parameters from steering file
  int _transient;
  int _outputType;///< Output type: event based or total.
  int _nEvt;
  double _binLength;///< Length of the zBins.

  //TPC parameters
  double _maxDriftLength;
};

}//end of namespace marlintpc

#endif 
