#ifndef ADCElectronicsListener_h
#define ADCElectronicsListener_h 1

//C++
#include <iostream>
#include <utility>
#include <stdexcept>
#include <map>

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "IMPL/LCCollectionVec.h"
#include "UTIL/LCTime.h"

//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"
#include "lccd/IConditionsChangeListener.hh"
#include "lccd/IConditionsHandler.hh"

//Marlin
#include "marlin/ConditionsProcessor.h"

//TPCCondData
#include "GenericADCElectronicsParameters.h"

// the package schould be capsulated in a name space
namespace marlintpc
{

/** Helper Class which provides access to the parameters of ADC Electroncis.
 *  Its designed as interface for different ADC electronics conditions objects.
 *  It is derived from IConditionsChangeListener,
 *  The parameters are updated every time the conditions data change.
 * @author R. Diener (DESY)
 */

class ADCElectronicsListener : public lccd::IConditionsChangeListener
{

    public:

        /// Map of module GenericADCElectronicsParameters
        typedef std::map< int, tpcconddata::GenericADCElectronicsParameters* > GenericADCElectronicsParametersMap_t;

        /** Constructor
         * @param ColName name of the collection in which the ADC electronics parameters are stored
         */
        ADCElectronicsListener(std::string ConditionsName, std::string ColName);

        /// Destructor
        virtual ~ADCElectronicsListener();

        /// Returns the default name of the conditions data collection.
        static std::string getDefaultColName()
        {
            return  tpcconddata::GenericADCElectronicsParameters::getDefaultColName();
        };

        /// Updates the stored values if conditions have changed.
        /** Called by the Marlin ConditionsProcessor, if the condition change
         *  inplemetentation of the IConditionsChangeListener interface
         */
        void conditionsChanged(lcio::LCCollection* col);

        /// Implementation of the listener class, takes care of default collection settings and registration of the handler.
        void registeredWithHandler(lccd::IConditionsHandler* condHandler);

        /// Implementation of the listener class, cleans up after deregistration and gives warning message.
        void deRegisteredWithHandler(lccd::IConditionsHandler* condHandler);

        /// Get the polarity value, should be -1 or +1.
        int getPolarity(int readoutGroupID);
        /// Get the precision value of the electronics (maximum possible ADC value).
        int getMaxADCValue(int readoutGroupID);
        /// Get the read out frequency of the read-out electronics (in MHz).
        float getFrequency(int readoutGroupID);
        /// returns true, if readout electronics is automatically substracting pedestals
        bool isPedestalSubstracted(int readoutGroupID);
        /// returns true, if readout electronics does zero suppression
        bool isZeroSupressed(int readoutGroupID);

        /// Gget the validity start time of the current conditions collection or default collection.
        lcio::long64 collectionValidSince();
        /// Gget the validity end time of the current conditions collection or default collection.
        lcio::long64 collectionValidTill();

        /// Set if an exception should the thrown in case of no data available in the database.
        void useLastValidIfNoDataAvailable(bool useLastValid);
        /// Rreturn if currently data is available in the database/file or not.
        bool noConditionsDataAvailable();

    protected:

    private:

        /// Fills the internal conditions map from the given collection.
        void fillConditionsMap(lcio::LCCollection* col);

        /// Holds the name of the conditions data collection
        std::string _collectionName;
        /// Holds the name of the conditions. Currently, only "GenericADCElectronicsParameters" is implemented.
        std::string _conditionsName;

        /// Default collection that is registered and will be returned from the handler if no data is available.
        lcio::LCCollectionVec* _thisEmptyCollection;

        /// Map of registered default collections for differnet handlers. Will hold here only one entry, since this listener registers itself with only one handler.
        std::map<lcio::LCCollection* , lccd::IConditionsHandler* > _handlerDefaultCollectionMap;

        /// Conditions object for the GenericADCElectronicsParameters type.
        tpcconddata::GenericADCElectronicsParameters* _genericADCElectronicsParameters;
        /// Map for conditions objects of the GenericADCElectronicsParameters type.
        GenericADCElectronicsParametersMap_t _genericADCElectronicsParametersMap;

        /// Internal member to hold the "since" of the validity range of the current conditions collection.
        lcio::long64 _timeStampSince;
        /// Internal member to hold the "till" of the validity range of the current conditions collection.
        lcio::long64 _timeStampTill;

        /// Used to set if the last valid collection should be used if no current data are in the database.
        bool _useLastValidIfNoDataAvailable;
        /// Holds the indormation, if there are no current data available in the database.
        bool _noDataAvailable;

        /// Pointer to hold the address of the handler this listener is registered with.
        lccd::IConditionsHandler* _condHandler;
};

}
#endif
