#ifndef ChannelMappingProcessor_h
#define ChannelMappingProcessor_h 1

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "marlin/Exceptions.h"

// MarlinTPC
#include "ADCChannelMapping.h"
#include "ChannelMappingListener.h"

// C++
#include <string>


/// the MarlinTPC Package: reconstruction of TPC data
namespace marlintpc
{

/** Channel Mapping processor of the MarlinTPC package.
 *
 *  This processor translates the channel mapping from hardware channel and readout group to gear pad ID and module ID.
 *  Mapping is provided via LCCD (simple file, database file or database).
 *
 *  <h4>Input - Prerequisites</h4>
 *  \li A collection of TrackerData
 *  \li A collection with the mapping (ADCChannelMapping)
 *  <h4>Output</h4>
 *  Collection of TrackerData.
 *
 * @param InputCollectionName Name of input data collection (default: TPCPulses)
 * @param OutputCollectionName \c optional Name of output data collection (optional, if not given input collection is modified)
 * @param WriteOutputToStorage:bool output collection will be written if set to true (which is the default)
 *   Only applies if the \c OutputCollectionName is given and a new collection is created.
 *   If the input collection is modified, the status from the creating processor is preserved.
 * @param ChannelMappingCollectionName Name of the collection containing the channel mapping (default: TPCChannelMapping)
 * @param RemoveChannelsWithoutMapping If "true", found signals without mapping in the conditions data will be removed from collection,
 *   otherwise a tpcconddata::CondDataNotAvailableException exception is thrown an processing will stop
 * @param MissingConditionsDataOverrideTimespan \c optional  Sets the time in seconds that missing conditions data will be ignored by using the last valid conditions data collection (optional parameter, if not set, processing stops if no conditions data available for event time).
**/
/**
*
* @author R. Diener, DESY
*/

class ChannelMappingProcessor : public marlin::Processor
{

    public:

        virtual Processor*  newProcessor()
        {
            return new ChannelMappingProcessor ;
        }

        ChannelMappingProcessor() ;

        virtual void init() ;

        virtual void processRunHeader(lcio::LCRunHeader* run) ;

        virtual void processEvent(lcio::LCEvent * evt) ;

        virtual void check(lcio::LCEvent * evt) ;

        virtual void end() ;


    protected:

        /// the name of the input collection
        std::string _inputColName ;
        /// the name of the output collection
        std::string _outputColName ;
        /// the name of the ChannelMapping collection
        std::string _channelMappingColName ;

        /// set the persistency flag of the output collection
        bool _outputIsPersistent;

        /** Is true till the first output collection is written to an event
         *  used to to store parameters only in the first collection
         *  (also done, if collection is transient)
         */
        bool isFirstCollection;

        /** The timepan in seconds for which the last valid collection will be used if no conditions data are available via LCCD.
         */
        int _missingConditionsDataOverrideTimespan;
        bool _removeChannelsWithoutMapping;
    private:

        ChannelMappingListener* _ChannelMappingListener;

        /// Internal member to hold the time for the missing conditions overide. Unit: nanoseconds.
        lcio::long64 _missingConditionsDataOverrideLCIOTime;

        //fg: add a map of channels that have been removed from the collection as
        //    their hardware address is not mapped to any known software address (padID)
        typedef std::map< std::pair< int, int > , unsigned > MAP ;
        MAP _cMap ;
} ;
}
#endif
