#ifndef GAINCORRECTORPROCESSOR_H
#define GAINCORRECTORPROCESSOR_H 1

// MarlinTPC
#include "ChannelCorrectionListener.h"

// Marlin
#include <marlin/Processor.h>

// LCIO
#include <lcio.h>

// stl
#include <string>

namespace marlintpc
{

/** GainCorrectorProcessor: Applies a gain correction to LCIO::TrackerPulse collection by dividing the stored charge values by the factor given in the ChannelCorrection conditions collection.
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of TrackerData
 *
 *  <h4>Output</h4>
 *  Collection of TrackerData.
 *
 * @param InputCollectionName Name of input data collection
 * @param OutputCollectionName Name of output data collection
 * @param WriteOutputToStorage:bool output collection will be written if set to true (which is the default)
 * @param ChannelCorrectionCollectionName Name of the conditions data collection containing the channel correction values
 * @param TypeOfIndex Type of the cellID0 and cellID1 indices (readout electronics or gear)
 * @param MissingConditionsDataOverrideTimespan Sets the time in seconds that missing conditions data will be ignored by using the last valid conditions data collection (optional parameter, if not set, processing stops if no conditions data available for event time)
 **/
/**
 *
 * @author R. Diener, DESY; P. Wienemann, University of Bonn
 */

class GainCorrectorProcessor : public marlin::Processor
{

    public:
        /**
         * Constructor
         */
        GainCorrectorProcessor() ;

        /**
         * Overload newProcessor() method of marlin::Processor base class
         * to get a new instance of this processor
         */
        virtual Processor*  newProcessor()
        {
            return new GainCorrectorProcessor();
        }

        virtual void init() ;

        /**
         * Write information into the run header
         *
         * @param run pointer to LCRunHeader to be processed
         */

        virtual void processRunHeader(lcio::LCRunHeader* run);

        /**
         * Called for every event - the working horse.
         *
         * @param run pointer to LCEvent to be processed
         */

        virtual void processEvent(lcio::LCEvent * evt);

        /**
         * Optional checks
         *
         * @param run pointer to LCEvent to be checked
         */

        virtual void check(lcio::LCEvent * evt);

        /**
         * Called after data processing for clean up.
         */

        virtual void end();

    private:

        /** the name of the collection the input pulses will be stored under
         */
        std::string _inputColName;

        /** the name of the collection the output pulses will be stored under
         */
        std::string _outputColName;

        /** the name of the collection the channel corrections are stored with
         */
        std::string _channelCorrectionColName;

        // how many runs / events have been processed sofar
        int _number_of_runs;
        int _number_of_events;

        /** the conditions listener for the channel corrections
         */
        ChannelCorrectionListener* _channelCorrectionListener;
        /** The timepan in seconds for which the last valid collection will be used if no conditions data are available via LCCD.
         */
        int _missingConditionsDataOverrideTimespan;

        /// set the persistency flag of the output collection
        bool _outputIsPersistent;

        /// type of the index (readout electronics or gear)
        std::string _indexType;
        /// Internal member to hold the time for the missing conditions overide. Unit: nanoseconds.
        lcio::long64 _missingConditionsDataOverrideLCIOTime;

};

}
#endif                           // GAINCORRECTORPROCESSOR_H
