#ifndef TIMESHIFTCORRECTORPROCESSOR_H
#define TIMESHIFTCORRECTORPROCESSOR_H

// Marlin
#include "marlin/Processor.h"
#include "marlin/Exceptions.h"
// LCIO
#include "lcio.h"
// C++
#include <string>
// MarlinTPC
#include "ChannelCorrection.h"
#include "ChannelCorrectionListener.h"

namespace marlintpc
{

/**
 * Processor to apply time offset correction to individual channels.
 * The offset for each channel is read from LCCD via a ConditionsListener.
 *
 * @author Ralf Diener, DESY; Peter Wienemann, University of Bonn
 */

/** Pedestal subtracting processor of the MarlinTPC package.
 *
 *  Thisprocessor substracts pedestals provided via LCCD (simple file, database file or database).
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of TrackerData
 *
 *  <h4>Output</h4>
 *  Collection of TrackerData.
 *
 * @param InputCollectionName Name of input data collection
 * @param OutputCollectionName Name of output data collection
 * @param WriteOutputToStorage:bool output collection will be written if set to true (which is the default)
 * @param ChannelCorrectionCollectionName Name of the conditions data collection containing the channel correction values
 * @param TypeOfIndex Type of the cellID0 and cellID1 indices (readout electronics or gear)
 * @param MissingConditionsDataOverrideTimespan Sets the time in seconds that missing conditions data will be ignored by using the last valid conditions data collection (optional parameter, if not set, processing stops if no conditions data available for event time)
 **/
/**
 *
 * @author R. Diener, DESY; P. Wienemann, University of Bonn
 */

class TimeShiftCorrectorProcessor : public marlin::Processor
{
    public:

        /**
         * Constructor
         */
        TimeShiftCorrectorProcessor();

        /**
         * Overload newProcessor() method of marlin::Processor base class
         * to get a new instance of this processor
         */
        virtual marlin::Processor* newProcessor()
        {
            return new TimeShiftCorrectorProcessor;
        }

        /**
         * Write information into the run header
         *
         * @param run pointer to LCRunHeader to be processed
         */
        virtual void processRunHeader(EVENT::LCRunHeader* run);

        virtual void init();

        /**
         * Called for every event - the working horse.
         *
         * @param run pointer to LCEvent to be processed
         */
        virtual void processEvent(EVENT::LCEvent* evt);

        /**
         * Optional checks
         *
         * @param run pointer to LCEvent to be checked
         */
        virtual void check(lcio::LCEvent* evt)
        {
            ;
        }

        /**
         * Called after data processing for clean up.
         */
        virtual void end();

    protected:

        ChannelCorrectionListener* _channelCorrectionListener;

        /// the name of the input collection
        std::string _inputColName ;
        /// the name of the output collection
        std::string _outputColName ;
        /// the name of the channelcorrection collection
        std::string _channelCorrectionColName ;
        /// type of the index (readout electronics or gear)
        std::string _indexType;
        /** The timepan in seconds for which the last valid collection will be used if no conditions data are available via LCCD.
         */
        int _missingConditionsDataOverrideTimespan;

        /// set the persistency flag of the output collection
        bool _outputIsPersistent;

    private:
        /// Internal member to hold the time for the missing conditions overide. Unit: nanoseconds.
        lcio::long64 _missingConditionsDataOverrideLCIOTime;

};

}                                // namespace marlintpc
#endif                           // TIMESHIFTCORRECTOR_H
