#include "ChannelCorrectionListener.h"

//C++
//#include <utility>
#include <map>
#include <utility>
#include <iostream>
#include <string.h>
#include <exception>
#include <sstream>

//lccd
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include <lccd_exceptions.h>

//lcio
#include "lcio.h"
#include "EVENT/LCIO.h"
#include "UTIL/LCFixedObject.h"
#include "UTIL/LCTime.h"

//MarlinTPC
#include "TPCCondDataExeptions.h"

namespace marlintpc
{

ChannelCorrectionListener::ChannelCorrectionListener(std::string ColName)
{
    //initialize first the members, else it goes wrong
    _collectionName = ColName;
    _readoutChannelCorrectionMap.clear();
    _gearChannelCorrectionMap.clear();
    _thisEmptyCollection = new LCCollectionVec(LCIO::LCGENERICOBJECT);
    _thisEmptyCollection->parameters().setValue("CollectionName", "EmptyCollectionChannelCorrectionListener");
    _timeStampSince = 0;
    _timeStampTill = 0;
    _useLastValidIfNoDataAvailable = false;
    _noDataAvailable = true;
    _condHandler = NULL;
    // try to register with the Marlin conditions processor
    if (!(marlin::ConditionsProcessor::registerChangeListener(this, ColName)))
        {
            std::stringstream message;
            message << " ChannelCorrectionListener couldn't register collection " << ColName << " with the ConditionsProcessor.";
            throw lccd::LCCDException(message.str());
        }
}



ChannelCorrectionListener::~ChannelCorrectionListener()
{
    delete _thisEmptyCollection;
}



void ChannelCorrectionListener::conditionsChanged(EVENT::LCCollection* col)
{
    m_out(DEBUG) << "ChannelCorrectionListener::conditionsChanged()" << std::endl;

    // get collection time parameters and set internal members
    std::stringstream sstrSince(col->getParameters().getStringVal(lccd::DBSINCE));
    sstrSince >> _timeStampSince;
    lcio::LCTime timeSince(_timeStampSince) ;
    std::stringstream sstrTill(col->getParameters().getStringVal(lccd::DBTILL));
    sstrTill >> _timeStampTill;
    lcio::LCTime timeTill(_timeStampTill) ;
    m_out(DEBUG) << "ChannelCorrectionListener::conditionsChanged(): _timeStampSince=" << _timeStampSince << " _timeStampTill=" << _timeStampTill << std::endl;

    // look into the map to see if we have accepted this collection as a default
    std::map<lcio::LCCollection* , lccd::IConditionsHandler* >::iterator itDefaultCollectionMap = _handlerDefaultCollectionMap.find(col);

    // check if the collection is our default collection
    if (itDefaultCollectionMap != _handlerDefaultCollectionMap.end())
        {
            _noDataAvailable = true;

            m_out(MESSAGE) << "ChannelCorrectionListener::conditionsChanged(): default collection " << col->getParameters().getStringVal("CollectionName") << " sent " << std::endl;
            m_out(DEBUG) << "ChannelCorrectionListener::conditionsChanged(): address of default collection: " << col << std::endl;

            //give some information on the missing conditions data
            m_out(DEBUG) << "ChannelCorrectionListener::conditionsChanged(): CollectionName: " << _collectionName << " not available at this timestamp." << std::endl;
            m_out(DEBUG) << "ChannelCorrectionListener::conditionsChanged(): No data in database from: " << timeSince.timeStamp() << " (" << timeSince.year() << " " << timeSince.month() << " " << timeSince.day() << " " << timeSince.hour() << " " << timeSince.min() << " " << timeSince.sec() << " " << timeSince.ns() << ")" << std::endl;
            m_out(DEBUG) << "ChannelCorrectionListener::conditionsChanged(): No data in database till: " << timeTill.timeStamp() << " (" << timeTill.year() << " " << timeTill.month() << " " << timeTill.day() << " " << timeTill.hour() << " " << timeTill.min() << " " << timeTill.sec() << " " << timeTill.ns() << ")" << std::endl;

            //check what to do if there are not data available
            if (_useLastValidIfNoDataAvailable)
                {
                    // check if there has been a last valid collection
                    if (_condHandler->lastValidCollection() != NULL)
                        {
                            fillConditionsMaps(_condHandler->lastValidCollection());
                        }
                    else
                        {
                            // no last valid collection available, so we bail out by throwing an exception
                            std::stringstream out;
                            out << "ChannelCorrectionListener::conditionsChanged(): Tried to use last valid collection but found none. Exiting processing";
                            throw lccd::DataNotAvailableException(out.str());
                        }
                }
            else
                {
                    // no conditions data available, so we bail out by throwing an exception
                    std::stringstream out;
                    out << "ChannelCorrectionListener::conditionsChanged(): No conditions data stored for current time stamp. Exiting processing";
                    throw lccd::DataNotAvailableException(out.str());
                }
        }
    else                     // it is not a default so we can do the updating
        {
            _noDataAvailable = false;

            m_out(MESSAGE) << "ChannelCorrectionListener::conditionsChanged(): new values available. Updating parameter map." << std::endl;
            m_out(DEBUG) << "ChannelCorrectionListener::conditionsChanged(): CollectionName: " << col->getParameters().getStringVal("CollectionName") << std::endl;
            m_out(DEBUG) << "ChannelCorrectionListener::conditionsChanged(): Valid Since: " << col->getParameters().getStringVal("DBSince") << std::endl;
            m_out(DEBUG) << "ChannelCorrectionListener::conditionsChanged(): Valid Till: " << col->getParameters().getStringVal("DBTill") << std::endl;

            fillConditionsMaps(col);
        }
}



void ChannelCorrectionListener::fillConditionsMaps(lcio::LCCollection* col)
{
    _readoutChannelCorrectionMap.clear();
    _gearChannelCorrectionMap.clear();
    int nEntries = col->getNumberOfElements();

    tpcconddata::ChannelCorrection*    Entry = 0;
    std::pair<ChannelCorrectionMap_t::iterator, bool> iInsertReturn;
    m_out(DEBUG) << "ChannelCorrectionListener: update ChannelCorrections  (data for " << nEntries << " channels in collection)" << std::endl;

    // fill LCCD collection into the map(map(ChannelCorrection)) structure
    for (int iEntry = 0; iEntry < nEntries; ++iEntry)
        {
            Entry = new tpcconddata::ChannelCorrection(dynamic_cast<LCGenericObject*>(col->getElementAt(iEntry)));
            if (((Entry->getHardwareID()).first != -1) && ((Entry->getHardwareID()).second != -1))
                {
                    iInsertReturn = _readoutChannelCorrectionMap.insert(std::make_pair(Entry->getHardwareID(), Entry));
                    if (!iInsertReturn.second)
                        {
                            m_out(ERROR) << "ChannelCorrectionListener: could not store ChannelCorrection for hardwareID:     (" << (Entry->getHardwareID()).first << "," << (Entry->getHardwareID()).second << ") in the readoutChannelCorrection map because there was already a value stored for this channel" << std::endl;
                        }
                }
            if (((Entry->getGearID()).first != -1) && ((Entry->getGearID()).second != -1))
                {
                    iInsertReturn = _gearChannelCorrectionMap.insert(std::make_pair(Entry->getGearID(), Entry));
                    if (!iInsertReturn.second)
                        {
                            m_out(ERROR) << "ChannelCorrectionListener: could not store ChannelCorrection for gearID: (" << (Entry->getGearID()).first << "," << (Entry->getGearID()).second << ") in the gearChannelCorrection map because there was already a value stored for this channel" << std::endl;
                        }
                }
        } //loop over all entries in the LCCollectionVec
}

std::pair<int, int> ChannelCorrectionListener::getGearMapping(Int_Pair_t hardwareIDpair)
{
    tpcconddata::ChannelCorrection* tmpChannelCorrection;
    try
        {
            tmpChannelCorrection = this->getChannelCorrectionFromHardwareID(hardwareIDpair);
        }
    catch (tpcconddata::CondDataNotAvailableException& e)
        {
            throw e;
        }
    return tmpChannelCorrection->getGearID();
}

std::pair<int, int> ChannelCorrectionListener::getHardwareMapping(Int_Pair_t gearIDpair)
{
    tpcconddata::ChannelCorrection* tmpChannelCorrection;
    try
        {
            tmpChannelCorrection = this->getChannelCorrectionFromGearID(gearIDpair);
        }
    catch (tpcconddata::CondDataNotAvailableException& e)
        {
            throw e;
        }
    return tmpChannelCorrection->getHardwareID();
}


tpcconddata::ChannelCorrection* ChannelCorrectionListener::getChannelCorrectionFromHardwareID(Int_Pair_t hardwareIDpair)
{
    ChannelCorrectionMap_t::iterator result = _readoutChannelCorrectionMap.find(hardwareIDpair);
    if (result != _readoutChannelCorrectionMap.end())
        {
            return dynamic_cast<tpcconddata::ChannelCorrection*>((*result).second);
        }
    else
        {
            std::stringstream message;
            message << "No entry could be found for channel/readgroup: " << hardwareIDpair.first << "/" << hardwareIDpair.second << ".";
            throw tpcconddata::CondDataNotAvailableException(message.str());
        }
}



tpcconddata::ChannelCorrection* ChannelCorrectionListener::getChannelCorrectionFromGearID(Int_Pair_t gearIDpair)
{
    ChannelCorrectionMap_t::iterator result = _gearChannelCorrectionMap.find(gearIDpair);
    if (result != _gearChannelCorrectionMap.end())
        {
            return dynamic_cast<tpcconddata::ChannelCorrection*>((*result).second);
        }
    else
        {
            std::stringstream message;
            message << "No entry could be found for pad/module: " << gearIDpair.first << "/" << gearIDpair.second << ".";
            throw tpcconddata::CondDataNotAvailableException(message.str());
        }
}



float ChannelCorrectionListener::getCalibrationFactorFromGearID(Int_Pair_t gearIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromGearID(gearIDpair);
    return Entry->getCalibrationFactor();
}



float ChannelCorrectionListener::getCalibrationFactorFromHardwareID(Int_Pair_t hardwareIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromHardwareID(hardwareIDpair);
    return Entry->getCalibrationFactor();
}



float ChannelCorrectionListener::getTimeOffsetFromGearID(Int_Pair_t gearIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromGearID(gearIDpair);
    return Entry->getTimeOffset();
}



float ChannelCorrectionListener::getTimeOffsetFromHardwareID(Int_Pair_t hardwareIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromHardwareID(hardwareIDpair);
    return Entry->getTimeOffset();
}



int   ChannelCorrectionListener::getQualityFromGearID(Int_Pair_t gearIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromGearID(gearIDpair);
    return Entry->getQuality();
}



int   ChannelCorrectionListener::getQualityFromHardwareID(Int_Pair_t hardwareIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromHardwareID(hardwareIDpair);
    return Entry->getQuality();
}



bool  ChannelCorrectionListener::isDeadFromGearID(Int_Pair_t gearIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromGearID(gearIDpair);
    return Entry->isDead();
}



bool  ChannelCorrectionListener::isDeadFromHardwareID(Int_Pair_t hardwareIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromHardwareID(hardwareIDpair);
    return Entry->isDead();
}



bool  ChannelCorrectionListener::isNoisyFromGearID(Int_Pair_t gearIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromGearID(gearIDpair);
    return Entry->isNoisy();
}



bool  ChannelCorrectionListener::isNoisyFromHardwareID(Int_Pair_t hardwareIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromHardwareID(hardwareIDpair);
    return Entry->isNoisy();
}



bool  ChannelCorrectionListener::isAddonBitASetFromGearID(Int_Pair_t gearIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromGearID(gearIDpair);
    return Entry->isAddonBitASet();
}



bool  ChannelCorrectionListener::isAddonBitASetFromHardwareID(Int_Pair_t hardwareIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromHardwareID(hardwareIDpair);
    return Entry->isAddonBitASet();
}



bool  ChannelCorrectionListener::isAddonBitBSetFromGearID(Int_Pair_t gearIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromGearID(gearIDpair);
    return Entry->isAddonBitBSet();
}



bool  ChannelCorrectionListener::isAddonBitBSetFromHardwareID(Int_Pair_t hardwareIDpair)
{
    tpcconddata::ChannelCorrection* Entry = getChannelCorrectionFromHardwareID(hardwareIDpair);
    return Entry->isAddonBitBSet();
}



void ChannelCorrectionListener::registeredWithHandler(lccd::IConditionsHandler* condHandler)
{
    _condHandler = condHandler;
    m_out(MESSAGE) << "ChannelCorrectionListener::registeredWithHandler(): registered: " << condHandler->name() << std::endl;
    m_out(DEBUG) << "ChannelCorrectionListener::registeredWithHandler(): register default collection:" << std::endl;
    // try to get the default collection
    try
        {
            LCCollection* col = condHandler->defaultCollection();

            if (! col)           // it will be null if none has so far been registered. So let's register ours
                {
                    condHandler->registerDefaultCollection(_thisEmptyCollection);
                    m_out(DEBUG) << "ChannelCorrectionListener::registeredWithHandler(): default collection registered; address: " << _thisEmptyCollection << std::endl;
                    // and put in the map for this handler
                    _handlerDefaultCollectionMap[_thisEmptyCollection] = condHandler;
                }
            //then the default handler was already registered, that's odd ... ;)
            else if (col == _thisEmptyCollection)
                {
                    m_out(DEBUG) << "ChannelCorrectionListener::registeredWithHandler(): default collection is already set to thisEmptyCollection:" << std::endl;
                }
            else                 // somebody has got there before us, let's see if we like the default
                {
                    // if it has the same CollectionName, we can use it, no problem ...
                    if (col->getParameters().getStringVal("CollectionName") == "EmptyCollectionChannelCorrectionListener")
                        {
                            m_out(DEBUG) << "ChannelCorrectionListener::registeredWithHandler(): I like your default ;)" << std::endl;
                            _handlerDefaultCollectionMap[col] = condHandler;
                        }
                    // if not, an exception is thrown because things should be handled in a uniform standardized way
                    else
                        {
                            m_out(ERROR) << "ChannelCorrectionListener::registeredWithHandler(): a non-standard empty default collection has already been registered with the conditions handler" << std::endl;
                            throw lccd::InconsistencyException("ChannelCorrectionListener::registeredWithHandler() found a default collection that differs from the standard one has already been registered with the handler");
                        }
                }
        }
    // catch MemberNotImplemented exception, which happens, if a simple file is used as input
    catch (lccd::MemberNotImplementedException)
        {
            m_out(MESSAGE) << "ChannelCorrectionListener::registeredWithHandler: Could not register default collection. Seems simple file input is used for ChannelCorrection conditions." << std::endl;
        }
}



void ChannelCorrectionListener::deRegisteredWithHandler(lccd::IConditionsHandler* condHandler)
{
    m_out(WARNING) << "ChannelCorrectionListener::deRegisteredWithHandler() deregistered with:" << condHandler->name() << std::endl;
    _condHandler = NULL;
}



lcio::long64 ChannelCorrectionListener::collectionValidSince()
{
    return _timeStampSince;
}



lcio::long64 ChannelCorrectionListener::collectionValidTill()
{
    return _timeStampTill;
}



void ChannelCorrectionListener::useLastValidIfNoDataAvailable(bool useLastValid)
{
    _useLastValidIfNoDataAvailable = useLastValid;
}



bool ChannelCorrectionListener::noConditionsDataAvailable()
{
    return _noDataAvailable;
}


}
