#include "ChannelMappingListener.h"

//C++
//#include <utility>
#include <map>
#include <utility>
#include <iostream>
#include <string.h>
#include <exception>
#include <sstream>

//lccd
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include <lccd_exceptions.h>

//lcio
#include "lcio.h"
#include "EVENT/LCIO.h"
#include "UTIL/LCFixedObject.h"
#include "UTIL/LCTime.h"

//MarlinTPC
#include "TPCCondDataExeptions.h"

namespace marlintpc
{

ChannelMappingListener::ChannelMappingListener(std::string ColName)
{
    //initialize first the members, else it goes wrong
    _collectionName = ColName;
    _readoutChannelMappingMap.clear();
    _gearChannelMappingMap.clear();
    _thisEmptyCollection = new LCCollectionVec(LCIO::LCGENERICOBJECT);
    _thisEmptyCollection->parameters().setValue("CollectionName", "EmptyCollectionChannelMappingListener");
    _timeStampSince = 0;
    _timeStampTill = 0;
    _useLastValidIfNoDataAvailable = false;
    _noDataAvailable = true;
    _condHandler = NULL;
    // try to register with the Marlin conditions processor
    if (!(marlin::ConditionsProcessor::registerChangeListener(this, ColName)))
        {
            std::stringstream message;
            message << " ChannelMappingListener couldn't register collection " << ColName << " with the ConditionsProcessor.";
            throw lccd::LCCDException(message.str());
        }
}



ChannelMappingListener::~ChannelMappingListener()
{
    delete _thisEmptyCollection;
}



void ChannelMappingListener::conditionsChanged(EVENT::LCCollection* col)
{
    m_out(DEBUG) << "ChannelMappingListener::conditionsChanged()" << std::endl;

    // get collection time parameters and set internal members
    std::stringstream sstrSince(col->getParameters().getStringVal(lccd::DBSINCE));
    sstrSince >> _timeStampSince;
    lcio::LCTime timeSince(_timeStampSince) ;
    std::stringstream sstrTill(col->getParameters().getStringVal(lccd::DBTILL));
    sstrTill >> _timeStampTill;
    lcio::LCTime timeTill(_timeStampTill) ;
    m_out(DEBUG) << "ChannelMappingListener::conditionsChanged(): _timeStampSince=" << _timeStampSince << " _timeStampTill=" << _timeStampTill << std::endl;

    // look into the map to see if we have accepted this collection as a default
    std::map<lcio::LCCollection* , lccd::IConditionsHandler* >::iterator itDefaultCollectionMap = _handlerDefaultCollectionMap.find(col);

    // check if the collection is our default collection
    if (itDefaultCollectionMap != _handlerDefaultCollectionMap.end())
        {
            _noDataAvailable = true;

            m_out(MESSAGE) << "ChannelMappingListener::conditionsChanged(): default collection " << col->getParameters().getStringVal("CollectionName") << " sent " << std::endl;
            m_out(DEBUG) << "ChannelMappingListener::conditionsChanged(): address of default collection: " << col << std::endl;

            //give some information on the missing conditions data
            m_out(DEBUG) << "ChannelMappingListener::conditionsChanged(): CollectionName: " << _collectionName << " not available at this timestamp." << std::endl;
            m_out(DEBUG) << "ChannelMappingListener::conditionsChanged(): No data in database from: " << timeSince.timeStamp() << " (" << timeSince.year() << " " << timeSince.month() << " " << timeSince.day() << " " << timeSince.hour() << " " << timeSince.min() << " " << timeSince.sec() << " " << timeSince.ns() << ")" << std::endl;
            m_out(DEBUG) << "ChannelMappingListener::conditionsChanged(): No data in database till: " << timeTill.timeStamp() << " (" << timeTill.year() << " " << timeTill.month() << " " << timeTill.day() << " " << timeTill.hour() << " " << timeTill.min() << " " << timeTill.sec() << " " << timeTill.ns() << ")" << std::endl;

            //check what to do if there are not data available
            if (_useLastValidIfNoDataAvailable)
                {
                    // check if there has been a last valid collection
                    if (_condHandler->lastValidCollection() != NULL)
                        {
                            fillConditionsMaps(_condHandler->lastValidCollection());
                        }
                    else
                        {
                            // no last valid collection available, so we bail out by throwing an exception
                            std::stringstream out;
                            out << "ChannelMappingListener::conditionsChanged(): Tried to use last valid collection but found none. Exiting processing";
                            throw lccd::DataNotAvailableException(out.str());
                        }
                }
            else
                {
                    // no conditions data available, so we bail out by throwing an exception
                    std::stringstream out;
                    out << "ChannelMappingListener::conditionsChanged(): No conditions data stored for current time stamp. Exiting processing";
                    throw lccd::DataNotAvailableException(out.str());
                }
        }
    else                     // it is not a default so we can do the updating
        {
            _noDataAvailable = false;

            m_out(MESSAGE) << "ChannelMappingListener::conditionsChanged(): new values available. Updating parameter map." << std::endl;
            m_out(DEBUG) << "ChannelMappingListener::conditionsChanged(): CollectionName: " << col->getParameters().getStringVal("CollectionName") << std::endl;
            m_out(DEBUG) << "ChannelMappingListener::conditionsChanged(): Valid Since: " << col->getParameters().getStringVal("DBSince") << std::endl;
            m_out(DEBUG) << "ChannelMappingListener::conditionsChanged(): Valid Till: " << col->getParameters().getStringVal("DBTill") << std::endl;

            fillConditionsMaps(col);
        }
}



void ChannelMappingListener::fillConditionsMaps(lcio::LCCollection* col)
{
    _readoutChannelMappingMap.clear();
    _gearChannelMappingMap.clear();
    int nEntries = col->getNumberOfElements();

    tpcconddata::ADCChannelMapping*    Entry = 0;
    std::pair<ChannelMappingMap_t::iterator, bool> iInsertReturn;
    m_out(DEBUG) << "ChannelMappingListener: update ChannelMappings  (data for " << nEntries << " channels in collection)" << std::endl;

    // fill LCCD collection into the map(map(ChannelMapping)) structure
    for (int iEntry = 0; iEntry < nEntries; ++iEntry)
        {
            Entry = new tpcconddata::ADCChannelMapping(dynamic_cast<LCGenericObject*>(col->getElementAt(iEntry)));
            iInsertReturn = _readoutChannelMappingMap.insert(std::make_pair(std::make_pair(Entry->getChannelID(), Entry->getReadoutGroup()), Entry));
            if (!iInsertReturn.second)
                {
                    m_out(ERROR) << "ChannelMappingListener: could not store ChannelMapping for hardwareID: (" << Entry->getChannelID() << "," << Entry->getReadoutGroup() << ") in the readoutChannelMapping map because there was already a value stored for this channel" << std::endl;
                }
            iInsertReturn = _gearChannelMappingMap.insert(std::make_pair(std::make_pair(Entry->getPadID(), Entry->getModuleID()), Entry));
            if (!iInsertReturn.second)
                {
                    m_out(ERROR) << "ChannelMappingListener: could not store ChannelMapping for gearID: (" << Entry->getPadID() << "," << Entry->getModuleID() << ") in the gearChannelMapping map because there was already a value stored for this channel" << std::endl;
                }
        } //loop over all entries in the LCCollectionVec
}

std::pair<int, int> ChannelMappingListener::getGearMapping(Int_Pair_t hardwareIDpair)
{
    tpcconddata::ADCChannelMapping* tmpChannelMapping;
    try
        {
            tmpChannelMapping = this->getADCChannelMappingFromHardwareID(hardwareIDpair);
        }
    catch (tpcconddata::CondDataNotAvailableException& e)
        {
            throw e;
        }
    return std::make_pair(tmpChannelMapping->getPadID(), tmpChannelMapping->getModuleID());
}

std::pair<int, int> ChannelMappingListener::getHardwareMapping(Int_Pair_t gearIDpair)
{
    tpcconddata::ADCChannelMapping* tmpChannelMapping;
    try
        {
            tmpChannelMapping = this->getADCChannelMappingFromGearID(gearIDpair);
        }
    catch (tpcconddata::CondDataNotAvailableException& e)
        {
            throw e;
        }
    return std::make_pair(tmpChannelMapping->getChannelID(), tmpChannelMapping->getReadoutGroup());
}


tpcconddata::ADCChannelMapping* ChannelMappingListener::getADCChannelMappingFromHardwareID(Int_Pair_t hardwareIDpair)
{
    ChannelMappingMap_t::iterator result = _readoutChannelMappingMap.find(hardwareIDpair);
    if (result != _readoutChannelMappingMap.end())
        {
            return dynamic_cast<tpcconddata::ADCChannelMapping*>((*result).second);
        }
    else
        {
            std::stringstream message;
            message << "No entry could be found for channel/readgroup: " << hardwareIDpair.first << "/" << hardwareIDpair.second << ".";
            throw tpcconddata::CondDataNotAvailableException(message.str());
        }
}



tpcconddata::ADCChannelMapping* ChannelMappingListener::getADCChannelMappingFromGearID(Int_Pair_t gearIDpair)
{
    ChannelMappingMap_t::iterator result = _gearChannelMappingMap.find(gearIDpair);
    if (result != _gearChannelMappingMap.end())
        {
            return dynamic_cast<tpcconddata::ADCChannelMapping*>((*result).second);
        }
    else
        {
            std::stringstream message;
            message << "No entry could be found for pad/module: " << gearIDpair.first << "/" << gearIDpair.second << ".";
            throw tpcconddata::CondDataNotAvailableException(message.str());
        }
}



void ChannelMappingListener::registeredWithHandler(lccd::IConditionsHandler* condHandler)
{
    _condHandler = condHandler;
    m_out(MESSAGE) << "ChannelMappingListener::registeredWithHandler(): registered: " << condHandler->name() << std::endl;
    m_out(DEBUG) << "ChannelMappingListener::registeredWithHandler(): register default collection:" << std::endl;
    // try to get the default collection
    try
        {
            LCCollection* col = condHandler->defaultCollection();

            if (! col)           // it will be null if none has so far been registered. So let's register ours
                {
                    condHandler->registerDefaultCollection(_thisEmptyCollection);
                    m_out(DEBUG) << "ChannelMappingListener::registeredWithHandler(): default collection registered; address: " << _thisEmptyCollection << std::endl;
                    // and put in the map for this handler
                    _handlerDefaultCollectionMap[_thisEmptyCollection] = condHandler;
                }
            //then the default handler was already registered, that's odd ... ;)
            else if (col == _thisEmptyCollection)
                {
                    m_out(DEBUG) << "ChannelMappingListener::registeredWithHandler(): default collection is already set to thisEmptyCollection:" << std::endl;
                }
            else                 // somebody has got there before us, let's see if we like the default
                {
                    // if it has the same CollectionName, we can use it, no problem ...
                    if (col->getParameters().getStringVal("CollectionName") == "EmptyCollectionChannelMappingListener")
                        {
                            m_out(DEBUG) << "ChannelMappingListener::registeredWithHandler(): I like your default ;)" << std::endl;
                            _handlerDefaultCollectionMap[col] = condHandler;
                        }
                    // if not, an exception is thrown because things should be handled in a uniform standardized way
                    else
                        {
                            m_out(ERROR) << "ChannelMappingListener::registeredWithHandler(): a non-standard empty default collection has already been registered with the conditions handler" << std::endl;
                            throw lccd::InconsistencyException("ChannelMappingListener::registeredWithHandler() found a default collection that differs from the standard one has already been registered with the handler");
                        }
                }
        }
    // catch MemberNotImplemented exception, which happens, if a simple file is used as input
    catch (lccd::MemberNotImplementedException)
        {
            m_out(MESSAGE) << "ChannelMappingListener::registeredWithHandler: Could not register default collection. Seems simple file input is used for ChannelMapping conditions." << std::endl;
        }
}



void ChannelMappingListener::deRegisteredWithHandler(lccd::IConditionsHandler* condHandler)
{
    m_out(WARNING) << "ChannelMappingListener::deRegisteredWithHandler() deregistered with:" << condHandler->name() << std::endl;
    _condHandler = NULL;
}



lcio::long64 ChannelMappingListener::collectionValidSince()
{
    return _timeStampSince;
}



lcio::long64 ChannelMappingListener::collectionValidTill()
{
    return _timeStampTill;
}



void ChannelMappingListener::useLastValidIfNoDataAvailable(bool useLastValid)
{
    _useLastValidIfNoDataAvailable = useLastValid;
}



bool ChannelMappingListener::noConditionsDataAvailable()
{
    return _noDataAvailable;
}


}
