#include "ChannelMappingProcessor.h"

// Marlin
#include <marlin/ConditionsProcessor.h>

//MarliNTPC
#include "ChannelMappingListener.h"

// LCIO
#include <EVENT/LCCollection.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCFlagImpl.h>
#include <UTIL/LCFixedObject.h>

#include "LCObjectCopier.h"

// tpcconditions
#include "TPCCondDataExeptions.h"
#include "ADCChannelMapping.h"

// C++
#include <iostream>

using namespace lcio ;
using namespace marlin ;
using namespace tpcconddata;
using namespace std;

namespace marlintpc
{

ChannelMappingProcessor aChannelMappingProcessor ;

ChannelMappingProcessor::ChannelMappingProcessor() : Processor("ChannelMappingProcessor")
{

    /* description of the processor which will be displayed in the
     * steering file autometicly generated by
     * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
     */
    _description = "ChannelMappingProcessor to translate the channel mapping from hardware channel and readout group to gear pad ID and module ID. Mappings provided via LCCD";

    /* register steering parameters: name, description, class-variable, default value
     * the type will by definded by the type of the default value
     * string, double, float and int are possible
     * plaese note to get a float you must cast in C++: (float) 1.3
     */
    registerInputCollection(LCIO::TRACKERPULSE,
                            "InputCollectionName" ,
                            "Name of the input collection"  ,
                            _inputColName ,
                            std::string("TPCPulses")
                           );

    registerOptionalParameter("OutputCollectionName",
                              "Optional: Name of output collection",
                              _outputColName ,
                              std::string("TPCPulsesMapped"));

    registerProcessorParameter("WriteOutputToStorage",
                               "if true the output collection is written",
                               _outputIsPersistent,
                               true);

    registerProcessorParameter("ChannelMappingCollectionName" ,
                               "Name of the collection containing the ChannelMappings",
                               _channelMappingColName,
                               tpcconddata::ADCChannelMapping::getDefaultColName()
                              );

    registerOptionalParameter("MissingConditionsDataOverrideTimespan",
                              "The timepan in seconds for which the last valid collection will be used if no conditions data are available via LCCD.",
                              _missingConditionsDataOverrideTimespan,
                              (int)0);

    registerProcessorParameter("RemoveChannelsWithoutMapping",
                               "If true, found signals without mapping in the conditions data will be removed from collection, else processing will stop in such a case.",
                               _removeChannelsWithoutMapping,
                               false);
}

void ChannelMappingProcessor::init()
{
    printParameters() ;
    // check if a new output collection will be used
    if (!(parameterSet("OutputCollectionName")))
        {
            m_out(WARNING) << _processorName << " : No output collection name given. During processing, readout electronics mapping values in the original collection will be overwritten with gear mapping values!" << std::endl;
        }

    // take care of the LCCD conditions listener
    _ChannelMappingListener = new ChannelMappingListener(_channelMappingColName);
    if (parameterSet("MissingConditionsDataOverrideTimespan"))
        {
            _missingConditionsDataOverrideLCIOTime = 1000000000 * (lcio::long64)(_missingConditionsDataOverrideTimespan);
            m_out(WARNING) << _processorName << " : Processing will go on using the last valid conditions collection even if there are no data available in the database for " << _missingConditionsDataOverrideTimespan << " seconds!" << std::endl;
        }
}



void ChannelMappingProcessor::processRunHeader(LCRunHeader* run)
{
    run->parameters().setValue(_processorName + "_revision", "$Rev: 2453 $");
    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }
}

void ChannelMappingProcessor::processEvent(LCEvent * evt)
{
    LCCollectionVec *inputCol = NULL;
    // get the input collection from the event
    try
        {
            inputCol = dynamic_cast<LCCollectionVec*>(evt->getCollection(_inputColName)) ;
            if (inputCol->size() == 0)
                {
                    throw DataNotAvailableException(std::string("Empty pulses collection ") + _inputColName);
                }
        }
    catch (DataNotAvailableException &e)
        {
            m_out(WARNING) << _processorName << " : Empty event " << evt->getEventNumber() << endl;
            //      throw marlin::SkipEventException( this);
            return;
        }

    LCCollectionVec* outputCol = NULL;
    // create the output collection
    if (parameterSet("OutputCollectionName"))
        {
            outputCol = new LCCollectionVec(LCIO::TRACKERPULSE);
            // set the transient flag as defined by the processor parameter
            for (LCCollectionVec::iterator it = inputCol->begin(); it < inputCol->end() ; ++it)
                {
                    outputCol->addElement(LCObjectCopier::copy(dynamic_cast<TrackerPulseImpl*>(*it)));
                }
            outputCol->setTransient(!_outputIsPersistent);   // transience is the opposite of persistency
        }
    else
        {
            // test if the input collection is writeable
            try
                {
                    dynamic_cast<TrackerPulseImpl*>(inputCol->getElementAt(0))->setCellID0(
                        dynamic_cast<TrackerPulse*>(inputCol->getElementAt(0))->getCellID0());
                }
            catch (ReadOnlyException &e)
                {
                    m_out(ERROR) <<  _processorName << " :" << std::endl;
                    m_out(ERROR) << "  Error: input collection is not writeable." << std::endl;
                    m_out(ERROR) << "  You have to specify an OutputCollectionName." << std::endl;
                    m_out(ERROR) << "  Remember to adapt the input collection name of the next processor!" << std::endl;
                    throw lcio::ReadOnlyException(_inputColName + " not writeable");
                }
            outputCol = inputCol;
        }
    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1) ;
    outputCol->setFlag(trkFlag.getFlag());

    //check if the conditions data are proper or if we want to use the last valid collection
    if (parameterSet("MissingConditionsDataOverrideTimespan"))
        {
            lcio::LCTime eventTime(evt->getTimeStamp());
            if (_ChannelMappingListener->noConditionsDataAvailable())
                {
                    m_out(DEBUG4) << _processorName << "::processEvent:Timecheck: eventTime.timeStamp()=" << eventTime.timeStamp() << " _ChannelMappingListener->collectionValidSince())=" << _ChannelMappingListener->collectionValidSince() << " _missingConditionsDataOverrideLCIOTime=" << _missingConditionsDataOverrideLCIOTime << std::endl;
                    m_out(DEBUG4) << _processorName << "::processEvent:Timecheck: ( (eventTime.timeStamp())-(_ChannelMappingListener->collectionValidSince()))=" << ((eventTime.timeStamp()) - (_ChannelMappingListener->collectionValidSince())) << std::endl;

                    // check if we are in a timespan where we want to use the last valid collection
                    if (((eventTime.timeStamp()) - (_ChannelMappingListener->collectionValidSince())) > _missingConditionsDataOverrideLCIOTime)
                        {
                            // we are outside the timespan that we want to allow, so we bail out by throwing an exception
                            std::stringstream out;
                            out << _processorName << "::processEvent: Tried to use last valid collection for ChannelMappings but found none in allowed timespan. Exiting processing";
                            throw tpcconddata::CondDataNotAvailableException(out.str());
                        }
                    else
                        {
                            // if we are inside the allowed timespan we just go on
                            m_out(DEBUG) << _processorName << "::processEvent: no current ChannelMapping conditions data available. Using last valid collection at time " << eventTime.timeStamp() << " (" << eventTime.year() << " " << eventTime.month() << " " << eventTime.day() << " " << eventTime.hour() << " " << eventTime.min() << " " << eventTime.sec() << " " << eventTime.ns() << ")" << std::endl;
                        }
                }
        }

    // vector to store "bad" channels that couldn't be mapped
    std::vector<int> badChannels ;

    // get the number of elements in the input collection
    int nData = outputCol->getNumberOfElements();
    // loop over the number of elements of the input collection
    for (int iData = 0; iData < nData ; iData++)
        {
            /* getElement return only a LCObject
                * you have to cast it the get the right type
                * please use dynamic_cast which will return 0 if the object can not be casted
                */
            TrackerPulseImpl* pulse = dynamic_cast<TrackerPulseImpl*>(outputCol->getElementAt(iData)) ;

            try
                {
                    // try to change the cell id's from electronics to gear
                    std::pair<int, int> gearIDpair = _ChannelMappingListener->getGearMapping(std::make_pair(pulse->getCellID0(), pulse->getCellID1()));
                    pulse->setCellID0(gearIDpair.first);
                    pulse->setCellID1(gearIDpair.second);

                }
            catch (tpcconddata::CondDataNotAvailableException& e)
                {
                    if (_removeChannelsWithoutMapping)
                        {
                            // if the channel was not found, the data will be removed from the collection
                            // and a later warning be given at the end of the reconstruction
                            // if verbostiy is set to DEBUG, already here a message will be displayed
                            m_out(DEBUG) << _processorName << " :" <<
                            "************************************************************************"  << std::endl
                            << "*   will remove channel [" << pulse->getCellID0() << "," << pulse->getCellID1() << "] "
                            << "*   from collection !  " << std::endl
                            << "************************************************************************" << std::endl ;
                            // store channel data for later removal from the collection
                            badChannels.push_back(iData) ;
                            _cMap[ std::make_pair(pulse->getCellID0(), pulse->getCellID1())] ++ ;
                        }
                    else
                        {
                            throw e;
                        }
                }
        }

    if (_removeChannelsWithoutMapping)
        {
            // remove "bad" elements now - need to start from end of vector !!
            for (std::vector<int>::reverse_iterator it = badChannels.rbegin() ; it != badChannels.rend() ; ++it)
                {
                    outputCol->removeElementAt(*it) ;
                }
        }
    // if the pulses collection has been created by cloning
    // we have to add it to the event
    if (parameterSet("OutputCollectionName"))
        {
            evt->addCollection(outputCol, _outputColName);
        }

    // save the processor parameters in the first collection stored
    if (isFirstCollection)
        {
            outputCol->parameters().setValue("InputCollectionName", _inputColName);
            outputCol->parameters().setValue("SetOutputTransient", (!_outputIsPersistent));
            outputCol->parameters().setValue("ChannelMappingCollectionName", _channelMappingColName);
            isFirstCollection = false;
        }
}



void ChannelMappingProcessor::check(LCEvent * evt)
{
    // nothing here currently - could be extended to fill checkplots
}



void ChannelMappingProcessor::end()
{
    if (_removeChannelsWithoutMapping)
        {
            // output if there have been "bad" channels (which could not be mapped)
            if (_cMap.size())
                {

                    m_out(WARNING3) << _processorName << " :"
                    << "******************************************************************************" << std::endl
                    << "*  The following hardware channels have been dropped from the output          " << std::endl
                    << "*  collection, due to unknown channel mapping (possibly unconnected channels) " << std::endl
                    << "*  cellID0 \t cellID1 \t count (#events)                                      " << std::endl
                    << "*                                                                             " << std::endl;

                    for (MAP::iterator it = _cMap.begin() ; it != _cMap.end() ; ++it)
                        {

                            streamlog_out(WARNING3) << "*   "  <<  it->first.first << "\t" << it->first.second << "\t" << it->second  << std::endl ;
                        }

                    m_out(WARNING3) << "*                                                                              " << std::endl
                    << "*******************************************************************************" << std::endl ;
                }
        }
}

}
