#ifndef HIT_CANDIDATE_H
#define HIT_CANDIDATE_H

//LCIO
#include <lcio.h>
#include <EVENT/TrackerPulse.h>
#include <IMPL/TrackerHitImpl.h>

//GEAR
#include <gear/TPCParameters.h>

//stl
#include <vector>

namespace marlintpc
{
/** A helper class for hit parameter calculation. It contains a vector of pulses and
 *  functions to calculate coordinates and errors. This functionality is needed at
 *  several places, so it has been decoupled from the processors to avoid code duplication.
 *
 *  \todo FIXME: Calculation of covariance matrix is missing, currently is unity.
 *  It should be the estimated error of the coordinates, not the varianve of the charge distribution.
 */
class HitCandidate
{
    public:
        /// The constructor needs a reference to TPCParameters and the dirft velocity
        /// Optionally the first pulse can be given.
        HitCandidate(const gear::TPCParameters & tpcParams, double vDrift,
                     EVENT::TrackerPulse * pulse = 0);

        /// Add a pulse to the hit.
        void addPulse(EVENT::TrackerPulse*);
        void addDeadPad();

        /// Manually turn on the boundary flag
        void setBoundaryFlag();


        /// Return the highest pulse in this hit.
        const EVENT::TrackerPulse* getMaximumPulse() const
        {
            return _maxPulse;
        };

        /// Get the overall charge in the hit
        double getCharge();

        /// Get the maximal pulse heigt in this hit
        double getMaxPulseHeight() const;

        /// Get a vector with the pulses contributing to this hits candidate.
        const std::vector<EVENT::TrackerPulse*>& getPulses() const
        {
            return _thePulses;
        };

        /// Return the number of pulses contributing to this hits candidate.
        /// This might not be the real width if there are multiple pulses per channel.
        unsigned int getWidth() const;

        /// Return whether the hit contains pulses on the boundary pads of a module.
        /// Currently this flag has to be set manually
        /// \todo FIXME: Calculation of boudary condition should be done automatically when adding a pulse
        bool   isAtBoundary() const
        {
            return _isAtBoundary;
        };

        /// Get the middle of the pad row (y or r)
        double getLocalFixedCoordinate();

        /// Get the coordinate along the pad row (x or phi)
        double getLocalFreeCoordinate();

        /// Get the z coordinate
        double getZCoordinate();

        /// Get the variance (estimated error) perpendicular to the pad row (y or r)
        double getLocalFixedVariance();

        /// Get the variance (estimated error) along the pad row (in x or phi)
        double getLocalFreeVariance();

        /// Get the skewness along the pad row
        double getLocalSkewness();

        /// Get the variance (estimated error) in z
        double getZVariance();

        /// Get the local coordinates
        /// (polar or cartesian, depending on the pad plane).
        lcio::DoubleVec getLocalCoordinates();

        /// Get the global coordinates
        /// (polar or cartesian, depending on the coordinate system).
        lcio::DoubleVec getGlobalCoordinates();

        /// Get the global coordinates in cartesian format
        /// (in lcio it is always cartesian).
        lcio::DoubleVec getGlobalCartesianCoordinates();

        /// Get the covariance matrix in global coordinates
        /// (polar or cartesian, depending on the coordinate system).
        lcio::FloatVec getGlobalCovarianceMatrix();

        /// Get the global covariance matrix in cartesian coordinates.
        /// This is what will be in the lcio hit
        lcio::FloatVec getGlobalCartecianCovarianceMatrix();

        /// Convenience function that allocates a new TrackerHitImpl in memory,
        /// sets all parameters and adds the pulses in this HitCandidate
        IMPL::TrackerHitImpl * createTrackerHit();

    protected:
        std::vector<EVENT::TrackerPulse*> _thePulses;
        EVENT::TrackerPulse* _maxPulse;
        double _maxPulseCharge;
        int _numberOfDeadPads;
        bool _isAtBoundary;
        double _vDrift;
        double _zAnode;

        const gear::TPCParameters & _tpcParameters;
        const gear::TPCModule * _tpcModule;

        // the parameters calculated by calculateHitProperties

        lcio::DoubleVec _localCoordinates;
        lcio::DoubleVec _globalCoordinates;
        lcio::FloatVec _globalCovariance;

        double _charge;
        double _localFixedVariance;
        double _localFreeVariance;
        double _localSkewness; ///<  The skewness in local coordinates
        double _zVariance;

        /// (re)calculate the properties of this hit candidate
        void calculateHitProperties();

        bool _hitParametersAreValid; /// A flag to store whether the hit parameters have been
        /// recalculated after the last pulse has been added
};

}

#endif //HIT_CANDIDATE_H
