//C++
#include <iostream>
#include <stdexcept>
#include <map>
#include <cassert>
#include <cmath>
#include <complex>
#include <typeinfo>

//LCIO
#include <EVENT/TrackerPulse.h>
#include <IMPL/TrackerHitImpl.h>
#include <EVENT/TrackerData.h>

#include <marlin/Global.h>

#include "TopoFinder.h"
#include <streamlog/streamlog.h>
#include <gear/PadRowLayout2D.h>
#include <gear/TPCModule.h>

using namespace std;
using namespace lcio;
using namespace gear;

namespace marlintpc
{

TopoFinder::TopoFinder() : _maxEmptyPads(1), _minHitSize(1), _maxTimeSpread(100.) //ns
        , _vDrift(0.), _zAnode(0), _negativeZDirection(0),
        _pulsesByIndex(0),  _recursionCounter(0)
{} // just an empty constructor...


EVENT::TrackerHit* TopoFinder::findHit(std::multimap<gear::GlobalPadIndex, TrackerPulse*>::iterator currentPulseIter)
{
    // Check that the multimap has been initialized correctly
    // As it is a programming bug if this is not the case we don't want to waste
    // calculation time if not in debug mode since this function might be called often:
    // use assert!
    assert(_pulsesByIndex);

    // instantiate a new hitCandidate and add the first pulse to it
    TrackerHitImpl* hitCandidate = new TrackerHitImpl();
    // this would add the pulse twice
    //    hitCandidate->rawHits().push_back( (*currentPulse).second );

    const gear::TPCModule &module =
        marlin::Global::GEAR->getTPCParameters().getModule(currentPulseIter->first.getModuleID());

    const gear::PadRowLayout2D &localPadPlane = module.getLocalPadLayout();

    // store the pulse pointer and the pad index from the iterator.
    // afterwards remove the iterator from the map to avoid refinding it in the recursive search
    gear::GlobalPadIndex currentPadIndex = currentPulseIter->first;
    TrackerPulse * currentPulse = currentPulseIter->second;

    _pulsesByIndex->erase(currentPulseIter);

    std::list<TrackerPulse *> pulsesInThisHit = findNeighbouringPulses(currentPadIndex, currentPulse, module);

    for (std::list<TrackerPulse *>::iterator pulseIter = pulsesInThisHit.begin() ;
            pulseIter != pulsesInThisHit.end() ; ++pulseIter)
        {
            hitCandidate->rawHits().push_back(*pulseIter);
        }

    // check whether the cluster is large enough
    if (hitCandidate->rawHits().size() >= static_cast<unsigned int>(_minHitSize))
        {
            // hitCandidate is valid

            // calculate the hit from the cluster
            // the sums needed for centre of gravity:
            double sum_xphi_q = 0; // sum(x*q) or sum(phi*q)
            double sumzq = 0;    // sum(z*q)
            double sumq = 0;  // sum(q)
            double xphi = 0;   // coordinate along pad row
            double yr = 0;   // coordinate orthogonal to pad row

            // loop over all pulses in cluster and sum up
            for (EVENT::LCObjectVec::iterator pulseIter = hitCandidate->rawHits().begin();
                    pulseIter < hitCandidate->rawHits().end(); pulseIter++)
                {
                    // Raw hits is a vector of LCObjects.
                    // We have to cast the iterator to TrackerPulse to access the member functions.
                    TrackerPulse* pulse = dynamic_cast<TrackerPulse*>(*pulseIter);

                    // get pad coordinates of current pulse from GEAR layout
                    Vector2D padCenterLocal = localPadPlane.getPadCenter(pulse->getCellID0());

                    // get x / phi coordinate from first pad
                    switch (localPadPlane.getCoordinateType())
                        {
                        case PadRowLayout2D::CARTESIAN:
                            xphi = padCenterLocal[0]; // x
                            break;
                        case PadRowLayout2D::POLAR:
                            xphi = padCenterLocal[1]; // phi
                            break;
                        default:
                            throw gear::Exception("TopoFinder: unkown coordinate type");
                        };

                    // take y / r coordinate from first pad, it's the same for all pads
                    if (pulseIter == hitCandidate->rawHits().begin())
                        {
                            switch (localPadPlane.getCoordinateType())
                                {
                                case PadRowLayout2D::CARTESIAN:
                                    yr = padCenterLocal[1]; // y
                                    break;
                                case PadRowLayout2D::POLAR:
                                    yr = padCenterLocal[0]; // r
                                    break;
                                default:
                                    throw gear::Exception("TopoFinder: unkown coordinate type");
                                };
                        }

                    // summ correctly if handling pads from the first and thr fourth quadrant
                    // in polar coordinates (wrap around from 2 pi to 0)
                    if ((pulseIter != hitCandidate->rawHits().begin())  // for the first pad there is nothing to compare with
                            && (localPadPlane.getCoordinateType() == PadRowLayout2D::POLAR))  // problem only occurs in polar coordinates
                        {
                            // there can be thwo cases
                            //1.
                            if ((sum_xphi_q / sumq < M_PI_2) // pulse up to now is in first quadrant (or in fourth, already mapped to <0)
                                    && (xphi > 1.5*M_PI))  // current pulse in fourth quadrant
                                xphi -= 2 * M_PI; // map to range [-pi:pi] so averaging works

                            // 2.
                            // no else needed; statements are mutually excluse because due to first condition
                            // even if xphi was changed in the first statement
                            if ((sum_xphi_q / sumq > 1.5* M_PI) // pulse up to now is in fourth quadrant (or in first, already mapped to >2*pi)
                                    && (xphi < M_PI_2))  // current pulse in first quadrant
                                xphi += 2 * M_PI; // map to range [pi:3*pi] so averaging works
                        }

                    // add to the sums
                    sum_xphi_q += xphi * pulse->getCharge();  // x * q or phi * q
                    sumzq += pulse->getTime() * _vDrift / 1000 // t * vDrift = z // 1000 ns/us
                             * pulse->getCharge() ;   // * q
                    sumq += pulse->getCharge() ;    // q

                }// for (pulse)

            // global 2D position, cartesian ot polar
            gear::Vector2D positionGlobal2D;

            switch (localPadPlane.getCoordinateType())
                {
                case PadRowLayout2D::CARTESIAN:
                    // in cartesian coordinates xphi is x, yr is y
                    positionGlobal2D = module.localToGlobal(sum_xphi_q / sumq , yr);
                    break;
                case PadRowLayout2D::POLAR:
                    positionGlobal2D = module.localToGlobal(yr , sum_xphi_q / sumq);
                    break;
                };

            // the positon in global cartesian for lcio
            double positionGlobal[3];
            switch (module.getTPCCoordinateType())
                {
                case PadRowLayout2D::CARTESIAN:
                    // in cartesian coordinates xphi is x, yr is y
                    positionGlobal[0] = positionGlobal2D[0];
                    positionGlobal[1] = positionGlobal2D[1];
                    break;
                case PadRowLayout2D::POLAR:
                    // convert polar to cartesian
                    positionGlobal[0] = positionGlobal2D[0] * cos(positionGlobal2D[1]);
                    positionGlobal[1] = positionGlobal2D[0] * sin(positionGlobal2D[1]);
                    break;
                };

            // sumzq contains the weighted drift distance
            // calculate z coordinate wrt. origin, which is at the cathode
            positionGlobal[2] = _zAnode - sumzq / sumq;

            // the second half TPC look in the negative z direction
            // (rotated by 180 degress around y axis wrt. first endplate)
            // multiply x and z with -1 in this case
            if (_negativeZDirection)
                {
                    positionGlobal[0] *= -1.;
                    positionGlobal[2] *= -1.;
                }

            // set position and dEdx of the hit
            hitCandidate->setPosition(positionGlobal);
            hitCandidate->setdEdx(sumq);

            // return the valid hit
            return hitCandidate;
        } // ifMinClusterSize
    else
        {
            streamlog_out(DEBUG3) << "rejecting hit, too small" << endl;
            delete hitCandidate;
            return 0;
        }

} // findHit


std::list<TrackerPulse *> TopoFinder::findNeighbouringPulses(
    gear::GlobalPadIndex currentPadIndex ,
    EVENT::TrackerPulse * currentPulse,
    const gear::TPCModule & module)
{
    std::list<TrackerPulse *> pulsesReturnList;
    pulsesReturnList.push_back(currentPulse);

    std::list<std::multimap<gear::GlobalPadIndex, EVENT::TrackerPulse*>::iterator > pulsesFound;

    // look for neighbouring pads to the left and to the right of current pad
    // This loop just implements looking in both directions :
    // searchLeftNeighbours = 0 : search to the right
    // searchLeftNeighbours = 1 : search to the left
    for (unsigned int searchLeftNeighbours = 0 ; searchLeftNeighbours < 2 ; searchLeftNeighbours++)
        {
            unsigned int nEmptyPads = 0;

            // start search at pad index of currentPulse
            gear::GlobalPadIndex lastCheckedPadIndex = currentPadIndex;

            // loop neighbouring pads to search for pulses
            // until we are _maxEmptyPads away from the last pad contributing to this hit
            while (nEmptyPads <= static_cast<unsigned int>(_maxEmptyPads))
                {
                    // search for next matching pulse
                    gear::GlobalPadIndex neighbouringPadIndex(-1, -1);

                    // determine index of pad neighbouring the one which was already checked
                    try
                        {
                            if (searchLeftNeighbours)
                                neighbouringPadIndex = gear::GlobalPadIndex(
                                                           module.getLeftNeighbour(lastCheckedPadIndex.getPadIndex()) ,
                                                           module.getModuleID());
                            else
                                neighbouringPadIndex = gear::GlobalPadIndex(
                                                           module.getRightNeighbour(lastCheckedPadIndex.getPadIndex()) ,
                                                           module.getModuleID());
                        }
                    catch (std::exception &e)
                        {
                            streamlog_out(DEBUG0) << e.what() << std::endl; // can be printed for debug

                            // no neighbouring pad in the pad plane, stop searching
                            break;
                        }

                    // remember that this pad has been checked
                    lastCheckedPadIndex = neighbouringPadIndex;

                    // Increase the number of empty pads
                    // This is done before the actual search is performed because it avoids
                    // introducing an additional variable to remember whether the search was successful.
                    // If a matching pulse was found this counter is resert to 0 anyway.
                    // If not the counter had been increased.
                    nEmptyPads++;

                    // check if the pad index has been marked as bad ( dead or noisy)
                    // tread is as an empty pad and continue
                    // !!!FIXME!!! To be implemented
                    // if (neighbouringPadIndex is invalid)
                    //    continue;

                    // loop all pulses on the neighbouring pad:

                    // The pulsesByIndex map provides the first and the last iterator to a matching index as a pair.
                    // So this is a pair of iterators to the entries of the multimap, which contains an int and an
                    // pointer to a TrackerPulse.
                    // The pair variable is called 'neighbours'
                    std::pair < std::multimap<gear::GlobalPadIndex, TrackerPulse*>::iterator,
                    std::multimap<gear::GlobalPadIndex, TrackerPulse*>::iterator > neighbours;

                    for (neighbours = _pulsesByIndex->equal_range(neighbouringPadIndex);
                            neighbours.first != neighbours.second; neighbours.first++)
                        {
                            // neighbours.first is the iterator to the map entry we want to check
                            // the second member in the map entry is the pointer to the TrackerPulse
                            TrackerPulse* pulseCandidate = neighbours.first->second;

                            // check that the pulse is in the correct time window


                            // if (fabs( pulseCandidate->getTime() - (*currentPulse).second->getTime() )
                            //     < _maxTimeSpread )

                            // The pulses have to overlap (assuming they have been split correclty)
                            // no additional cut required
                            // FIXME: What if two pulses overlap?
                            // - use the one with the smaller distance in time?
                            // - use the one with more overlapping bins?
                            // - take both!?

                            // Sufficient for overlap:
                            // start time or end time of the second pulse is between start and end time of the first
                            float start1 = currentPulse->getTrackerData()->getTime();
                            float end1 = start1 + currentPulse->getTrackerData()->getChargeValues().size();
                            float start2 = pulseCandidate->getTrackerData()->getTime();
                            float end2 = start2 + pulseCandidate->getTrackerData()->getChargeValues().size();

                            if (!((end2 < start1) || (start2 > end1)))
                                {
                                    // pulse belongs to the hit
                                    // add it to the list of found iterators
                                    pulsesFound.push_back(neighbours.first);

                                    // reset the counter for invalid pads (empty, noisy or dead)
                                    nEmptyPads = 0;
                                    streamlog_out(DEBUG2) << "added new pad: padIndex " << neighbouringPadIndex.getPadIndex()
                                    << " = padNr " << (neighbouringPadIndex.getPadIndex() & 0xffff)  << std::endl;
                                }
                            else
                                {
                                    streamlog_out(DEBUG2) << "pulse rejected on pad "
                                    << neighbouringPadIndex.getPadIndex()
                                    << ": pulse1(" << start1 << "," << end1
                                    << "), pulse2(" << start2 << "," << end2
                                    << "), distance"
                                    << pulseCandidate->getTime() - currentPulse->getTime()
                                    << std::endl;
                                }

                            //      total_charge += pulse_candidate->getCharge();
                            //      weighted_charge += pulse_candidate->getCharge() * pulse_candidate->getTime();

                        } // for (neighbours)

                }// while ( nInvalidPads <= _maxEmptyPads )

        }// for (searchLeftNeighbours)

    // Search is complete.

    // delete all iterators and store the pulses and indices in a list for recursion
    std::list< std::pair<gear::GlobalPadIndex, EVENT::TrackerPulse*> > recursionList;
    for (std::list< std::multimap<gear::GlobalPadIndex, EVENT::TrackerPulse*>::iterator >::iterator pulseIterIter
            = pulsesFound.begin(); pulseIterIter != pulsesFound.end() ; ++pulseIterIter)
        {
            // delete the iterator from the map before starting the recursive search to avoid double finding
            recursionList.push_back(*(*pulseIterIter));
            _pulsesByIndex->erase(*pulseIterIter);
        }

    // Now perform the recursive search and store all pulses in the iterators to the return list.
    for (std::list< std::pair<gear::GlobalPadIndex, EVENT::TrackerPulse*> >::iterator pulseIter
            = recursionList.begin(); pulseIter != recursionList.end() ; ++pulseIter)
        {
            // findNeighbouringPulses returns the pulse to the iter in the return list
            std::list<TrackerPulse *> neighboursOfThisPulse = findNeighbouringPulses(pulseIter->first,
                                                                                     pulseIter->second,
                                                                                     module);

            // insert (splice) the return value into the return list (splice empties the souce list)
            pulsesReturnList.splice(pulsesReturnList.end(), neighboursOfThisPulse);
        }

    return pulsesReturnList;
}


EVENT::TrackerHitVec* TopoFinder::findHitsInNeighbouringRows(EVENT::TrackerHit* hit, int moduleID)
{
    streamlog_out(DEBUG4) << "recursion depth " << _recursionCounter++ << std::endl;

    // get the module and pad plane, only work on one module
    const TPCModule &module = marlin::Global::GEAR->getTPCParameters().getModule(moduleID);
    const PadRowLayout2D &localPadLayout = module.getLocalPadLayout();


    LCObjectVec pulseVec = hit->getRawHits();
    streamlog_out(DEBUG1) << "size of pulseVec is " << pulseVec.size() << endl;

    TrackerHitVec *neighbouringHits = 0;

    for (LCObjectVec::iterator it = pulseVec.begin(); it < pulseVec.end(); it++)
        {
            TrackerPulse *pulse = dynamic_cast<TrackerPulse*>(*it);
            int padIndex = pulse->getCellID0();

            // first calculate the left and right edge of the pad
            Vector2D positionLocal = localPadLayout.getPadCenter(padIndex);
            streamlog_out(DEBUG3) << "local pad coordinates are " << positionLocal[0]
            << " " << positionLocal[1] << endl;

            double leftPadEdge = 0, rightPadEdge = 0;
            double padPitch = localPadLayout.getPadPitch(padIndex);

            switch (localPadLayout.getCoordinateType())
                {
                case PadRowLayout2D::CARTESIAN : // x is used in cartesian coordinates
                    leftPadEdge = positionLocal[0] - padPitch / 2;
                    rightPadEdge = positionLocal[0] + padPitch / 2;
                    break;
                case PadRowLayout2D::POLAR : // phi is used in polar coordinates, it's the second cordinate
                    leftPadEdge = positionLocal[1] - padPitch / 2;
                    rightPadEdge = positionLocal[1] + padPitch / 2;
                    break;
                default :
                    throw gear::NotImplementedException(std::string("Unknown pad layout!"));
                }

            float minADCIndex = pulse->getTrackerData()->getTime();
            float maxADCIndex = minADCIndex + pulse->getTrackerData()->getChargeValues().size();

            int currentRow = module.getRowNumber(padIndex);

            // now search the neighbouring rows, forward and backward
            for (int forward = 0; forward <= 1 ; forward++)
                {
                    for (int deltaRow = 1; deltaRow <= 1 + _maxSkipRows ; deltaRow++)
                        {
                            //calculate y/r coordinate of the investigated row
                            // take pad 0 in this row, all pads have same y/r
                            Vector2D positionLocal;
                            try
                                {
                                    positionLocal = localPadLayout.getPadCenter(
                                                        localPadLayout.getPadIndex(currentRow + (forward ? +1 : -1) * deltaRow, 0));
                                }
                            catch (std::out_of_range &)
                                {   // The requested row does not exist, stop searching
                                    break;
                                }
                            double y_r;
                            switch (localPadLayout.getCoordinateType())
                                {
                                case PadRowLayout2D::CARTESIAN : // y is used in cartesian coordinates
                                    y_r = positionLocal[1];
                                    break;
                                case PadRowLayout2D::POLAR : // r is used in polar coordinates, it's the second cordinate
                                    y_r = positionLocal[0]  ;
                                    break;
                                default :
                                    throw gear::NotImplementedException(std::string("Unknown pad layout!"));
                                }

                            // get the pad indices nearest to
                            int leftmostIndex, rightmostIndex;
                            switch (localPadLayout.getCoordinateType())
                                {
                                case PadRowLayout2D::CARTESIAN : // x and y is used in cartesian coordinates
                                    leftmostIndex = localPadLayout.getNearestPad(leftPadEdge , y_r);
                                    rightmostIndex = localPadLayout.getNearestPad(rightPadEdge , y_r);
                                    break;
                                case PadRowLayout2D::POLAR : // r and phi is used in polar coordinates (in this order)
                                    leftmostIndex = localPadLayout.getNearestPad(y_r , leftPadEdge);
                                    rightmostIndex = localPadLayout.getNearestPad(y_r , rightPadEdge);
                                    break;
                                default :
                                    throw gear::NotImplementedException(std::string("Unknown pad layout!"));
                                }



                            TrackerHitVec *hitsInRange = findHitsInRange(gear::GlobalPadIndex(leftmostIndex, moduleID),
                                                                         gear::GlobalPadIndex(rightmostIndex, moduleID),
                                                                         minADCIndex, maxADCIndex);
                            // drift time in ns, calculated from z in mm
                            if (hitsInRange)
                                {
                                    // recursively search for neighbouring hits
                                    for (TrackerHitVec::iterator hitIter = hitsInRange->begin(); hitIter < hitsInRange->end(); hitIter++)
                                        {
                                            streamlog_out(DEBUG2) << "starting recursion..." << endl;
                                            TrackerHitVec *recursivelyFoundHits = findHitsInNeighbouringRows(*hitIter, moduleID);
                                            addTrackerHitVecs(&neighbouringHits, &recursivelyFoundHits);
                                        }
                                }

                            addTrackerHitVecs(&neighbouringHits, &hitsInRange);

                        } //for deltaRow
                }// for forward

        }// for pulseIter

    if (neighbouringHits)
        {
            streamlog_out(DEBUG2) << "returning " << neighbouringHits->size() << " hits" << std::endl;
        }
    else
        {
            streamlog_out(DEBUG2) << "returning 0 hits" << std::endl;
        }

    _recursionCounter--;


    return neighbouringHits;
}

EVENT::TrackerHitVec* TopoFinder::findHitsInRange(gear::GlobalPadIndex index1,
                                                  gear::GlobalPadIndex index2,
                                                  float minADCIndex, float maxADCIndex)
{
    streamlog_out(DEBUG3) << "TopoFinder::findHitsInRange(" << index1.getPadIndex()
    << "," << index2.getPadIndex()
    << ", timeIndex " << minADCIndex << " - " << maxADCIndex << ")" << endl;
    TrackerHitVec *hitsInRange = new  TrackerHitVec;

    const TPCModule &module = marlin::Global::GEAR->getTPCParameters().getModule(index1.getModuleID());

    // in case of wrap around at phi = 0 for polar geometry there are two ranges:
    // [index1, lastIndexInRow] and [0, index2]
    // store them in a vextor
    std::vector< std::pair<gear::GlobalPadIndex, gear::GlobalPadIndex> > indexRanges;
    if (!(index2 < index1))
        {
            indexRanges.push_back(std::make_pair(index1, index2));
        }
    else
        {
            int row = module.getRowNumber(index1.getPadIndex());
            indexRanges.push_back(std::make_pair(index1, gear::GlobalPadIndex(*(module.getPadsInRow(row).rbegin()), index1.getModuleID())));
            //                                                                                              ^^^^^^ last pad in row
            indexRanges.push_back(std::make_pair(gear::GlobalPadIndex(*(module.getPadsInRow(row).begin()) , index1.getModuleID()) , index2));
            //                                                                                       ^^^^^ first pad in row
            streamlog_out(DEBUG3) << "modified ranges to [" << indexRanges[0].first.getPadIndex() << ","
            << indexRanges[0].second.getPadIndex() << "] and ["
            << indexRanges[1].first.getPadIndex() << ","
            << indexRanges[1].second.getPadIndex() << "]" << std::endl;
        }

    for (std::vector< std::pair<gear::GlobalPadIndex, gear::GlobalPadIndex> >::iterator indexRangeIter = indexRanges.begin() ;
            indexRangeIter < indexRanges.end(); ++indexRangeIter)
        {

            // The first pulse with index1
            std::multimap<gear::GlobalPadIndex, TrackerPulse*>::iterator firstPulse = _pulsesByIndex->lower_bound(indexRangeIter->first);
            // The pulse after the last pulse with index2
            std::multimap<gear::GlobalPadIndex, TrackerPulse*>::iterator lastPulse = _pulsesByIndex->upper_bound(indexRangeIter->second);;

            while (firstPulse != lastPulse)
                {
                    streamlog_out(DEBUG2) << "testin pulse with index " << firstPulse->first.getPadIndex()
                    << " , time " << firstPulse->second->getTrackerData()->getTime()
                    << endl;
                    float start = firstPulse->second->getTrackerData()->getTime();
                    float end = start + firstPulse->second->getTrackerData()->getChargeValues().size();

                    if (!((end < minADCIndex || (start > maxADCIndex))))
                        {
                            TrackerHit *hit = findHit(firstPulse);
                            if (hit)
                                hitsInRange->push_back(hit);

                            // get new values for the iterators, they might have been deleted by findHit
                            firstPulse = _pulsesByIndex->lower_bound(indexRangeIter->first);
                            lastPulse = _pulsesByIndex->upper_bound(indexRangeIter->second);
                        }
                    else
                        firstPulse++;
                }
        }

    if (hitsInRange->size() > 0)
        {
            streamlog_out(DEBUG3) << "Found " << hitsInRange->size() << " hits in range" << endl;
            return hitsInRange;
        }
    else
        {
            delete hitsInRange;
            return 0;
        }
}

void TopoFinder::addTrackerHitVecs(EVENT::TrackerHitVec **vectorToAddTo, EVENT::TrackerHitVec **vectorToBeAdded)
{
    if ((*vectorToBeAdded) == 0)
        return;

    if ((*vectorToAddTo) == 0)
        {
            *vectorToAddTo = *vectorToBeAdded;
            return;
        }

    for (TrackerHitVec::iterator it = (*vectorToBeAdded)->begin(); it < (*vectorToBeAdded)->end(); it++)
        {
            (*vectorToAddTo)->push_back(*it);
        }
    delete *vectorToBeAdded;
    *vectorToBeAdded = 0;

}

std::string TopoFinder::getRevision()
{
    return std::string("$Rev: 2275 $");
}

} // namespace marlintpc
