//LCIO
#include <EVENT/LCIO.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerPulseImpl.h>

//MarlinTPC
#include "PhotodotReconstructionProcessor.h"
#include "ElectronCloudDrifterProcessor.h"

//Marlin
#include <marlin/Global.h>

//C++
#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <cmath>
#include <iomanip>
#include <set>

//GEAR
#include <gearimpl/RectangularPadRowLayout.h>
#include <gearimpl/FixedPadSizeDiskLayout.h>
#include <gear/TPCParameters.h>

//Root
#include <TMath.h>

using namespace marlin;
using namespace gear;
using namespace lcio;

namespace marlintpc
{

// Create an instance of the processoer to register it with Marlin
PhotodotReconstructionProcessor aPhotodotReconstructionProcessor;

/* Call the constructor to initialize the private member values
 */
PhotodotReconstructionProcessor::PhotodotReconstructionProcessor() :
        Processor("PhotodotReconstructionProcessor"),
        /* Variables to be initialized:
         */
        _geometry_filename("photo_geometry.txt"),
        _input_pulses_collection_name("TPCPulses"),
        _compared_photodots_filename("compared_photodots.txt"),
        _pulsesByIndex(0),
        /* TPC geometry parameters:
         */
        _diffusion(1.),
        _omega_tau(10.),
        _drift_velocity(40.),
        _tpc_outer_radius(1500.),
        _tpc_inner_radius(305.),
        _outer_z(100.),
        _inner_z(10.),
        _sigma(0.034692)

{

    /* Description of the processor which will be displayed in the steering
     * file automatically generated by Marlin.
     */
    _description = "PhotodotReconstructionProcessor will compare the location of the centre of a deposited electron cloud with the position of the photodot that generated it. Photodot positions are adjusted for the presence of a known magnetic field.";

    registerInputCollection(LCIO::TRACKERPULSE,
                            "InputTrackerPulses",
                            "Input collection carrying a deposited photelectric pulse",
                            _input_pulses_collection_name,
                            std::string("TPCPulses"));

    registerOptionalParameter(
        "GeometryFilename",
        "The name of the file containing the positions of photodots",
        _geometry_filename,
        std::string("photo_geometry.txt"));

    registerOptionalParameter(
        "ComparedPhotodotsFilename",
        "The name of the output file containing the differences between observed and actual photodot positions",
        _compared_photodots_filename,
        std::string("compared_photodots.txt"));

    registerProcessorParameter(
        "TransverseDiffusion",
        "Transverse diffusion of photodots along the drift distance.",
        _diffusion,
        50.);

    registerProcessorParameter(
        "OmegaTau",
        "The gas parameterization for this section of the TPC",
        _omega_tau,
        10.);

    registerProcessorParameter(
        "DriftVelocity",
        "The parameterized drift velocity in this section of the TPC (mm/us)",
        _drift_velocity,
        40.);

    registerProcessorParameter(
        "TPCOuterRadius",
        "The outer radius of the TPC (mm)",
        _tpc_outer_radius,
        1500.);

    registerProcessorParameter(
        "TPCInnerRadius",
        "The inner radius of the TPC (mm)",
        _tpc_inner_radius,
        300.);

    registerProcessorParameter(
        "OuterZ",
        "The absolute value of the z coordinate of the outermost position of the TPC (mm)",
        _outer_z,
        100.);

    registerProcessorParameter(
        "InnerZ",
        "The absolute value of the z coordinate of the innermost position of the TPC (mm)",
        _inner_z,
        10.);

    //Pad layout information is initialized to disallowed values to cover errors
    _padLayoutType = -1;
    _padLayout = NULL;

} // /Constructor

/* Destructor
 */
PhotodotReconstructionProcessor::~PhotodotReconstructionProcessor() {}

void PhotodotReconstructionProcessor::init()
{
    streamlog_out(DEBUG) << "PhotodotReconstructionProcessor::init\n";

    // Usually a good idea to
    printParameters();

    _diffusion *= _sigma;

    // Get the pad layout from gear
    try
        {

            _padLayoutType = Global::GEAR->getTPCParameters().getPadLayout().getPadLayoutType();

            switch (_padLayoutType)
                {

                case PadRowLayout2D::CARTESIAN:
                    _padLayout = &dynamic_cast< const RectangularPadRowLayout & >
                                 (Global::GEAR->getTPCParameters().getPadLayout());
                    break;

                case PadRowLayout2D::POLAR:
                    _padLayout = &dynamic_cast< const FixedPadSizeDiskLayout & >
                                 (Global::GEAR->getTPCParameters().getPadLayout());
                    break;

                default:
                    throw gear::Exception("Unknown pad layout type");

                }; // /switch setting pad layout

        } // /try to set pad layout

    catch (std::exception & e)
        {

            streamlog_out(ERROR) << e.what() << std::endl;
            throw e;

        } // /catch setting pad layout

} // /init

void PhotodotReconstructionProcessor::processRunHeader(LCRunHeader * run)
{
    streamlog_out(DEBUG) << "PhotodotReconstructionProcessor::processRunHeader()\n";

    /* Get the photodot geometry
     */
    std::ifstream input_geometry;
    input_geometry.open(_geometry_filename.c_str());

    if (input_geometry.is_open())
        {

            input_geometry >> _geometry_type;

            // Get the geometry type (CARTESIAN or POLAR)
            streamlog_out(DEBUG) << "Using geometry type " << _geometry_type
            << std::endl;

            /* Check the type of photogeometry information
             */
            if (_geometry_type == "CARTESIAN")
                {

                    //Get the type of the emitting surface (photodots or photolines)
                    //This line exists to separate the photodot and photoline information.
                    //The string is not used.
                    std::string emittorType;
                    input_geometry.clear();
                    input_geometry >> emittorType;

                    //Get the actual photodot locations
                    _photodots.clear();

                    double x, y, diameter;
                    while (input_geometry >> x >> y >> diameter)
                        {

                            _photodots.push_back(Photodot(x, y, diameter));

                            streamlog_out(DEBUG) << "Added a photodot at " << x << ", " << y
                            << std::endl;
                        } // /while adding photodots

                    // There are no more photodots when input fails:
                    // the next piece of text describes the photolines.
                    input_geometry.close();

                } // /if Cartesian geometry

            else if (_geometry_type == "POLAR")
                {

                    // Separator between photodot and photoline information
                    // The string is not used.
                    std::string emittorType;
                    input_geometry.clear();
                    input_geometry >> emittorType;

                    // Get the photodot locations
                    _photodots.clear();

                    double r, phi, diameter;
                    while (input_geometry >> r >> phi >> diameter)
                        {

                            _photodots.push_back(Photodot(r, phi, diameter, 7));

                            streamlog_out(DEBUG) << "Added a photodot at " << r << ", " << phi
                            << std::endl;

                        } // /while adding photodots

                    // The next piece of text describes the photolines
                    input_geometry.close();

                } // if Polar photogeometry
            else

                streamlog_out(ERROR) << "Unrecognised photogeometry type."
                << std::endl;

        } // /if geometry file is open
    else

        streamlog_out(ERROR) << "Could not open " << _geometry_filename
        << std::endl;

    /* Create the output file
     */
    std::ofstream output_geometry;
    output_geometry.open(_compared_photodots_filename.c_str());

    if (output_geometry.is_open())
        {

            streamlog_out(DEBUG) << "Outputting file headers\n";

            output_geometry << _geometry_type << std::endl;

            output_geometry << " Photodot position  "
            << "  Expected image    "
            << "  Image position\n";

            for (int i = 0; i < 3; i++)
                {

                    if (_geometry_type == "CARTESIAN")
                        output_geometry << "(       x,       y) ";
                    else if (_geometry_type == "POLAR")
                        output_geometry << "(       r,     phi) ";

                } // /for each piece of output data

            output_geometry << std::endl;
            output_geometry.close();

        } // /if output file is open

    else

        streamlog_out(ERROR) << "Could not open "
        << _compared_photodots_filename
        << std::endl;

} // /processRunHeader

void PhotodotReconstructionProcessor::processEvent(LCEvent * evt)
{
    streamlog_out(DEBUG) << "IdealPhotodotComparisonProcessor::processEvent\n";

    try
        {

            // Get the input collection
            LCCollection * input_pulses = evt->getCollection(_input_pulses_collection_name);

            // Initialize _pulsesByIndex
            _pulsesByIndex = new std::multimap< int, TrackerPulse * >;

            /* Copy each input pulse into the map
             */
            for (int iRawData = 0;
                    iRawData < input_pulses->getNumberOfElements();
                    iRawData++)
                {

                    // Retrieve a pulse
                    TrackerPulse * input_pulse = dynamic_cast< TrackerPulse * >
                                                 (input_pulses->getElementAt(iRawData));

                    if (input_pulse)   // only proceed if the pulse is charged
                        {

                            // Insert the pulse into the map
                            _pulsesByIndex->insert(std::pair< const int, TrackerPulse * >
                                                   (input_pulse->getCellID0(), input_pulse));

                        } // / if valid input pulse

                    else

                        throw(EVENT::Exception("Input collection "
                                               + _input_pulses_collection_name
                                               + "does not contain TrackerPulses."));

                } // /for each input pulse

        } // /try getting input pulses

    catch (DataNotAvailableException e)
        {

            streamlog_out(ERROR) << e.what() << std::endl;
            throw e;

        } // /catch getting input pulses

    // Determine which pulses are in a valid time range.
    findPhotodotPulses();

    streamlog_out(MESSAGE) << "Processing " << _photodots.size()
    << " pulses." << std::endl;

    /* Create an output file for photodot information
     */
    streamlog_out(MESSAGE) << "Creating output file\n";

    double rError, phiError;
    rError = phiError = 0;

    /* Open the output file for appending
     */

    std::ofstream output_geometry(_compared_photodots_filename.c_str(),
                                  std::ios::app);

    /*
     * Process the photodots
     */
    streamlog_out(MESSAGE) << "Processing photodots" << std::endl;
    for (unsigned int i = 0; i < _photodots.size(); i++)
        {

            /* Find the nearest hit to the photodot.
             */
            int closestHit = findNearestHit(_photodots[i]);

            /* Find the central column of the nearest hit.
             */
            int hitCentreColumn = findCentreColumn(closestHit);

            /* Find the geometric centre of the nearest hit
             * Its output is (x,y) or (r,phi) as appropriate
             */
            Vector2D hitCentre = findHitCentre(hitCentreColumn);

            rError +=
                (hitCentre[0] - _expected_photodots[i]._r) /
                _expected_photodots[i]._r / _photodots.size();
            phiError +=
                (hitCentre[1] - _expected_photodots[i]._phi) /
                _expected_photodots[i]._phi / _photodots.size();

            /* Print out the photodot information
             */
            if (_geometry_type == "CARTESIAN")
                {

                    streamlog_out(DEBUG) << "Outputting Cartesian information\n";

                    // Original photodot
                    output_geometry << std::setw(9) << _photodots[i]._x;
                    output_geometry << std::setw(9) << _photodots[i]._y;
                    output_geometry << "  ";

                    // expected image
                    output_geometry << std::setw(9) << _expected_photodots[i]._x;
                    output_geometry << std::setw(9) << _expected_photodots[i]._y;
                    output_geometry << "  ";

                    if (_padLayoutType == PadRowLayout2D::POLAR)
                        {

                            output_geometry << std::setw(9) << hitCentre[0] *
                            std::cos(hitCentre[1]);
                            output_geometry << std::setw(9) << hitCentre[0] *
                            std::sin(hitCentre[1]);

                        } // /if polar pad layout

                    else
                        {

                            output_geometry << std::setw(9) << hitCentre[0];
                            output_geometry << std::setw(9) << hitCentre[1];

                        } // /if Cartesian pad layout

                } // /if Cartesian geometry

            else if (_geometry_type == "POLAR")
                {

                    streamlog_out(DEBUG) << "Outputting polar information\n";

                    // Original photodot
                    output_geometry << std::setw(9) << _photodots[i]._r;
                    output_geometry << std::setw(9) << _photodots[i]._phi;
                    output_geometry << "  ";

                    // Expected image
                    output_geometry << std::setw(9) << _expected_photodots[i]._r;
                    output_geometry << std::setw(9) << _expected_photodots[i]._phi;
                    output_geometry << "  ";

                    if (_padLayoutType == PadRowLayout2D::CARTESIAN)
                        {

                            output_geometry << std::setw(9) << std::sqrt(hitCentre[0] *
                                                                         hitCentre[0] +
                                                                         hitCentre[1] *
                                                                         hitCentre [1]);
                            output_geometry << std::setw(9) << std::atan2(hitCentre[1],
                                                                          hitCentre[0]);

                        } // /if Cartesian pad layout

                    else
                        {

                            output_geometry << std::setw(9) << hitCentre[0];
                            output_geometry << std::setw(9) << hitCentre[1];

                        } // /if polar pad layout

                } // /if polar geometry

            output_geometry << std::endl;

        } // /for each photodot

    output_geometry.close();

} // /processEvent()

void PhotodotReconstructionProcessor::check(LCEvent * evt) {}

void PhotodotReconstructionProcessor::end() {}

int PhotodotReconstructionProcessor::findNearestHit(Photodot thisPhotodot)
{

    // Vector holding ( padIndex, distanceToPad )
    Vector2D closestSoFar(0.0, 999999999.9);

    /* Analyse each pulse for the closest one
     */
    for (std::multimap< int, TrackerPulse * >::iterator iPulse =
                _pulsesByIndex->begin();
            iPulse != _pulsesByIndex->end();
            iPulse++)
        {

            // Get the centre of the pad being tested:
            Vector2D currentPad = _padLayout->getPadCenter(iPulse->first);
            // Distance to that pad
            double padDistance = 0.;

            /* Calculate the distance to that pad
             */
            switch (_padLayoutType)
                {

                case PadRowLayout2D::POLAR:
                    padDistance =
                        thisPhotodot._r * thisPhotodot._r +
                        currentPad[0] * currentPad[0] -
                        2 * thisPhotodot._r * currentPad[0] *
                        cos(fabs(thisPhotodot._phi - currentPad[1]));
                    break;

                case PadRowLayout2D::CARTESIAN:
                    padDistance =
                        (thisPhotodot._x - currentPad[0]) *
                        (thisPhotodot._x - currentPad[0]) +
                        (thisPhotodot._y - currentPad[1]) *
                        (thisPhotodot._y - currentPad[1]);
                    break;
                } // /switch calculate pad distance

            if (padDistance < closestSoFar[1])
                {
                    closestSoFar[0] = iPulse->first;
                    closestSoFar[1] = padDistance;
                } // /if change closest pulse

        } // /for each pulse

    return (int)closestSoFar[0];

} // /findNearestHit()



int PhotodotReconstructionProcessor::findCentreColumn(int currentPulse)
{

    /* This loop looks at the existence of the adjacent pad (assigning the pad
     * to nextPulse if it exists), then checks if the charge on that pad exceeds
     * the charge on the current pad. If it does, the loop repeats and reassigns
     * nextPad to currentPulse.
     */
    // Cell ID of the next pulse to be compared to the current pulse
    int nextPad = currentPulse;

    bool newPad = 1;
    while (newPad)
        {

            // Update the ID of the pulse being tested
            currentPulse = nextPad;
            newPad = 0;

            // Check the left pad
            if (_pulsesByIndex->find(nextPad = _padLayout->getLeftNeighbour(currentPulse)) != _pulsesByIndex->end() &&
                    _pulsesByIndex->find(nextPad)->second->getCharge() > _pulsesByIndex->find(currentPulse)->second->getCharge())
                newPad = 1;

            // Check the right pad
            else if (_pulsesByIndex->find(nextPad = _padLayout->getRightNeighbour(currentPulse)) != _pulsesByIndex->end() &&
                     _pulsesByIndex->find(nextPad)->second->getCharge() > _pulsesByIndex->find(currentPulse)->second->getCharge())
                newPad = 1;

            // move the current pad to the bottom row, to simplify later calculations
            else
                switch (_padLayoutType)
                    {

                    case PadRowLayout2D::CARTESIAN:
                        if (_pulsesByIndex->find(nextPad = _padLayout->getNearestPad((_padLayout->getPadCenter(currentPulse))[0], (_padLayout->getPadCenter(currentPulse))[1] + _padLayout->getPadHeight(currentPulse))) != _pulsesByIndex->end())
                            newPad = 1;
                        break;

                    case PadRowLayout2D::POLAR:
                        if (_pulsesByIndex->find(nextPad = _padLayout->getNearestPad((_padLayout->getPadCenter(currentPulse))[0] - _padLayout->getPadHeight(currentPulse), (_padLayout->getPadCenter(currentPulse))[1])) != _pulsesByIndex->end())
                            newPad = 1;
                        break;

                    }// /switch pad type

        } // /while searching for new pads

    // The current pad is now the bottom of the hit's centre column
    return currentPulse;

} // /findCentreColumn()

Vector2D PhotodotReconstructionProcessor::findHitCentre(int centreColumn)
{

    // Centre of the row in the each case (to be calculated repeatedly)
    double phiHitCentre = 0;
    double xHitCentre = 0;

    // Total charge
    double totalCharge = 0;

    // Charge on the lower part of the hit
    double lowerCharge = 0;

    // Vector holding the charge of each row
    std::vector< double > chargeByRow;

    // Distance to the lower edge of the upper part of the hit
    double upperRowBase = 0;

    // cell ID of the next row to be tested
    int thisRowCentre = centreColumn;

    /* Find the sum of the charges in each row
     */
    do
        {

            // Sum of the charges on each side
            double leftChargeSum, rightChargeSum;
            leftChargeSum = rightChargeSum = 0;

            /* sum the charges to the right of the centre column
             */
            int thisPad = thisRowCentre;

            while (_pulsesByIndex->find(thisPad = _padLayout->getRightNeighbour(thisPad)) != _pulsesByIndex->end())

                rightChargeSum += getPadCharge(thisPad);

            /* sum the charges to the left of the centre column
             */
            thisPad = thisRowCentre;
            do

                leftChargeSum += getPadCharge(thisPad);

            while (_pulsesByIndex->find(thisPad = _padLayout->getLeftNeighbour(thisPad)) != _pulsesByIndex->end());

            chargeByRow.push_back(leftChargeSum + rightChargeSum);
            totalCharge += leftChargeSum + rightChargeSum;

            /* Find the next row, and other operations as appropriate
             */
            switch (_padLayoutType)
                {

                case PadRowLayout2D::POLAR:

                    /* First, find the row's hit centre in a polar layout
                     */
                    phiHitCentre +=
                        // Get the upper side of this pad (boundary between left and right)
                        (_padLayout->getPadCenter(thisRowCentre)[1] +
                         0.5 * _padLayout->getPadWidth(thisRowCentre) -
                         // Get the charge centre for this row
                         std::sqrt(2.) * _padLayout->getPadWidth(thisRowCentre) * _diffusion *
                         TMath::ErfInverse(2 * leftChargeSum / (leftChargeSum +
                                                                rightChargeSum) - 1.0))
                        // Prepare the weighted mean
                        * (leftChargeSum + rightChargeSum);

                    /* Now find the next row
                     */
                    thisRowCentre = _padLayout->getNearestPad((_padLayout->getPadCenter(thisRowCentre))[0] + _padLayout->getPadHeight(thisRowCentre), (_padLayout->getPadCenter(thisRowCentre))[1]);
                    break;

                case PadRowLayout2D::CARTESIAN:

                    /* Find the row's hit centre
                     */
                    xHitCentre +=
                        // Get the upper edge of this pad)
                        (_padLayout->getPadCenter(thisRowCentre)[0] +
                         0.5 * _padLayout->getPadWidth(thisRowCentre) -
                         // Get the charge centre for this row
                         std::sqrt(2.) * _diffusion *
                         TMath::ErfInverse(2 * leftChargeSum / (leftChargeSum +
                                                                rightChargeSum) - 1.0))
                        // Prepare the weighted mean
                        * (leftChargeSum + rightChargeSum);

                    /* Now find the next row
                     */
                    thisRowCentre = _padLayout->getNearestPad((_padLayout->getPadCenter(thisRowCentre))[0], (_padLayout->getPadCenter(thisRowCentre))[1] + _padLayout->getPadHeight(thisRowCentre));
                    break;

                } // /switch

        } // do sum the charges on a line
    while (_pulsesByIndex->find(thisRowCentre) != _pulsesByIndex->end());

    std::vector< double >::iterator iCharge;

    /*
     * Calculate the centre of the hit
     */
    Vector2D hitCentre;
    switch (_padLayoutType)
        {

        case PadRowLayout2D::POLAR:


            /* Calculate the weighted mean of the phi-positions found for the rows
             */
            phiHitCentre /= totalCharge;

            /* Columns:
             */
            iCharge = chargeByRow.begin();

            double rHitCentre;

            /* Make sure there is sufficient charge on the bottom row for an accurate
             * result. If not, there are probably at least three rows.
             * If there are not three rows, it doesn't have a choice.
             */
            if (((lowerCharge += *iCharge) / totalCharge < 0.2)
                    && ++(++iCharge) != chargeByRow.end())
                {

                    iCharge--;
                    lowerCharge += *iCharge;
                    upperRowBase =
                        _padLayout->getPadCenter(centreColumn)[0] +
                        1.5 * _padLayout->getPadHeight(centreColumn);

                } // /if the charge on the bottom row is too small
            else

                upperRowBase =
                    _padLayout->getPadCenter(centreColumn)[0] +
                    0.5 * _padLayout->getPadHeight(centreColumn);

            rHitCentre =
                // Get the lower edge of the upper half of the hit
                (upperRowBase -
                 // Get the radial charge centre (radial diffusion must be found)
                 std::sqrt(2.) * _diffusion *
                 TMath::ErfInverse(2 * lowerCharge / totalCharge - 1.0));

            hitCentre = Vector2D(rHitCentre, phiHitCentre);

            break;

            /* Cartesian case
             */
        case PadRowLayout2D::CARTESIAN:

            /* Calculate the weighted mean of the x-positions for the rows
             */
            xHitCentre /= totalCharge;

            /* Columns:
             */
            iCharge = chargeByRow.begin();

            double yHitCentre;

            /* Make sure there is sufficient charge on the bottom row for an accurate
             * result.
             */
            if (((lowerCharge += *iCharge) / totalCharge < 0.01)
                    && ++(++iCharge) != chargeByRow.end())
                {

                    iCharge--;
                    lowerCharge += *iCharge;
                    upperRowBase =
                        _padLayout->getPadCenter(centreColumn)[1] +
                        1.5 * _padLayout->getPadHeight(centreColumn);

                } // /if the charge on the bottom row is too small
            else

                upperRowBase =
                    _padLayout->getPadCenter(centreColumn)[1] +
                    0.5 * _padLayout->getPadHeight(centreColumn);

            yHitCentre =
                // Get the lower edge of the upper half of the hit
                (upperRowBase -
                 // Get the y charge centre
                 std::sqrt(2.) * _diffusion *
                 TMath::ErfInverse(2 * lowerCharge / totalCharge - 1.0));

            hitCentre = Vector2D(xHitCentre, yHitCentre);

            break;

        } // /switch pad layout type

    return hitCentre;

} // /findHitCentre()

void PhotodotReconstructionProcessor::findPhotodotPulses()
{

    /* The upper and lower limits of the times to be considered to be valid
     * pulses.
     * These values are set in case the more complex method defined below
     * doesn't work -- this will happen if the photopulses were not created at
     * time 0.
     */
    double arrivalTime_low = 64100.;
    double arrivalTime_high = 64400.;

    /* Valid arrival times are calculated from the expected drift times for
     * each photodot.
     */
    double currentLowTime = 999999999.9;
    double currentHighTime = 0.;
    double meanTime = 0.;
    for (std::vector< Photodot >::iterator iPhotodot = _photodots.begin();
            iPhotodot != _photodots.end();
            iPhotodot++)
        {

            /* Drift the photodots through the tracker
             */
            double expectedHitCentre[4] = { iPhotodot->_x, iPhotodot->_y, 0., 0. };
            ElectronCloudDrifterProcessor::drift_through_cylinder(expectedHitCentre,
                                                                  _tpc_inner_radius,
                                                                  _tpc_outer_radius,
                                                                  _inner_z,
                                                                  _outer_z,
                                                                  _drift_velocity,
                                                                  _omega_tau);

            _expected_photodots.push_back(Photodot(expectedHitCentre[0],
                                                   expectedHitCentre[1],
                                                   iPhotodot->_diameter));

            /* Adjust the current time range, if need be.
             */
            currentLowTime = expectedHitCentre[3] < currentLowTime ?
                             expectedHitCentre[3] :
                             currentLowTime;
            currentHighTime = expectedHitCentre[3] > currentHighTime ?
                              expectedHitCentre[3] :
                              currentHighTime;

            meanTime += expectedHitCentre[3] / _photodots.size();

        } // /for each photodot

    /* The mean arrival time for expected photodots is for some reason 70ns
     * over the arrival time for images.
     */
    meanTime -= 0.070;
    /* 50ns added to give a wider range to search over, to ensure all pulses
     * are covered.
     */
    double timeRange = currentHighTime - currentLowTime + 0.05;

    /* It *appears* that the most prevalent pulse time in the event occurs at
     * 2/5 of the distribution.
     */
    arrivalTime_low = meanTime - 0.4 * timeRange;
    arrivalTime_high = meanTime + 0.6 * timeRange;

    std::set< double > hitPads;

    /* Check that there are no pulses outside the desired range and that no
     * pad gets multiple pulses
     */
    for (std::multimap< int, EVENT::TrackerPulse * >::iterator iPulse =
                _pulsesByIndex->begin();
            iPulse != _pulsesByIndex->end();
            iPulse++)
        {

            if (iPulse->second->getTime() < arrivalTime_low * 1.E3 ||
                    iPulse->second->getTime() > arrivalTime_high * 1.E3)
                {

                    streamlog_out(WARNING) << "Pulse at " << iPulse->first
                    << " deleted due to invalid time ("
                    << iPulse->second->getTime() << ")"
                    << std::endl;

                    _pulsesByIndex->erase(iPulse);

                } // /if the pulse's time is invalid
            else

                hitPads.insert(iPulse->first);

        } // /for each pulse

} // /indPhotodotPulses()

double PhotodotReconstructionProcessor::getPadCharge(int padIndex)
{

    double totalCharge = 0;

    /* The summing only needs to occur if there exists more than one pad
     * with a certain index
     */
    if (_pulsesByIndex->count(padIndex) > 1)
        {

            streamlog_out(WARNING) << "Pulse on pad #" << padIndex
            << " is a duplicate." << std::endl;

            for (std::multimap< int, EVENT::TrackerPulse * >::iterator iPulse =
                        _pulsesByIndex->lower_bound(padIndex);
                    iPulse != _pulsesByIndex->upper_bound(padIndex);
                    iPulse++)

                totalCharge += iPulse->second->getCharge();

        } // if multiple pulses exist at one index
    else

        totalCharge = _pulsesByIndex->find(padIndex)->second->getCharge();

    return totalCharge;

} // findPhotodotPulses()

} // /namespace marlintpc
