// MarlinTPC
#include "PhotoelectricLikelihoodSeederProcessor.h"

// Marlin
#include <marlin/Exceptions.h>
#include <marlin/Global.h>

// AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/ITupleFactory.h>
#include <AIDA/ITuple.h>

// CLHEP
#include <CLHEP/Vector/TwoVector.h>

// GEAR
#include <gear/TPCModule.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <gearxml/tinyxml.h>

// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerPulse.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>

using namespace CLHEP;
using namespace EVENT;
using namespace IMPL;
using namespace lcio;
using namespace gear;
using namespace marlin;
using namespace std;

namespace marlintpc
{


PhotoelectricLikelihoodSeederProcessor aPhotoelectricLikelihoodSeederProcessor;


PhotoelectricLikelihoodSeederProcessor::PhotoelectricLikelihoodSeederProcessor() : Processor("PhotoelectricLikelihoodSeederProcessor"),

        _input_photoelectric_tracker_pulses_collection_name("TPCPhotoelectricPulses"),
        _output_photodot_collection_name("TPCPhotodots"),
        _output_photoline_collection_name("TPCPhotolines"),
        _output_is_transient(true),
        _projected_photogeometry_file_name("projected_photogeometry.xml"),
        _sigma(2.5)
{

    /* description of the processor which will be displayed in the
    * steering file autometicly generated by
    * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
    */

    _description = "PhotoelectricLikelihoodSeederProcessor collects the TrackerPulses which belong to each projected photogeometry element and produces \"seed\" objects for the likelihood reconstruction algorithm.";

    /* register steering parameters: name, description, class-variable, default value
    * the type will by definded by the type of the default value
    * string, double, float and int are possible
    * plaese note to get a float you must cast in C++: (float) 1.3
    */

    registerInputCollection(LCIO::TRACKERHIT ,
                            "InputPhotoelectricTrackerPulses",
                            "The name of the input collection of TPCTrackerPulses from the photoelectric system.",
                            _input_photoelectric_tracker_pulses_collection_name,
                            std::string("TPCPhotoelectricPulses"));

    registerOutputCollection(LCIO::TRACKERHIT ,
                             "OutputSeedPhotodots",
                             "The name of the output collection of TrackerHits which represent seed photodots.",
                             _output_photodot_collection_name,
                             std::string("TPCSeedPhotodots"));

    registerOutputCollection(LCIO::TRACK ,
                             "OutputSeedPhotolines",
                             "The name of the output collection of Tracks which represent seed photolines.",
                             _output_photoline_collection_name,
                             std::string("TPCSeedPhotolines"));

    registerOptionalParameter("SetOutputTransient",
                              "Controls whether the output collection should be saved in the LCIO file (false) or if it should be deleted at the end (true - default).",
                              _output_is_transient,
                              true);

    registerOptionalParameter("ProjectedPhotogeometryFilename",
                              "This is the name of the collection which stores the true positions of the photodots as SimTrackerHit objects. It is optional but it provides the number of photodots and, if provided, the processor will order the seed photodots to match the order of true photodots (useful for analysis).",
                              _projected_photogeometry_file_name,
                              _projected_photogeometry_file_name);

}


void PhotoelectricLikelihoodSeederProcessor::init()
{

    // usually a good idea to
    printParameters();

    // load AIDA (if it's available)
    AIDA::ITuple *projected_photodot_tuple = 0;
    try
        {

            projected_photodot_tuple = AIDAProcessor::tupleFactory(this)->create("PhotodotProjectionInfo", "Photodot Projection Info", "int i_dot, int max_pads", "");
        }
    catch (lcio::Exception &e)
        {

            streamlog_out(DEBUG) << e.what() << endl;
            return;
        }

    // open the projection file and get the time interval we're watching for and the list of pads to take
    TiXmlDocument document(_projected_photogeometry_file_name.c_str());

    if (!document.LoadFile())
        {

            streamlog_out(ERROR) << "Couldn't load projected geometry file " << _projected_photogeometry_file_name << "." << endl;
            return;
        }

    int i_dot = 0;

    // get a list of global pad indices to search for
    TiXmlNode *photodot_node = document.RootElement()->FirstChild("photodot");

    for (; photodot_node; photodot_node = photodot_node->NextSibling("photodot"), i_dot++)
        {

            // create a set to hold the pulses for this photodot
            // even if there is no projection and it won't be used for fitting it must be stored in the lcio collection to keep all of the photodots in the correct order
            std::list< std::set< EVENT::TrackerPulse* > >::iterator new_pulse_set = _photodot_pulse_sets.insert(_photodot_pulse_sets.end(), set< EVENT::TrackerPulse* >());

            TiXmlNode *projection_node = photodot_node->FirstChild("projection");

            if (! projection_node)
                {

                    streamlog_out(DEBUG) << "photodot " << i_dot << " doesn't have a projection." << endl;
                    projected_photodot_tuple->fill(0, i_dot);
                    projected_photodot_tuple->fill(1, 0);
                    projected_photodot_tuple->addRow();
                    continue;
                }

            TiXmlElement *pad_element = projection_node->FirstChildElement("pad");

            int pad_count = 0;

            // add the pads
            for (; pad_element; pad_element = pad_element->NextSiblingElement("pad"))
                {

                    int pad_index;
                    int module_id;

                    pad_element->QueryIntAttribute("pad_index", &pad_index);
                    pad_element->QueryIntAttribute("module_id", &module_id);

                    GlobalPadIndex new_index(pad_index, module_id);

                    // assign this global index to this pulse collection
                    _photodot_pulse_set_by_global_index[ new_index ] = new_pulse_set;

                    pad_count += 1;
                }

            projected_photodot_tuple->fill(0, i_dot);
            projected_photodot_tuple->fill(1, pad_count);
            projected_photodot_tuple->addRow();

        }

    TiXmlNode *photoline_node = document.RootElement()->FirstChild("photoline");

    for (; photoline_node; photoline_node = photoline_node->NextSibling("photoline"))
        {

            TiXmlElement *pad_element = photoline_node->FirstChildElement("pad");

            for (; pad_element; pad_element = pad_element->NextSiblingElement("pad"))
                {

                    int pad_index;
                    int module_id;

                    pad_element->QueryIntAttribute("index", &pad_index);
                    pad_element->QueryIntAttribute("module", &module_id);

                    streamlog_out(DEBUG) << "Valid pad index " << pad_index << " module " << module_id << "." << endl;
                }
        }

}


void PhotoelectricLikelihoodSeederProcessor::processRunHeader(LCRunHeader* run)
{

}


void PhotoelectricLikelihoodSeederProcessor::processEvent(LCEvent * evt)
{

    // Step 1. Reset the pulse data because the TrackerPulses will be different from the last event
    for (std::list< std::set< EVENT::TrackerPulse* > >::iterator pulse_set = _photodot_pulse_sets.begin(); pulse_set != _photodot_pulse_sets.end(); pulse_set++)
        {

            pulse_set->clear();
        }

    // Step 2. Associate the TrackerPulses with the correct photodot or photoline
    LCCollection* input_photoelectric_pulses = evt->getCollection(_input_photoelectric_tracker_pulses_collection_name);

    streamlog_out(MESSAGE) << "Using " << input_photoelectric_pulses->getNumberOfElements() << " pulses from the collection " << _input_photoelectric_tracker_pulses_collection_name << " to seed " << _photodot_pulse_sets.size() << " photodots." << endl;

    // now determine if this is a photoelectric pulse
    for (int i_tracker_pulse = 0; i_tracker_pulse < input_photoelectric_pulses->getNumberOfElements(); i_tracker_pulse++)
        {

            TrackerPulse* input_pulse = dynamic_cast<TrackerPulse*>(input_photoelectric_pulses->getElementAt(i_tracker_pulse)) ;

            if (!input_pulse)
                {

                    throw(SkipEventException(this));
                }

            const GlobalPadIndex global_index(input_pulse->getCellID0(), input_pulse->getCellID1());

            // I can't wait for the new auto keyword...
            const std::map< gear::GlobalPadIndex, std::list< std::set< EVENT::TrackerPulse* > >::iterator >::iterator i_pulse_set = _photodot_pulse_set_by_global_index.find(global_index);

            if (i_pulse_set != _photodot_pulse_set_by_global_index.end())
                {

                    i_pulse_set->second->insert(input_pulse);
                }
        }

    // Step 3. Loop through the photodots, creating TrackerHits with seed positions
    const TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();

    int i_dot = 0;

    // create a collection to store the photodot seeds
    LCCollectionVec* output_seed_photodots = new LCCollectionVec(LCIO::TRACKERHIT);
    output_seed_photodots->setTransient(_output_is_transient);
    output_seed_photodots->setSubset(false);

    for (list< set< TrackerPulse* > >::const_iterator i_photodot = _photodot_pulse_sets.begin(); i_photodot != _photodot_pulse_sets.end(); i_photodot++, i_dot++)
        {

            TrackerHitImpl *seed_photodot = new TrackerHitImpl();

            // again, make sure that the photodot is added to the collection to preserve ordering
            output_seed_photodots->addElement(seed_photodot);

            if (i_photodot->size() == 0)
                {

                    streamlog_out(DEBUG) << "photodot " << i_dot << " did not receive any data this event." << endl;
                    continue;
                }

            Hep2Vector seed_position;
            double total_charge = 0.;

            // add all of the TrackerPulses to the hit
            for (set< TrackerPulse* >::const_iterator i_pulse = i_photodot->begin(); i_pulse != i_photodot->end(); i_pulse++)
                {

                    Vector2D pad_centre = tpc_parameters.getModule((*i_pulse)->getCellID1()).getPadCenter((*i_pulse)->getCellID0());

                    if (tpc_parameters.getCoordinateType() == PadRowLayout2D::POLAR)
                        {

                            // convert to Cartesian....
                            const double r = pad_centre[0];
                            pad_centre[0] = r * cos(pad_centre[1]);
                            pad_centre[1] = r * sin(pad_centre[1]);
                        }

                    seed_position += (*i_pulse)->getCharge() * Hep2Vector(pad_centre[0], pad_centre[1]);
                    total_charge += (*i_pulse)->getCharge();

                    // add the TrackerPulse to the photodot....
                    seed_photodot->rawHits().push_back((*i_pulse));
                }

            seed_position *= 1. / total_charge;

            double position[3];
            position[0] = seed_position[0];
            position[1] = seed_position[1];
            position[2] = 0.;

            seed_photodot->setPosition(position);

            streamlog_out(DEBUG) << "Seeded photodot " << i_dot << " with position " << seed_position << "." << endl;
        }

    // now add the collection to the event
    evt->addCollection(output_seed_photodots, _output_photodot_collection_name);
}



void PhotoelectricLikelihoodSeederProcessor::check(LCEvent * evt)
{

    // nothing to check here - could be used to fill checkplots in reconstruction processor
}


void PhotoelectricLikelihoodSeederProcessor::end()
{

}

} // namespace marlintpc
