#ifndef CalibParam_h
#define CalibParam_h 1

//C++
#include <iostream>
#include <string>
#include <utility>

//LCIO
#include "lcio.h"
#include "UTIL/LCFixedObject.h"

#define CalibParamNINTVals 2       // parameter index and chipID 
#define CalibParamNFLOATVals 1     // Parameters of the calibration (4-ToT calibration, 2-THL offset, 2-Time Walk)
#define CalibParamNDOUBLEVals 0

namespace tpcconddata
{

class CalibParam;

typedef CalibParam LCCalibParam;

/** Class that reads the parameters from the calibration data,
 *  for a single chip the data format should be: (1)ToT (2)THL offset (3)TimeWalk
 *  4 ToT + 2 THL offset + 2 TimeWalk = 8 parameters.
 *  for a quad board the data format in "pixelman notation" should be: (1)chip1 (2)chip0 (3)chip2 (4)chip3
 *  so that together will have 32 parameters, differed by their index.
 *
 *  Combining them into a single object of type LCGenericObject.
 *  Based on the LCFixedObject template.
 *  This can be used to be stored in a LCCollectionVec.
 *
 * @author R.Ulman, Bonn
 */


class CalibParam : public UTIL::LCFixedObject < CalibParamNINTVals,
        CalibParamNFLOATVals,
        CalibParamNDOUBLEVals >
{

    public:

        /** Convenient constructor using two integers and a  float for calibration parameters.
         *  With the chipID (chip), the parameter index (param),
         *  and the calibration parameter (cP)
         */
        CalibParam(int chip, int param, float cP);

        /** 'Copy constructor' needed to interpret LCCollection read from file/database.
         */
        CalibParam(EVENT::LCObject* obj): UTIL::LCFixedObject < CalibParamNINTVals,
                CalibParamNFLOATVals,
                CalibParamNDOUBLEVals > (obj) { }

        /** Important for memory handling*/
        virtual ~CalibParam();


        // the class interface:
        int getChip() const;    ///<- get chipID
        int getParam() const;   ///<- get parameter index
        float getCP() const;    ///<- get calibration parameter

        void setChip(const int &chip) ;
        void setParam(const int &param) ;
        void setCP(const float &cP) ;

        static std::string getRevision();
        static std::string getDefaultColName();

        void print(std::ostream& os = std::cout) const;

        // -------- need to implement abstract methods from LCGenericObject
        const std::string getTypeName() const
        {
            return std::string("CalibrationParameters");
        }
        const std::string getDataDescription() const
        {
            return std::string("i:chipID,i:index,f:parameter");
        }


    protected:


};// end of class

std::ostream &operator<<(std::ostream &os, const CalibParam &p);

} //end namespace tpcconddata
#endif
