#ifndef TRACK_FINDER_RECTANGULAR_PROCESSOR_H
#define TRACK_FINDER_RECTANGULAR_PROCESSOR_H

// C++
#include <string>
#include <vector>

// LCIO
#include <lcio.h>

// Marlin
#include <marlin/Processor.h>
#include <EVENT/LCEvent.h>
#include <EVENT/LCRunHeader.h>
#include <IMPL/TrackerHitImpl.h>
#include <EVENT/Track.h>
#include <gear/PadRowLayout2D.h>

namespace marlintpc
{

/** Processor to find linear tracks using linear regression,
 *  based on a modified version of the track following algorithm.
 *  It assumes the hits being equidistantly spaced in y,
 *  which is the case for a TPC with rectangular pad layout (straight rows).
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of TrackerHits, sorted in y, x and z (in this order).
 *  A sorting function might be implemented in a later version
 *  or provided as a separate processor. For the time being
 *  the HitFinder has provide sorted hits.
 *
 *  <h4>Output</h4>
 *  Collection of Tracks
 *
 *  @param InputCollectionName Name of input data collection (default: TPCHits)
 *  @param OutputCollectionName Name of output data collection (default: TPCSeedTracks)
 *  @param SetOutputTransient If not 0 the output collection is set transient (default: 0)
 *
 *  This processor is intended to work as a finder, the fitting is done by a more advanced fitting algorithm.
 *  In this case the output can be set transient, so the seed tracks are not written to disk.<br>
 *  As the finding algorithm performs a linear regression, which is equivalent to a
 *  \f$\chi^2\f$ minimisation for straight tracks, all track parameters
 *  are available and the track fitting can be skipped.
 *  In this case the output collection name should be changed to TPCTracks.
 *
 *  @param MinTrackHits Minimum number of hits on the track
 *  @param MaxSkipRows Maximum number of subsequently missing hits
 *  @param DeltaX  Maximum x distances between predicted and measured hit
 *  @param DeltaY  Maximum y distances between predicted and measured hit
 *  @param DeltaZ  Maximum z distances between predicted and measured hit
 */
class TrackFinderRectangularProcessor : public marlin::Processor
{

    public:

        virtual Processor*  newProcessor()
        {
            return new TrackFinderRectangularProcessor ;
        }


        TrackFinderRectangularProcessor() ;

        /** Called at the begin of the job before anything is read.
         * Use to initialize the processor, e.g. book histograms.
         */
        virtual void init() ;

        /** Called for every run.
         */
        virtual void processRunHeader(EVENT::LCRunHeader* run) ;

        /** Called for every event - the working horse.
         */
        virtual void processEvent(EVENT::LCEvent * evt) ;


        virtual void check(EVENT::LCEvent * evt) ;


        /** Called after data processing for clean up.
         */
        virtual void end() ;


    protected:

        std::string _inputColName ;    ///<- the name of the input collection
        std::string _outputColName ;   ///<- the name of the output collection

        /** give the status which is set to the transient flag of the output collection
         *  it is an int instead of a boolean, because the processor parameter cannot be a boolean
         */
        int _outputIsTransient;

        /// The pad layout as defined in GEAR
        gear::PadRowLayout2D const * _padLayout;


        int _rowHeight;
        int _minTrackHits; ///< Minimum number of hits on track
        int _maxSkipRows;    ///< Maximum number of missing hits

        /// Maximum distances between predicted and measured hit
        float _deltaX, _deltaY, _deltaZ;

        /** !!!!! This information should come from the conditions data!
         *  Implemented as processor parameter for first tests!
         */


        int _nRun ;///< The number of runs
        int _nEvt ;///< The total number of events

        /** Flag about verbosity of the processor.
         *  To help to understand and to debug the algorithm
         * a lot of output can be switched on.
         * \li \c 0 : quiet, only report errors
         * \li \c 1 : normal, default, status of processor parameters and event statistics
         * \li \c 2 : rebort about every event
         * \li \c 3 : report about every track candidate
         * \li \c 4 : report about every hit
         * \li \c 5 : rebort about everything
         */
        int _verbosityLevel;

        bool *isInTrack; ///<Array of flags whether a hit is in a track
        bool *isInTrackCandidate;  ///<Array of flags whether a hit is in the current track candidate
        unsigned int nHits; ///< Number of hits in the current event

        /// Set flags if hit is in a track to zero
        void zeroFlags(bool *IsIn);

        /** Add the hits of the track candidate to the
         *  isInTrack flag array.
         */
        void addTrackCandidateFlags();


        void createNewIsInTrackFlags();

        /// Extrapolate the next expected hit on the track
        IMPL::TrackerHitImpl extrapolateNextHit(double y);

        /** Straight line for linear regression.<br>
         *  The staright line is defined with 4 parameters:<br>
         *  x=ay+b  ;  z=cy+d
         */
        struct
        {
            float a, b, c, d;
        } straightLine;

        /** Calculate the track parameters from the hits.
         *  This is done by linear regression, which is
         *  equivalent to a \f$\chi^2\f$ minimisation for
         * a straight line.
         *
         * The straight line is parameterised
         * in the xy and the zy plane:                          <br>
         * x=ay+b  ; z=c*y+d                                    <br>
         * Calculate a and b  ; c and d analytically
         * by linear regression.
         * (As track is going predominantly along the y
         * direction, this coordinate is set to the centre
         * of the pad and regarded without uncertainty)
         * \f[a=\frac{\sum y_i\sum x_i-n\sum y_i x_i}
         *    {(\sum y_i)^2 - n  \sum y_i^2}\f]
         * \f[b=\frac{\sum y_i\sum y_i x_i - n\sum y_i^2 x_i}
         *    {(\sum y_i)^2 - n  \sum y_i^2}\f]
         * (see for instance W. Walcher, Praktikum der Physik,
         *  Teubner Taschenb&uuml;cher Physik, 1994)
         */
        void doRegression(EVENT::Track *trackCandidate);
} ;

} // namespace martintpc

#endif//TRACK_FINDER_LINEAR_ROW_BASED_H



