#include "HitsInTracksSplitterProcessor.h"
#include "HitSplitter.h"
#include "LCObjectCopier.h"
#include "TPCConditions.h"

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/TrackImpl.h>
#include <IMPL/LCFlagImpl.h>

using namespace lcio;
using namespace marlin;
using namespace std;


namespace marlintpc
{

HitsInTracksSplitterProcessor aHitsInTracksSplitterProcessor;


HitsInTracksSplitterProcessor::HitsInTracksSplitterProcessor() : Processor("HitsInTracksSplitterProcessor")
{
    // the processor description
    _description = "HitsInTracksSplitterProcessor splits the hits in tracks and keeps the association to the tracks (grouping).";

    /* register steering parameters: name, description, class-variable, default value

      for example:
    */

    registerInputCollection(LCIO::TRACK,
                            "InputCollectionName" ,
                            "Name of the tracks with the hits to be split, default: TPCTracks"  ,
                            _inputColName ,
                            string("TPCTracks")) ;

    registerOptionalParameter("VDriftOverride" ,
                              "Use VDriftOverrride to overwrite the drift velocity from conditions data (in mm / us )",
                              _vDrift,
                              float(50.0));

    registerOutputCollection(LCIO::TRACKERHIT,
                             "SplitHitsCollectionName" ,
                             string("Name of the collection with split hits,") +
                             string(" default: OnlySplitHits"),
                             _splitHitsColName,
                             string("OnlySplitHits")) ;


    registerOutputCollection(LCIO::TRACKERPULSE,
                             "PulsesInSplitHitsCollectionName" ,
                             string("Name of the collection with pulses contained in the split hits,") +
                             string(" default: PulsesInSplitHits"),
                             _pulsesInSplitHitsColName ,
                             string("PulsesInSplitHits")) ;


    registerOutputCollection(LCIO::TRACK,
                             "OutputTracksCollectionName" ,
                             string("Name of the collection of tracks with split hits,") +
                             string(" default: TPCTracksWithSplitHits"),
                             _outputColName ,
                             string("TPCTracksWithSplitHits")) ;

    registerProcessorParameter("TPCConditonsCollectionName",
                               "Name of the TPCConditons collection (for vDrift), default: TPCConditions",
                               _tpcConditionsColName,
                               tpcconddata::TPCConditions::getDefaultColName());

}


void HitsInTracksSplitterProcessor::init()
{
    m_out(DEBUG) << " init() called  " << endl;

    // you might want to list the used parameters by invoking
    printParameters();

}


void HitsInTracksSplitterProcessor::processRunHeader(LCRunHeader* run)
{
    // do something for each run
}


void HitsInTracksSplitterProcessor::processEvent(LCEvent* evt)
{
    // do something for each event, usually access the input collection:

    // use an LCCollectionVec instead of LCCollection so we can use iterators later
    LCCollectionVec* inputColVec = 0;

    // do a try / catch in case there is an empty event
    try
        {
            inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection(_inputColName));
        }
    catch (DataNotAvailableException &e)
        {
            // no input data, just return and process the next event
            return;
        }

    // check whether the drift velocity is stored as conditions data
    try
        {
            LCCollection *TPCConditionsCollection = evt->getCollection(_tpcConditionsColName);
            tpcconddata::TPCConditions myTPCConditions(TPCConditionsCollection->getElementAt(0));

            // conditions data are available, but can be
            // overwritten by using processor parameter
            // give a warning in this case
            if (parameterSet("VDriftOverride"))
                {
                    streamlog_out(WARNING4) << _processorName << ": WARNING: drift velicity from conditions data is"
                    << " overwritten by processor parameter!" << std::endl
                    << "Remove parameter \"VDriftOverride\" from steering file to use vDrift from TPCCondData."
                    << std::endl;
                }
            else
                {
                    _vDrift = myTPCConditions.getDriftVelocity();
                }
        }
    catch (DataNotAvailableException &e)
        {
            // no conditions data available, try to use processor parameter
            if (!parameterSet("VDriftOverride"))
                {
                    throw(DataNotAvailableException(_processorName +
                                                    std::string(" no drift velocity available, provide either as TPCCondData or processor parameter")));
                }
        }

    // The collection containing the hits which have not been split and the split results
    // This collection is a subset so the non-split hits (the majority) is not duplicated.
    // Unfortunately this means that the splits hits have to go into a separate collection.
    LCCollectionVec * outputCollection = new LCCollectionVec(LCIO::TRACK);
    LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(LCIO::TRBIT_HITS) ;
    outputCollection->setFlag(trkFlag.getFlag());

    // The collection containing the split hits, and the pulses they contain
    LCCollectionVec * splitHitsCollection = new LCCollectionVec(LCIO::TRACKERHIT);
    LCCollectionVec * pulsesInSplitHitsCollection = new LCCollectionVec(LCIO::TRACKERPULSE);

    // initialise the splitter class
    HitSplitter hitSplitter(splitHitsCollection, pulsesInSplitHitsCollection, _vDrift);

    // loop the data collection and do something
    for (LCCollectionVec::iterator inputTrackIter = inputColVec->begin() ;
            inputTrackIter < inputColVec->end() ;
            inputTrackIter++)
        {

            Track * inputTrack = dynamic_cast<Track *>(*inputTrackIter);

            TrackImpl * trackWithSplitHits =  LCObjectCopier::copyTrackWithoutHits(inputTrack);;

            for (TrackerHitVec::const_iterator inputHitIter = inputTrack->getTrackerHits().begin();
                    inputHitIter < inputTrack->getTrackerHits().end();
                    ++inputHitIter)
                {

                    TrackerHitVec splitHitsVec = hitSplitter.splitHit(*inputHitIter);

                    // fill the output collection
                    if (splitHitsVec.size() == 0) // the hit has not been split
                        { // add the original hit
                            trackWithSplitHits->addHit(*inputHitIter);
                        }
                    else // fill the split hits into the output collection
                        {
                            for (TrackerHitVec::iterator hitIter = splitHitsVec.begin();
                                    hitIter < splitHitsVec.end(); ++hitIter)
                                {
                                    trackWithSplitHits->addHit(*hitIter);
                                }// for hitIter
                        }// else splitHitsVec.size()==0

                }// for inputHitIter

            // copy all the parameters from the input track
            /// \FIXME copy all the parameters from the input track

            // add the new track to the output collection
            outputCollection->addElement(trackWithSplitHits);

        }// for inputTrackIter

    // fill the event
    if (splitHitsCollection->getNumberOfElements() > 0)
        {
            evt->addCollection(splitHitsCollection, "OnlySplitHits");

            if (pulsesInSplitHitsCollection->getNumberOfElements() > 0)
                {
                    evt->addCollection(pulsesInSplitHitsCollection, "PulsesInSplitHits");
                }
        }
    evt->addCollection(outputCollection, "TPCTracksWithSplitHits");
}


void HitsInTracksSplitterProcessor::check(LCEvent* evt)
{
    // check what is needed here
}


void HitsInTracksSplitterProcessor::end()
{
    // do something at the end of the processing, e.g. cleaning up
}


} // close namespace brackets
