/*
 * PathFinderInterfaceProcessor.cc
 *
 *  Created on: Nov 4, 2011
 *      Author: caiazza
 */
#ifdef USE_PATHFINDER
#include "PathFinderInterfaceProcessor.h"
#include "TrackParameterFull.h"

#include "marlin/Exceptions.h"
#include "marlin/Global.h"

//LCIO Inclusion
#include "IMPL/LCCollectionVec.h"
#include "IMPL/TrackerHitImpl.h"
#include "IMPL/TrackImpl.h"

namespace marlintpc
{

PathFinderInterfaceProcessor aPathFinderInterfaceProcessor;

PathFinderInterfaceProcessor::PathFinderInterfaceProcessor()
        : marlin::Processor("PathFinderInterfaceProcessor")
{
    _description = "This processor provides an interface with the Track Finding package PATHFINDER";
    _myFinderParameter = 0;
    _trackFinder = 0;

    registerInputCollection(LCIO::TRACKERHIT, "InputCollectionName",
            "Name of the input collection containing the hits data", _inputColName,
            std::string("TPCHits"));

    registerOutputCollection(LCIO::TRACK, "OutputCollectionName",
            "Name of the collection where to store the tracks found", _outColName,
            std::string("FoundTracks"));

    registerProcessorParameter("WriteOutputToStorage",
            "If true the output collection is written to file", _outputIsPersistent, true);

    registerProcessorParameter(
            "TrackModel",
            "Defines the type of track to search. 0:Straight Track, 1:Helix Track (Default) 2:Curlers",
            _trackModel, 1);

    registerProcessorParameter("MinimumHitsonTrack",
            "Defines the minimum number of hits of a selected track", _minHitonTrack, 5);

    registerProcessorParameter(
            "MaxDistXY",
            "Defines the maximum distance allowed between a hit and the track in the XY plane in mm",
            _maxHitDistXY, static_cast<float>(0.5));

    registerProcessorParameter(
            "MaxDistZ",
            "Defines the maximum distance allowed between a hit and the track along the Z axis in mm",
            _maxHitDistZ, static_cast<float>(1.0));

    registerProcessorParameter(
            "BinningLevel",
            "Defines how fine the binning of the hough transformation map is. The values must range between 1(Coarse) and 10(Fine)",
            _binningLevel, 5);

    registerOptionalParameter("ActivateNeighborhoodSearch",
            "If true activates the neighborhood search algorithm of the pathfinder package",
            _activateNeighSearch, false);

    /*
     * TODO Implement the Coordinate Redefinition system which will all the hits in an event,
     * calculate the center of mass of those hits, perform the calculation using that point
     * as origin, then translates back the coordinate to the original center. This way the
     * calculation should be much more accurate because of the d0 lever arm being as small
     * as possible
    registerOptionalParameter(
            "ActivateCoordRedefinition",
            "If true it will calculate a temporary origin in the center of mass of the hits to increase the precision of the algorithm",
            _activateCoordRedef, false);*/

    registerOptionalParameter(
            "OverrideD0XYBins",
            "Manually define the number of bin of the Hough Map for the D0 search in the XY projection. Must not exceed 1000.",
            _overrideD0XYBins, 1000);

    registerOptionalParameter(
            "OverrideD0SZBins",
            "Manually define the number of bin of the Hough Map for the D0 search in the SZ projection. Must not exceed 1000.",
            _overrideD0SZBins, 1000);

    registerOptionalParameter(
            "OverrideThetaXYBins",
            "Manually define the number of bin of the Hough Map for the Theta search in the XY projection. Must not exceed 1000.",
            _overrideThetaXYBins, 1000);

    registerOptionalParameter(
            "OverrideThetaSZBins",
            "Manually define the number of bin of the Hough Map for the Theta search in the SZ projection. Must not exceed 1000.",
            _overrideThetaSZBins, 1000);

    registerOptionalParameter(
            "OverrideOmegaBins",
            "Manually define the number of bin of the Hough Map for the Omega search. Must not exceed 1000.",
            _overrideOmegaBins, 1000);
}

void
PathFinderInterfaceProcessor::init()
{
    _myFinderParameter = new FinderParameter();

    //Set the Pathfinder parameters based on the steering file
    //The write to root option is a debugging option of the Pathfinder package that the end user
    //should not use
    _myFinderParameter->setSaveRootFile(false);

    //Set the track model in the finder parameters
    switch (_trackModel)
    {
        case 0:
            _myFinderParameter->setFindCurler(false);
            _myFinderParameter->setIsHelix(false);
            _myFinderParameter->setIsStraightLine(true);
            break;
        case 1:
            _myFinderParameter->setFindCurler(false);
            _myFinderParameter->setIsHelix(true);
            _myFinderParameter->setIsStraightLine(false);
            break;
        case 2:
            _myFinderParameter->setFindCurler(true);
            _myFinderParameter->setIsHelix(false);
            _myFinderParameter->setIsStraightLine(false);
            break;
        default:
            m_out(ERROR) << "Track Model not recognized." << std::endl;
            throw marlin::StopProcessingException(this);
    }

    //Set the minimum number of hit on track
    if (_minHitonTrack > 1)
    {
        _myFinderParameter->setMinimumHitNumber(_minHitonTrack);
    }
    else
    {
        m_out(ERROR)
                    << "Invalid value for the minimum numbers of hit on a track. It must be more than 1.";
        throw marlin::StopProcessingException(this);
    }

    //Set the maximum distance in the XY plane
    _myFinderParameter->setMaxXYDistance(_maxHitDistXY);

    /*
     * Set the maximum distance in the sz plane
     * In most of the cases the distance in sz is almost the same as the distance in z
     * because most of the tracks are perpendicular to z (at least in the test beam data)
     * FIXME When Isa will introduce the conversion from the Z distance to the SZ distance
     * in the Pathfinder package this conversion will be performed internally to the package
     */
    _myFinderParameter->setMaxSZDistance(_maxHitDistZ);

    _myFinderParameter->setSearchNeighborhood(_activateNeighSearch);

    //Set the binning parameters of the Hough transform.
    const int BASEBINNING = 100;
    //D0 XY
    if (parameterSet("OverrideD0XYBins"))
    {
        if (_overrideD0XYBins > 0 && _overrideD0XYBins <= 1000)
        {
            _myFinderParameter->setNumberXYDzeroBins(_overrideD0XYBins);
        }
        else
        {
            m_out(ERROR) << "Invalid value set for the D0 Binning Override for the XY projection"
                        << std::endl;
            throw marlin::StopProcessingException(this);
        }
    }
    else
    {
        _myFinderParameter->setNumberXYDzeroBins(_binningLevel * BASEBINNING);
    }

    //D0 SZ
    if (parameterSet("OverrideD0SZBins"))
    {
        if (_overrideD0SZBins > 0 && _overrideD0SZBins <= 1000)
        {
            _myFinderParameter->setNumberXYDzeroBins(_overrideD0SZBins);
        }
        else
        {
            m_out(ERROR) << "Invalid value set for the D0 Binning Override for the SZ projection"
                        << std::endl;
            throw marlin::StopProcessingException(this);
        }
    }
    else
    {
        _myFinderParameter->setNumberSZDzeroBins(_binningLevel * BASEBINNING);
    }

    //Theta XY
    if (parameterSet("OverrideThetaXYBins"))
    {
        if (_overrideThetaXYBins > 0 && _overrideThetaXYBins <= 1000)
        {
            _myFinderParameter->setNumberXYDzeroBins(_overrideThetaXYBins);
        }
        else
        {
            m_out(ERROR)
                        << "Invalid value set for the Theta Binning Override for the XY projection"
                        << std::endl;
            throw marlin::StopProcessingException(this);
        }
    }
    else
    {
        _myFinderParameter->setNumberXYThetaBins(_binningLevel * BASEBINNING);
    }

    //Theta SZ
    if (parameterSet("OverrideThetaSZBins"))
    {
        if (_overrideThetaSZBins > 0 && _overrideThetaSZBins <= 1000)
        {
            _myFinderParameter->setNumberXYDzeroBins(_overrideThetaSZBins);
        }
        else
        {
            m_out(ERROR)
                        << "Invalid value set for the Theta Binning Override for the SZ projection"
                        << std::endl;
            throw marlin::StopProcessingException(this);
        }
    }
    else
    {
        _myFinderParameter->setNumberSZThetaBins(_binningLevel * BASEBINNING);
    }

    //Omega XY
    if (parameterSet("OverrideOmegaBins"))
    {
        if (_overrideOmegaBins > 0 && _overrideOmegaBins <= 1000)
        {
            _myFinderParameter->setNumberXYDzeroBins(_overrideOmegaBins);
        }
        else
        {
            m_out(ERROR) << "Invalid value set for the Omega Binning Override" << std::endl;
            throw marlin::StopProcessingException(this);
        }
    }
    else
    {
        _myFinderParameter->setNumberXYOmegaBins(_binningLevel * BASEBINNING);
    }

    _myFinderParameter->print();

    _nRun = 0;
    _nEvt = 0;

}

void
PathFinderInterfaceProcessor::processRunHeader(EVENT::LCRunHeader* run)
{
    run->parameters().setValue(_processorName + "_revision", "$Rev: 2806 $");
    for (marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
    {
        if (!i->second->isOptional() || i->second->valueSet())
        {
            run->parameters().setValue(_processorName + "_" + i->second->name(),
                    i->second->value());
        }
    }

    _nRun++;
}

void
PathFinderInterfaceProcessor::processEvent(EVENT::LCEvent* evt)
{
    LCCollectionVec* inputCol = 0;
    try
    {
        inputCol = dynamic_cast<LCCollectionVec*>(evt->getCollection( _inputColName ));
    }
    catch (DataNotAvailableException &e)
    {
      // no input data, just return and process the next event
      m_out(MESSAGE) << "No hits in event " << evt->getEventNumber() << std::endl;
      return;
    }

    m_out(DEBUG) << "Initializing the Hough Transformation algorithm." << std::endl;
    _trackFinder = new HoughTrafoTrackFinder();
    _trackFinder->setFinderParameter(*_myFinderParameter);

    //Convert the LCIO Hits in Pathfinder Basic Hits
    std::vector<basicHit> PFHitVector;
    PFHitVector.clear();

    for(LCCollectionVec::iterator itHit = inputCol->begin(); itHit != inputCol->end(); itHit++)
    {
        TrackerHitImpl* LCIOHit = dynamic_cast<TrackerHitImpl*>(*itHit);
        basicHit PFHit = basicHit(LCIOHit->getPosition()[0], LCIOHit->getPosition()[1], LCIOHit->getPosition()[2], LCIOHit);
        PFHitVector.push_back(PFHit);
    }

    _trackFinder->setInitialHits(PFHitVector);
    m_out(DEBUG) << "Track Finding..." << std::flush;
    bool trackFound = _trackFinder->find();
    m_out(DEBUG) << "complete." << std::endl;

    LCCollectionVec* Tracks = new LCCollectionVec(LCIO::TRACK);
    Tracks->setTransient(!_outputIsPersistent);
    if(trackFound)
    {
        std::vector<TrackFinderTrack> foundTracks = _trackFinder->getTracks();
        m_out(MESSAGE) << foundTracks.size() << " tracks found in the event." << std::endl;
        for(std::vector<TrackFinderTrack>::iterator itTrack = foundTracks.begin() ; itTrack != foundTracks.end(); itTrack++)
        {
            TrackParameterFull TrackPar = itTrack->getTrackParameter();
            TrackImpl* LCIOTrack = new TrackImpl();
            LCIOTrack->setD0(TrackPar.getDZero());
            LCIOTrack->setPhi(TrackPar.getPhi());
            LCIOTrack->setTanLambda(TrackPar.getTanLambda());
            LCIOTrack->setZ0(TrackPar.getZZero());
            if(_trackModel == 0)
            {
                LCIOTrack->setOmega(0);
            }
            else
            {
                LCIOTrack->setOmega(TrackPar.getOmega());
            }

            PFHitVector.clear();
            PFHitVector = itTrack->getHitsOnTrack();
            for(std::vector<basicHit>::iterator PFHit = PFHitVector.begin(); PFHit != PFHitVector.end() ; PFHit++)
            {
                //retrieve original lcio hit and add to output track
                TrackerHitImpl* LCIOHit = static_cast<TrackerHitImpl*> (PFHit->getInitialHit());
                LCIOTrack->addHit(LCIOHit);
            }

            Tracks->addElement(LCIOTrack);
        }
        evt->addCollection(Tracks, _outColName);
    }
    else
    {
        m_out(MESSAGE) << "No tracks found in event: " << evt->getEventNumber() << std::endl;
    }
}

void
PathFinderInterfaceProcessor::check(EVENT::LCEvent* evt)
{

}

void
PathFinderInterfaceProcessor::end()
{
    if(_myFinderParameter != 0)
    {
        delete _myFinderParameter;
        _myFinderParameter= 0;
    }

    if(_trackFinder != 0)
    {
        delete _trackFinder;
        _trackFinder= 0;
    }
}

}

#endif
