#ifndef ELECTRONCLOUDDRIFTERPROCESSOR_H
#define ELECTRONCLOUDDRIFTERPROCESSOR_H 1

// CLHEP
#include <CLHEP/Vector/ThreeVector.h>

// Marlin
#include <marlin/Processor.h>

// GSL
#include <gsl/gsl_errno.h>
#include <gsl/gsl_odeiv.h>

namespace EVENT {

	class LCRunHeader;
	class LCEvent;
}

namespace marlintpc {

/** ElectronCloudDrifterProcessor : This processor takes a collection of electron clouds from a previous processor and drifts them a supplied distance. The parameters of the processor are designed to specify the area in which the drift occurs. It currently uses a adaptive Runge-Kutta-Fehlberg method for drifting the electrons in a non-uniform magnetic field.
 *
 *    @author J. Abernathy (UVic)<br>
 */
class ElectronCloudDrifterProcessor : public marlin::Processor {

public:
	/** constructor
	 */
	ElectronCloudDrifterProcessor();

	/** destructor
	 */
	~ElectronCloudDrifterProcessor();

	/** used by Marlin to create a new processor
	 */
	virtual Processor* newProcessor() { return new ElectronCloudDrifterProcessor() ; }

	/** Called at the begin of the job before anything is read.
	 * Use to initialize the processor, e.g. book histograms.
	 */
	virtual void init();

	/** Called for every run
	 */
	virtual void processRunHeader( EVENT::LCRunHeader* run );

	/** Called for every event - the working horse.
	 */
	virtual void processEvent( EVENT::LCEvent * evt );

	/** Called after the event has been processed
	 */
	virtual void check( EVENT::LCEvent * evt );

	/** Called after data processing for clean up.
	 */
	virtual void end();

	/** the derivative function used for the GSL ode solving function
	 */
	static int ion_drift_derivative( double t, const double y[], double f[], void *params );

	/** the static drifting function that can be used by other processors (ie the LikelihoodFitter)
	 */
	static int drift_through_cylinder( double (&)[4], const double min_radius, const double max_radius, const double min_z, const double max_z, const double drift_velocity, const double omega_tau );

protected:

	struct DriftParameters {

		double _drift_velocity;
		double _omega_tau;
		double _omega_tau_squared;
	};

private:

	/** the name of the input electron cloud collection
	 */
	std::string _input_collection_name;

	/** the name to use for the tpc conditions collection
	 */
	std::string _tpc_conditions_collection_name;

	/** the name of this section of the tpc
	 */
	std::string _tpc_section_name;

	/** set as true if this drift section is the main drift section of the TPC
	 */
	bool _primary_drift_section;

	/** the gas parameterization of this piece of the TPC
	 */
	double _omega_tau;

	/** the parameterized drift velocity in this section of the TPC (mm/us)
	 */
	double _drift_velocity;

	/**
	 * the parameterized longitudinal diffusion in this section of the TPC (um/sqrt(cm))
	 */
	double _longitudinal_diffusion;

	/**
	 * the parameterized transverse diffusion in this section of the TPC (um/sqrt(cm))
	 */
	double _transverse_diffusion;

	/**
	 * The outer radius of the TPC (mm)
	 */
	double _tpc_outer_radius;

	/**
	 * The inner radius of the TPC (mm)
	 */
	double _tpc_inner_radius;

	/**
	 * The absolute value of the z coordinate of the outermost position of this section of the TPC (mm)
	 */
	double _section_outer_z;

	/**
	 * The absolute value of the z coordinate of the innermost position of this section of the TPC (mm),
	 */
	double _section_inner_z;

	/** the ODE solving functions from the GSL
	 */
	static gsl_odeiv_step *stepper;

	static gsl_odeiv_control *step_control;

	static gsl_odeiv_evolve *evolver;
};

} // namespace marlintpc

#endif // ELECTRONCLOUDDRIFTERPROCESSOR_H
