
// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <EVENT/LCEvent.h>
#include <EVENT/TrackerData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerRawDataImpl.h>
#include <IMPL/LCFlagImpl.h>

#include <cassert>

// GEAR
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>

// Marlin
#include <marlin/Global.h>
#include <marlin/Exceptions.h>

// MarlinTPC
#include "GenericADCElectronicsParameters.h"
#include "Pedestal.h"
#include "SignalDigitisationProcessor.h"

using namespace std;
using namespace EVENT;
using namespace IMPL;
using namespace marlin;
using namespace tpcconddata;

namespace marlintpc {

// create one to register it with Marlin
SignalDigitisationProcessor aSignalDigitisationProcessor;


SignalDigitisationProcessor::SignalDigitisationProcessor():
	Processor("SignalDigitisationProcessor"),
	_input_collection_name("TPCTrackerData"),
	_output_collection_name("TPCTrackerRawData"),
	_tpc_pedestal_collection_name(Pedestal::getDefaultColName()),
	_adc_electronics_parameters_collection_name( GenericADCElectronicsParameters::getDefaultColName() ),
	_gain(1.),
	_bits_per_sample(10),
	_pedestal(0),
	_time_resolution(50.) {

	registerInputCollection(
		LCIO::TRACKERDATA,
		"InputTPCTrackerData",
		"Name of the input TrackerData collection containing ChargeSignals",
		_input_collection_name,
		string("TPCTrackerData") );

	registerOutputCollection(
		LCIO::TRACKERRAWDATA,
		"OutputTPCTrackerRawData",
		"Name of the output TrackerRawData collection.",
		_output_collection_name,
		string("TPCTrackerRawData") );

	registerOptionalParameter(
		"TPCPedestalCollectionName",
		"The name to use for the TPC pedestals collection.",
		_tpc_pedestal_collection_name,
		Pedestal::getDefaultColName());

	registerOptionalParameter(
		"TPCCADCElectronicsParametersCollectionName",
		"The name to use for the TPC adc electronics parameters collection.",
		_adc_electronics_parameters_collection_name,
		GenericADCElectronicsParameters::getDefaultColName());

	registerProcessorParameter(
		"Gain",
		"The amplification of this electronic digitiser.",
		_gain,
		1.);

	registerOptionalParameter(
		"BitsPerSample",
		"The number of bits put into each electronic sample.",
		_bits_per_sample,
		10);

	registerOptionalParameter(
		"Pedestal",
		"The pedestal value for all channels.",
		_pedestal,
		0);

	registerOptionalParameter(
		"TimeResolution",
		"The width of a signal sample (1/Hz) in ns.",
		_time_resolution,
		50.);
}


SignalDigitisationProcessor::~SignalDigitisationProcessor() {}


void SignalDigitisationProcessor::processRunHeader( LCRunHeader* run) {

	// store the processor parameters in the event
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {

        if ( ! i->second->isOptional() || i->second->valueSet() ) {

            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }
}


void SignalDigitisationProcessor::processEvent( LCEvent * evt ) {

	// grab the collection of tracker data objects from the simulation
	LCCollection *trackerdata_collection = NULL;
	try {

		trackerdata_collection = evt->getCollection( _input_collection_name );
	}
	catch (DataNotAvailableException &e) {

		// the collection of tracker data with that name wasn't in the LCIO event file, skip simulating this event
		throw SkipEventException( this );
	}

	streamlog_out( MESSAGE0 ) << "Digitising " << trackerdata_collection->getNumberOfElements() << " signals\n";

	// get the width of a time bin in ns
	if( !parameterSet("TimeResolution") ) {

		_time_resolution = 1.E09 / Global::GEAR->getTPCParameters().getReadoutFrequency();
	}

	const int mask = (1 << _bits_per_sample) - 1;
	
	// create a collection to store the TrackerRawData
	LCCollectionVec *trackerrawdata_collection = new LCCollectionVec( LCIO::TRACKERRAWDATA );
	trackerrawdata_collection->setTransient(false);
	trackerrawdata_collection->setSubset(false);

	// set the flags that cellID1 should be stored
	lcio::LCFlagImpl trkFlag(0) ;
	trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
	trackerrawdata_collection->setFlag( trkFlag.getFlag() );

	// create a collection to store the pedestals
	LCCollectionVec *pedestal_collection = new LCCollectionVec( LCIO::LCGENERICOBJECT );
	pedestal_collection->setTransient(false);
	pedestal_collection->setSubset(false);

	// now loop through the trackerdata objects digitising each one
	for( int trackerdata_number = 0; trackerdata_number < trackerdata_collection->getNumberOfElements(); trackerdata_number++ ) {

		bool keep = false;

		TrackerData *trackerdata = dynamic_cast<TrackerData*>( trackerdata_collection->getElementAt( trackerdata_number ) );
		assert(trackerdata);

		TrackerRawDataImpl *trackerrawdata = new TrackerRawDataImpl();

		trackerrawdata->setCellID0( trackerdata->getCellID0() );
		trackerrawdata->setCellID1( trackerdata->getCellID1() );
		trackerrawdata->setTime( (int) (trackerdata->getTime() / _time_resolution) );

		for(EVENT::FloatVec::const_iterator bin = trackerdata->getChargeValues().begin(); bin != trackerdata->getChargeValues().end(); bin++) {

			int adcValue_unmasked = (int)( (*bin * _gain) + _pedestal);

			int adcValue = (adcValue_unmasked > mask)? mask: adcValue_unmasked;

			if(adcValue > _pedestal) {

				keep = true;
			}

			trackerrawdata->adcValues().push_back(adcValue);
		}

		if( keep ) {

			trackerrawdata_collection->addElement( trackerrawdata );

			// create a pedestal for this channel/readout group
			Pedestal *pedestal = new Pedestal( trackerrawdata->getCellID0(), trackerrawdata->getCellID1(), _pedestal, 1. );
			pedestal_collection->addElement( pedestal );
		}
		else {

			delete trackerrawdata;
		}
	}

	// the electronic parameters don't change on a per-event basis in this simulation so just add them to the first event
	if( isFirstEvent() ) {

		add_electronic_parameters( evt );
	}

	// now add the tracker raw data and pedestal collections to the event so other processors may access it
	try {

		evt->addCollection( trackerrawdata_collection, _output_collection_name );
		evt->addCollection( pedestal_collection, _tpc_pedestal_collection_name );
	}
	catch( ... ) {

		throw SkipEventException( this );
	}
	
}


void SignalDigitisationProcessor::init() {

    // usually a good idea to
    printParameters() ;
}


void SignalDigitisationProcessor::check( LCEvent * evt ) {

}


void SignalDigitisationProcessor::end() { 

}

void SignalDigitisationProcessor::add_electronic_parameters( LCEvent *evt ) {

	// add the ADC electronics parameters to the event
	// create the adc electronics parameters. only one readout group (probably) so only make one object
	//  - polarity is positive
	//  - max ADC value is (1 << _bits_per_sample) - 1
	//  - the frequency is the inverse of the bin width after scaling in MHz
	const int max_adc = (1 << _bits_per_sample) - 1;

	GenericADCElectronicsParameters *adc_electronics_parameters = new GenericADCElectronicsParameters( 0, 1, max_adc, 1.E03 / _time_resolution, false, false );

	try {
		LCCollection *adc_electronics_parameters_collection = evt->getCollection( _adc_electronics_parameters_collection_name );
		adc_electronics_parameters_collection->addElement( adc_electronics_parameters );
	}
	catch (DataNotAvailableException &e) {

		LCCollectionVec *adc_electronics_parameters_collection = new LCCollectionVec( LCIO::LCGENERICOBJECT );
		adc_electronics_parameters_collection->setTransient(false);
		adc_electronics_parameters_collection->setSubset(false);
		adc_electronics_parameters_collection->addElement( adc_electronics_parameters );
		evt->addCollection( adc_electronics_parameters_collection, _adc_electronics_parameters_collection_name );
	}

}

}
