/*
 * SORBlock.cpp
 *
 *  Created on: Jul 21, 2011
 *      Author: caiazza
 */

#include "SORBlock.hh"
#include "AltroBlocksLib.h"

namespace A2Lconverter
{
uint32_t SORBlock::runNumber = 0;

SORBlock::SORBlock(std::ifstream * const inputFile)
        : AltroBlock(inputFile)
{

}

SORBlock::SORBlock(std::ifstream* const inputFile, int start_address, uint32_t length)
        : AltroBlock(inputFile, start_address, length)
{

}

SORBlock::SORBlock(const AltroBlock & theMotherBlock)
        : AltroBlock(theMotherBlock)
{

    if (this->CheckIntegrity())
    {
        id = AltroBlock::BLOCK_SOR;
    }
    else
    {
        throw AltroBlockException(
                "The Altro Block copied in the SOR block doesn't appear to be a valid SOR block");
    }
}

SORBlock::SORBlock(const std::vector<uint32_t> & refContent)
        : AltroBlock(refContent)
{

}

void
SORBlock::Decode()
{
    try
    {
        AltroBlock::Decode();
        if (content[0] != content[1] + 1)
            throw AltroBlockException("The total length doesn't match the header length +1");
        if (content[2] != AltroBlock::BLOCK_SOR)
            throw AltroBlockException("The block identification doesn't match with BLOCK_SOR");
        id = AltroBlock::BlockIds(content[2]);

        daqversion = (content[3] & 0x0000ffff);

        dataformat = (content[3] & 0xffff0000) >> 16;
        if (daqversion == 420 && dataformat == 0)
        {
            std::cout
                    << "WARNING: Because of a known bug in some run acquired with firmware version 4.2"
                    << " the data format was wrongly set to 0 even when the data were encoded with data format 1."
                    << " The data format is set to the correct value during the conversion."
                    << std::endl << std::endl;
            dataformat = 1;
        }
        if (daqversion > 450)
        {
            throw AltroBlockException(
                    "This converter is not compatible with firmware version superior to 4.5");
        }

        runNumber = content[4];

        int day = content[5] & 0x000000ff;
        int month = (content[5] & 0x0000ff00) >> 8;
        int year = (content[5] & 0xffff0000) >> 16;
        int second = content[6] & 0x000000ff;
        int minute = (content[6] & 0x0000ff00) >> 8;
        int hour = (content[6] & 0xffff0000) >> 16;
        timestamp = UTIL::LCTime(year, month, day, hour, minute, second);

        PCA16DAC = (content[7] & 0x0000ffff);

        registers.preampEnable = (content[7] & 0x00010000) >> 16;

        registers.gain = SORBlock::PCA16Registers::gainSettings((content[7] & 0x00060000) >> 17);

        registers.shaper = SORBlock::PCA16Registers::shaperSettings(
                (content[7] & 0x00380000) >> 19);

        registers.shutdown = (content[7] & 0x00400000) >> 22;

        registers.polarity = (content[7] & 0x00800000) >> 23;

        runSets.zeroSuppression = (content[8] & 0x00000001);

        runSets.pedestalSubtraction = (content[8] & 0x00000002) >> 1;

        if (daqversion <= 450)    //These other fields are only set from firmware version 4.6, so now everything will be set to -1
        {
            runSets.overThreshold = -1;
            runSets.postsamples = -1;
            runSets.presamples = -1;
            runSets.threshold = -1;
            runSets.pretrigger = -1;
        }
        else
        {
            throw AltroBlockException(
                    "This converter is not compatible with firmware version superior to 4.5");
        }
    }
    catch (std::ifstream::failure &f)
    {
        std::ostringstream err;
        err
                << "ERROR: Exception caught while decoding the SOR block. The file ended before it was supposed to"
                << std::endl;
        err << f.what() << std::endl;
        std::string tmp = err.str();
        throw AltroBlockException(tmp);
    }
    catch (std::exception &e)
    {
        std::ostringstream err;
        err << "ERROR: Exception caught while decoding the SOR block. " << std::endl;
        err << e.what() << std::endl;
        std::string tmp = err.str();
        throw AltroBlockException(tmp);
    }

}

void
SORBlock::Write2LCIO(IO::LCWriter* lcwriter) const
{
    std::map<PCA16Registers::gainSettings, std::string> PCA16Gain2String;
    PCA16Gain2String[SORBlock::PCA16Registers::MV12] = "12 mV/fC";
    PCA16Gain2String[SORBlock::PCA16Registers::MV15] = "15 mV/fC";
    PCA16Gain2String[SORBlock::PCA16Registers::MV19] = "19 mV/fC";
    PCA16Gain2String[SORBlock::PCA16Registers::MV27] = "27 mV/fC";

    std::map<SORBlock::PCA16Registers::shaperSettings, std::string> PCA16Shape2String;
    PCA16Shape2String[SORBlock::PCA16Registers::MS30] = "30 ns";
    PCA16Shape2String[SORBlock::PCA16Registers::MS60] = "60 ns";
    PCA16Shape2String[SORBlock::PCA16Registers::MS90] = "90 ns";
    PCA16Shape2String[SORBlock::PCA16Registers::MS120] = "120 ns";

    IMPL::LCRunHeaderImpl* runHeader = new IMPL::LCRunHeaderImpl();
    runHeader->setRunNumber(runNumber);
    runHeader->setDetectorName("LPTPC");
    runHeader->setDescription("Test beam run");

    std::ostringstream oss;
    oss << VERSION_MAJOR << '.' << VERSION_MINOR << '.' << VERSION_PATCH;
    runHeader->parameters().setValue("Altro Converter Version", oss.str());
    runHeader->parameters().setValue("RUN Date and time", timestamp.getDateString());
    runHeader->parameters().setValue("DAQ Version", daqversion);
    runHeader->parameters().setValue("ALTRO Data Format", dataformat);
    runHeader->parameters().setValue("PCA16DAC", PCA16DAC);
    runHeader->parameters().setValue("preampEnable", registers.preampEnable);
    runHeader->parameters().setValue("Gain", PCA16Gain2String[registers.gain]);
    runHeader->parameters().setValue("Shaping Time", PCA16Shape2String[registers.shaper]);
    runHeader->parameters().setValue("polarity", registers.polarity);
    runHeader->parameters().setValue("shutdown", registers.shutdown);
    runHeader->parameters().setValue("Pedestal Subtraction", runSets.pedestalSubtraction);
    runHeader->parameters().setValue("Zero Suppression", runSets.zeroSuppression);
    runHeader->parameters().setValue("EUDAQ Run Number", static_cast<int>(EUDAQRunNumber));

    lcwriter->writeRunHeader(runHeader);

    delete runHeader;
}

std::string
SORBlock::PrintSummary() const
{
    std::map<PCA16Registers::gainSettings, std::string> PCA16Gain2String;
    PCA16Gain2String[SORBlock::PCA16Registers::MV12] = "12 mV/fC";
    PCA16Gain2String[SORBlock::PCA16Registers::MV15] = "15 mV/fC";
    PCA16Gain2String[SORBlock::PCA16Registers::MV19] = "19 mV/fC";
    PCA16Gain2String[SORBlock::PCA16Registers::MV27] = "27 mV/fC";

    std::map<SORBlock::PCA16Registers::shaperSettings, std::string> PCA16Shape2String;
    PCA16Shape2String[SORBlock::PCA16Registers::MS30] = "30 ns";
    PCA16Shape2String[SORBlock::PCA16Registers::MS60] = "60 ns";
    PCA16Shape2String[SORBlock::PCA16Registers::MS90] = "90 ns";
    PCA16Shape2String[SORBlock::PCA16Registers::MS120] = "120 ns";

    std::ostringstream oss;
    oss << std::dec << "Run parameters summary" << std::endl << std::endl;
    oss << "Run Number " << runNumber << std::endl;
    oss << "Run start: " << timestamp.getDateString() << std::endl << std::endl;
    oss << "Decoding ALTRO Raw Data acquired with DAQ firmware version " << daqversion << std::endl;
    oss << "ALTRO raw data encoded with data format " << dataformat << std::endl << std::endl;
    oss << "PCA16 decay time set to " << PCA16DAC << std::endl;
    oss << "The internal shift registers of the PCA16 were set to the following values: "
            << std::endl;
    oss << "The preamp mode was set to " << (registers.preampEnable ? "enabled" : "disabled")
            << std::endl;
    oss << "The polarity of the amplifiers was set to "
            << (registers.polarity ? "enabled" : "disabled") << std::endl;
    oss << "The gain of the ADC was set to " << PCA16Gain2String[registers.gain] << std::endl;
    oss << "The shaping time of the ADC was set to " << PCA16Shape2String[registers.shaper]
            << std::endl;
    oss << "Pedestal subtraction " << (runSets.pedestalSubtraction ? "enabled" : "disabled")
            << std::endl;
    oss << "Zero Suppression " << (runSets.zeroSuppression ? "enabled" : "disabled") << std::endl
            << std::endl;

    std::string summary = oss.str();

    return summary;
}

bool
SORBlock::CheckIntegrity() const
{
    if (AltroBlock::CheckIntegrity())
    {
        if (content.size() == 0)
        {
            inputFile->seekg(2 * WORD32, std::ios::cur);
            return (SORBlock::ReadWord32() == AltroBlock::BLOCK_SOR);
        }
        else
        {
            return (content[2] == AltroBlock::BLOCK_SOR);
        }
    }
    return false;
}

const uint32_t
SORBlock::GetRunNumber()
{
    return runNumber;
}

uint16_t
SORBlock::GetPCA16DAC() const
{
    return PCA16DAC;
}

}
