#include "TimePixDistrDelayCorrectionMapReaderProcessor.h"
#include "TimePixPixelMode.h"

#include "marlin/ProcessorMgr.h"

#include "IO/LCWriter.h"
#include "IMPL/LCEventImpl.h"
#include "IMPL/LCRunHeaderImpl.h"
#include "IMPL/TrackerDataImpl.h"
#include "IMPL/LCCollectionVec.h"
#include <IMPL/LCFlagImpl.h>

//#include "UTIL/LCStdHepRdr.h"
#include "UTIL/LCTOOLS.h"

using namespace lcio;
using namespace tpcconddata;

namespace marlin{




  TimePixDistrDelayCorrectionMapReaderProcessor aTimePixDistrDelayCorrectionMapReaderProcessor ;
  
  
  TimePixDistrDelayCorrectionMapReaderProcessor::TimePixDistrDelayCorrectionMapReaderProcessor() : DataSourceProcessor("TimePixDistrDelayCorrectionMapReaderProcessor") {
    
    _description = std::string("Reads Calibration data files as input and creates a LCIO event with LCGenericObject Map.")
      + std::string(" Make sure to not specify any LCIOInputFiles in the steering in order to read TimePix TimeWalk Correction Map.") ;

  
  registerOutputCollection( LCIO::LCGENERICOBJECT,
			    "OutputCollectionName" , 
			    "Name of the output collection"  ,
			    _outputColName,
			    std::string( "TimePixCalibrationMap" ) ) ;
  
  registerProcessorParameter( "TimeWalkCorrectionMapFileName" , 
			      "input file with TimeWalkCorrectionMap"  ,
			      _TimeWalkCorrectionMapFileName,
			      std::string( "TimePixDistrDelayCorrectionMap.txt" ) ) ;

  registerProcessorParameter( "DetectorName" , 
			      "Name of the detector"  ,
			      _detName ,
			      std::string( "TPC" ) );
  
  registerProcessorParameter( "Description" , 
			      "Description of data"  ,
			      _dataDescription ,
			      std::string( "Calibration data" ) ) ;
  
  registerProcessorParameter( "SubDetectorName" , 
			      "Name of the subdetector"  ,
			      _subDetName ,
			      std::string( "TimePix" ) );

  registerProcessorParameter( "ChipIDs" , 
			      "A list of Chip IDs for which a map is created"  ,
			      _chipIDs,
			      _chipIDs ) ; 
  
  registerProcessorParameter( "NRow" , 
			      "Number of rows on the chip "  ,
			      _nRow,
			      256 ) ; 
  
  registerProcessorParameter( "NColumn" , 
			      "Number of columns on the chip "  ,
			      _nColumn,
			      256 ) ; 

  registerProcessorParameter( "SetOutputTransient" , 
			      "if not 0 the output collection is set transient"  ,
			      _outputIsTransient,
			      false );

  registerProcessorParameter( "TimePixChipID" , 
			      "Identification name of chip"  ,
			      _timePixChipID ,
			      std::string( "H09 - W0009" ) );
  
  }
  
  TimePixDistrDelayCorrectionMapReaderProcessor*  TimePixDistrDelayCorrectionMapReaderProcessor::newProcessor() { 
    return new TimePixDistrDelayCorrectionMapReaderProcessor ;
  }
  

  void TimePixDistrDelayCorrectionMapReaderProcessor::init() {    
    printParameters() ;    
    
    _runNr=0;
    _evtNr=0;

    //read in .dat files  
    try { _infile.open( _TimeWalkCorrectionMapFileName.c_str() ); }
    catch (IOException& e ){
      streamlog_out(WARNING0)<< e.what() <<std::endl;
      }
    if ( !_infile.good() ) {
      _infile.close();
      streamlog_out(ERROR) << "***********************************" << std::endl;
      streamlog_out(ERROR) << " *** " << _TimeWalkCorrectionMapFileName.c_str() <<" is not good() *** "<<std::endl;
      streamlog_out(ERROR) << "***********************************" << std::endl;     
      
    }

  }


  void TimePixDistrDelayCorrectionMapReaderProcessor::readDataSource( int numEvents ) {
 
 
    streamlog_out(MESSAGE1)<<"readDataSource"<<std::endl;

    if ( _infile.good() ){      
      
      // create new lcio event
      LCEventImpl*  evt = new LCEventImpl() ;
      evt->setRunNumber( _runNr ) ;
      evt->setEventNumber( _evtNr ) ;
      evt->setDetectorName( _detName ) ;
      
      
      // want to store a collection of LCGenericObject
      LCCollectionVec* calibDataMap = new LCCollectionVec( LCIO::LCGENERICOBJECT )  ;

      // set the flags that cellID1 should be stored
      lcio::LCFlagImpl trkFlag(0) ;
      trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
      calibDataMap->setFlag( trkFlag.getFlag() );

      // set the transient flag as defined by the processor parameter
      calibDataMap->setTransient( _outputIsTransient != 0 );
      
      // number of pixels in the map
      int nCells = (_nRow*_nColumn)  ; 

      // read in the calibration value for each pixel
      int data;
      for (std::vector<int>::iterator chipIter = _chipIDs.begin(); chipIter < _chipIDs.end(); chipIter++)    {

	for ( int cellid = 0; cellid < nCells; cellid++ ){     //loop over pixels	  

	  _infile >> data; 
	  if ( !_infile.good() ) break;
	  
	  TimePixPixelMode* mode = new TimePixPixelMode( *chipIter, cellid, 3, 1, data);
	  
	  // add the Data to the collection
	  calibDataMap->addElement(mode );
	}
      }

      // chip ID name
      calibDataMap->parameters().setValue( _timePixChipID, _evtNr );

      // add the Data collection to the event
      evt->addCollection( calibDataMap , _outputColName ) ;

	
      // create RunHeader
      streamlog_out(DEBUG0)<<"create RunHeader"<<std::endl;
      
      LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ; 
      
      runHdr->setRunNumber( _runNr ) ;
      runHdr->setDetectorName( _detName ) ;
      runHdr->setDescription( _dataDescription ) ;
      runHdr->addActiveSubdetector( _subDetName ) ;
      
      ProcessorMgr::instance()->processRunHeader( runHdr ) ;
      delete runHdr;
      
      
      // dump the event to the screen 
      streamlog_out(DEBUG0)<<"in readDataSource"<<std::endl;
      
      if( streamlog::out.write< streamlog::DEBUG2 >() )
	LCTOOLS::dumpEvent( evt ) ;
      
      ProcessorMgr::instance()->processEvent( evt ) ;     
      
      
      // we created the event so we need to delete it ...
      delete evt ;    

      streamlog_out(DEBUG0)<<"\nend of readDataSource"<<std::endl;
    }

  }    


  void TimePixDistrDelayCorrectionMapReaderProcessor::end() {
    // nothing done here
  }


}
