#include "TimePixModeMapCreaterProcessor.h"
#include "TimePixPixelMode.h"

#include "marlin/ProcessorMgr.h"

#include "IO/LCWriter.h"
#include "IMPL/LCEventImpl.h"
#include "IMPL/LCRunHeaderImpl.h"
#include "IMPL/TrackerRawDataImpl.h"
#include "IMPL/LCCollectionVec.h"

//#include "UTIL/LCStdHepRdr.h"
#include "UTIL/LCTOOLS.h"

using namespace lcio;
using namespace tpcconddata;

namespace marlin{



  TimePixModeMapCreaterProcessor aTimePixModeMapCreaterProcessor ;
  
  
    TimePixModeMapCreaterProcessor::TimePixModeMapCreaterProcessor() 
	: DataSourceProcessor("TimePixModeMapCreaterProcessor") ,
	  _chipIDs(1)
    {
    
	_chipIDs[0]=0;
    _description = "Reads TimePix raw data files as input and creates LCIO events with TrackerRawData collections."
      " Make sure to not specify any LCIOInputFiles in the steering in order to read TimePix raw data files." ;

    registerOutputCollection( LCIO::LCGENERICOBJECT,
			     "ModeMapCollectionName" , 
			     "Name of the output mode map collection"  ,
			      _modeMapColName,
			     std::string("TimePixModeMap") ) ;
    
   registerProcessorParameter( "DetectorName" , 
			       "Name of the detector"  ,
			       _detName ,
			       std::string("TPC" ) );
   
   registerProcessorParameter( "Description" , 
			       "Description of data"  ,
				_description ,
			       std::string("no description given" ) ) ;
   
   registerProcessorParameter( "SubDetectorName" , 
			       "Name of the subdetector"  ,
				_subDetName ,
			       std::string("TimePix" ) );
   
   registerProcessorParameter( "Mode" , 
			       "Operation mode of the chip: TOT, TIME, MM1 (mixed mode with CellID0 =  TIME), MM2 (mixed mode with CellID0 =  TOT)"  ,
			       _mode,
			       std::string("TOT" ) ) ;
   
   registerProcessorParameter( "NRow" , 
			       "Number of rows on the chip "  ,
			       _nRow,
			       (int) 256 ) ; 
   
    registerProcessorParameter( "NColumn" , 
				"Number of columns on the chip "  ,
				_nColumn,
				(int) 256 ) ; 
    
    registerProcessorParameter( "ChipIDs" , 
				"A list of Chip IDs for which a map is created"  ,
				_chipIDs,
				_chipIDs ) ; 
    
    
    registerProcessorParameter( "SetOutputTransient" , 
				"if not 0 the output collection is set transient"  ,
				_outputIsTransient,
				0);
    
  }
  
  TimePixModeMapCreaterProcessor*  TimePixModeMapCreaterProcessor::newProcessor() { 
    return new TimePixModeMapCreaterProcessor ;
  }
  
  void TimePixModeMapCreaterProcessor::init() {    
    printParameters() ;    
    
    _isFirstEvent = true; 
    _runNr=0;
    _evtNr=0;


  }


  void TimePixModeMapCreaterProcessor::readDataSource( int numEvents ) {
 
 
    streamlog_out(MESSAGE1)<<"readDataSource"<<std::endl;


      // create new lcio event
      LCEventImpl*  evt = new LCEventImpl() ;
      evt->setRunNumber(  _runNr   ) ;
      evt->setEventNumber( _evtNr ) ;
      evt->setDetectorName( _detName ) ;
      
      
	
      // add the conditions data to the first event  
      LCCollectionVec* modeMap = new LCCollectionVec( LCIO::LCGENERICOBJECT );
      
      //create the modeMap
      if ( _mode == "TOT" ) createTOTMap( modeMap );
      if ( _mode == "TIME") createTIMEPIXMap( modeMap );
      if ( _mode == "MM1" ) createMixedMode1Map( modeMap );
      if ( _mode == "MM2" ) createMixedMode2Map( modeMap );
      
      evt->addCollection( modeMap, _modeMapColName );
      
      //create RunHeader
      streamlog_out(DEBUG0)<<"create RunHeader"<<std::endl;
      
      LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ; 
      
      runHdr->setRunNumber( _runNr) ;
      runHdr->setDetectorName( _detName ) ;
      runHdr->setDescription( _description ) ;
      runHdr->addActiveSubdetector( _subDetName ) ;
      
      ProcessorMgr::instance()->processRunHeader( runHdr ) ;
      delete runHdr;
      

    
      // dump the event to the screen 
      streamlog_out(DEBUG0)<<"in readDataSource"<<std::endl;

      if( streamlog::out.write< streamlog::DEBUG2 >() )
	  LCTOOLS::dumpEvent( evt ) ;

      ProcessorMgr::instance()->processEvent( evt ) ;     
	
      
      // ------------ IMPORTANT ------------- !
      // we created the event so we need to delete it ...
      delete evt ;
      // -------------------------------------
      

    streamlog_out(DEBUG0)<<"\nend of readDataSource"<<std::endl;

  }
  


  void TimePixModeMapCreaterProcessor::end() {



  }

  void TimePixModeMapCreaterProcessor::createMixedMode1Map( LCCollectionVec* modeMap ) 
  {
    //creates map of chip operated in MixedMode starting with TIMEPIX
    int ncells = _nRow*_nColumn;
    
    for (std::vector<int>::iterator chipIter = _chipIDs.begin();
	 chipIter < _chipIDs.end(); chipIter++)
    {
	for ( int cellid = 0; cellid < ncells; cellid++ )
	{    
	    int x = cellid%_nColumn;
	    int y = static_cast<int>((cellid-x)/_nColumn);
	    
	    if ( ( x%2 == 0 && y%2 ==0 ) || (x%2 != 0 && y%2 != 0 ) ) {
		
		modeMap->addElement( new TimePixPixelMode( *chipIter, 
							   cellid, 
							   TimePixPixelMode::TIMEPIX, 
							   TimePixPixelMode::INTACT) );

	    }
	    else {
		
		modeMap->addElement( new TimePixPixelMode( *chipIter, 
							   cellid, 
							   TimePixPixelMode::TOT, 
							   TimePixPixelMode::INTACT) );  
		
	
	    }
	    
	}
    }
    
  }

  void TimePixModeMapCreaterProcessor::createMixedMode2Map( LCCollectionVec* modeMap ) 
  {
    //creates map of chip operated in MixedMode starting with TOT
    int ncells = _nRow*_nColumn;
    
    for (std::vector<int>::iterator chipIter = _chipIDs.begin();
	 chipIter < _chipIDs.end(); chipIter++)
    {
	for ( int cellid = 0; cellid < ncells; cellid++ )
	{	    
	    int x = cellid%_nColumn;
	    int y = static_cast<int>((cellid-x)/_nColumn);
	    
	    if ( ( x%2 == 0 && y%2 ==0 ) || (x%2 != 0 && y%2 != 0 ) ) {
		
		modeMap->addElement( new TimePixPixelMode( *chipIter, 
							   cellid, 
							   TimePixPixelMode::TOT, 
							   TimePixPixelMode::INTACT) );  
		
	    }
	    else {
		
		modeMap->addElement( new TimePixPixelMode( *chipIter, 
							   cellid, 
							   TimePixPixelMode::TIMEPIX, 
							   TimePixPixelMode::INTACT) );  
	    }
	} // for cellID
    }// for chipID
    
  }
  
  void TimePixModeMapCreaterProcessor::createTOTMap( LCCollectionVec* modeMap ) 
  {
    //creates map of chip operated in TOT mode alone
    int ncells = _nRow*_nColumn;
    
    for (std::vector<int>::iterator chipIter = _chipIDs.begin();
	 chipIter < _chipIDs.end(); chipIter++)
    {
	for ( int cellid = 0; cellid < ncells; cellid++ )
	{	    
	    modeMap->addElement( new TimePixPixelMode( *chipIter, 
						       cellid, 
						       TimePixPixelMode::TOT, 
						       TimePixPixelMode::INTACT) );  	    
	}
	
    }    
  }
    
  void TimePixModeMapCreaterProcessor::createTIMEPIXMap( LCCollectionVec* modeMap ) 
  {
  
    //creates map of chip operated in TIMEPIX mode alone
    int ncells = _nRow*_nColumn;
    
    for (std::vector<int>::iterator chipIter = _chipIDs.begin();
	 chipIter < _chipIDs.end(); chipIter++)
    {
	for ( int cellid = 0; cellid < ncells; cellid++ )
	{    
	    modeMap->addElement( new TimePixPixelMode( *chipIter, 
						       cellid, 
						       TimePixPixelMode::TIMEPIX, 
						       TimePixPixelMode::INTACT) );  
	    
	}
    }
    
  }

}
