#ifndef FieldSettings_h
#define FieldSettings_h 1

//C++
#include "iostream"
#include "string"

//LCIO
#include "lcio.h"
#include "EVENT/LCGenericObject.h"
#include "IMPL/LCGenericObjectImpl.h"

namespace tpcconddata{

class FieldSetting ;

/** class to store the field setting of a TPC with GEM-based amplification system
 */
class FieldSetting : public EVENT::LCGenericObject {

public:

    FieldSetting( int NGEMs );

    FieldSetting(EVENT::LCObject* Obj);
    
    ~FieldSetting();

    /// return the Number of GEMs
    int   getNGEMs() const;

    /** return the status of the GEM System
     *  0: off
     *  1: running normally
     *  2: ramping up
     *  3: ramping down
     *  4: tripped
     *  5: other error
     */
    int   getGEMStatus() const;

    /** return the status of the DriftField System
     *  0: off
     *  1: running normally
     *  2: ramping up
     *  3: ramping down
     */
    int getDriftFieldStatus() const;

    /** return the status of the Magnet
     *  0: off
     *  1: running normally
     *  2: ramping up
     *  3: ramping down
     */
    int getMagneticFieldStatus() const;

    /** return the voltage of the given number
     *  GEM 0 is the nearest the drift region
     */
    float getGEMVoltage( int GEMID ) const;
    
    /** return the field (in V/mm) between the GEM N and N+1
     */
    float getTransferField( int GEMID ) const;

    /** return the field (in V/mm) between the last GEM and the Pad-Readout-Plane
     */
    float getInductionField() const;

    /** return voltage (in V) of the cathode of the field cage
     */
    float getCathodeVoltage() const;

    /** return voltage (in V) of the anode of the field cage
     *  (e.g. the shield at the position of the first GEM)
     */
    float getAnodeVoltage() const;

    /** return the driftfield (in V/mm) if the length of the driftregion (in mm) is given
     */
    float getDriftField( float maxDriftLegth ) const;

    /// return the magnetic field (in T)
    float getMagneticField() const;

    ///set the status of a GEM (see getGEMStatus)
    void setGEMStatus( int Value );

    ///set the status of a drift field (see getDriftFieldStatus)
    void setDriftFieldStatus( int Value );

    ///set the status of a magnet (see getMagneticFieldStatus)
    void setMagneticFieldStatus( int Value );

    /** set the voltage of the given number
     *  GEM 0 is the nearest the drift region
     */
    void setGEMVoltage( int GEMID, float Value );
    
    /** set the field (in V/mm) between the GEM N and N+1
     */
    void setTransferField( int GEMID, float Value );

    /** set the field (in V/mm) between the last GEM and the Pad-Readout-Plane
     */
    void setInductionField( float Value );

    /** set voltage (in V) of the cathode of the field cage
     */
    void setCathodeVoltage( float Value );

    /** set voltage (in V) of the anode of the field cage
     *  (e.g. the shield at the position of the first GEM)
     */
    void setAnodeVoltage( float Value );

    /// return the magnetic field (in T)
    void setMagneticField( float Value );

    IMPL::LCGenericObjectImpl* obj();
    
    virtual int id();
    
    static std::string getRevision();
    static std::string getDefaultColName();

    void print ( std::ostream& os = std::cout ) const;
    
    // ---- need to implement LCGenericObject interface:
    
    int getNInt() const;   
    int getNFloat() const; 
    int getNDouble() const;
    
    int getIntVal(int index) const;
    float getFloatVal(int index) const;
    double getDoubleVal(int index) const;

    bool isFixedSize() const { return false; };

    virtual const std::string getTypeName() const{
	return std::string("FieldSetting");
    }
    
    virtual const std::string getDataDescription() const{
	return std::string("i:#GEMs,GEMStatus,DriftFieldStatus,MagnetFieldStatus")+
	    std::string("f:CathodeVoltage,AnodeVoltage,GEMVoltages,Transferfields,Inductionfield");	
    }
    // ---- end of LCGenericObject interface

protected:

    bool createObj;
    IMPL::LCGenericObjectImpl* myObj;

    /** set the Number of GEMs
     * is protected, numberof GEMs is to imported for consistence
     * can only be set by constructor
     */
    void  setNGEMs( int Value);
};

    std::ostream &operator<<(std::ostream &os, const FieldSetting &fs);
}
#endif
