#ifndef TimePixPixelMode_h
#define TimePixPixelMode_h 1

//C++
#include "iostream"
#include <string>
#include <sstream>
#include <ostream>

//LCIO
#include "lcio.h"
#include "UTIL/LCFixedObject.h"


#define TimePixPixelModeNINTVals 4 
#define TimePixPixelModeNFLOATVals 0
#define TimePixPixelModeNDOUBLEVals 0

namespace tpcconddata{

class TimePixPixelMode;

/** class to store the mapping of the pixel modes of the TimePix 
 *  a TimePixPixelMode contains three integers, 
 *  the ChipID, the PixelID and the PixelMode
 *  furthermore it contains a struct called PixelKey 
 *  of ChipID and PixelID as the key for the map
 *
 *  the possible modes are:
 *  MEDIPIX = 0
 *  TOT = 1
 *  ONEHIT = 2
 *  TIMEPIX = 3
 *  
 */

class TimePixPixelMode : public UTIL::LCFixedObject<TimePixPixelModeNINTVals,
			                            TimePixPixelModeNFLOATVals,
                                                    TimePixPixelModeNDOUBLEVals> {

public:

  TimePixPixelMode(int ChipID = 0, int PixelID = 0, int PixelMode = 0, int PixelStatus = 0, int PixelCalibration = 0 );

    TimePixPixelMode(EVENT::LCObject* obj) : 
    UTIL::LCFixedObject<TimePixPixelModeNINTVals,
                        TimePixPixelModeNFLOATVals,
                        TimePixPixelModeNDOUBLEVals>(obj) { } 
    
    virtual ~TimePixPixelMode();

    //PixelKey
    struct PixelKey{

      PixelKey( int _chipID, int _pixelID );

      int ChipID; ///< The chipID
      int PixelID;///< The pixelID

      bool operator()(PixelKey* pk1, PixelKey* pk2)
      {
	return(pk1->ChipID == pk2->ChipID ? 
	       pk1->PixelID < pk2->PixelID : pk1->ChipID < pk2->ChipID);
      }

      /** PixelKey1 < PixelKey2:
       *  if chipID1 < chipID2 or
       *  if chipID1 == chipID2 and pixelID1 < pixelID2.
       */ 
      bool operator<(PixelKey const pixelkey) const
        {
          return(ChipID == pixelkey.ChipID ? 
		 PixelID < pixelkey.PixelID : ChipID < pixelkey.ChipID);
        }

      /** PixelKey1 == PixelKey2:
       *  if chipID1 == chipID2 and pixelID1 == pixelID2.
       */ 
      bool operator==(PixelKey const pixelkey) const
        {
          return(ChipID == pixelkey.ChipID && PixelID == pixelkey.PixelID );
        }

      friend std::ostream& operator<<(std::ostream& out, 
				      const TimePixPixelMode::PixelKey& pixelkey)
	   // output
      {
	out << "TimePixPixelMode::PixelKey(" << pixelkey.ChipID << ", " 
	    << pixelkey.PixelID << ")";
	return out;
      }

    };

    /** Gets the chipID of the chip.
     */
    int getChipID() ;

    /** Gets the pixelID of the pixel
     *  count begins with 0.
     */
    int getPixelID() ;

    /** Gets the mode of the pixel
     *  four possibilities:
     *  MEDIPIX, TOT, TIMEPIX, ONEHIT.
     */
    int getPixelMode() ;

    /** Gets the status of the pixel
     *  three possibilities:
     *  INTACT (default), NOISY, DEAD.
     */
    int getPixelStatus() ;

    /** Gets the status word which is a combination of the pixel mode
     *  and the pixel status.
     */
    int getPixelStatusWord();

    /** Gets the pixel calibration value 
     *	to correct clock distribution 
     */
    int getPixelCalibration();


    /** Gets the pixelKey of the pixel.
     *  The PixelKey consists of 
     *  the chipID and the pixelID.
     */
    PixelKey getPixelKey() ;

    static std::string getRevision();

    // -------- need to implement abstract methods from LCGenericObject
    virtual const std::string getTypeName() const{
	return std::string("TimePixPixelMode");
    }
    
    virtual const std::string getDataDescription() const{
	return std::string("i:ChipID,i:PixelID,i:PixelStatusWord,i=PixelCalibration");	
    }

    /** Sets the mode of this pixel.
     *  From the given \c int only the bits describing the mode are taken
     *  to replace the mode bits in the status word.
     *  This means you can pass the Mode or a complete StatusWord as argument.
     */ 
    void setPixelMode( int Mode );

    /** Sets the status of this pixel.
     *  From the given \c int only the bits describing the status are taken
     *  to replace the status bits in the status word.
     *  This means you can pass the Status or a complete StatusWord as argument.
     */ 
    void setPixelStatus( int Status );

    /// Sets the status word of a pixel.
    void setPixelStatusWord(int StatusWord); 

    /// Sets the calibration correction value
    void setPixelCalibration(int Calibration);   

    
    static const int MEDIPIX = 0;
    static const int TOT = 1;
    static const int ONEHIT = 2;
    static const int TIMEPIX = 3;

    static const int INTACT = 0;
    static const int NOISY = 0x10000;
    static const int DEAD = 0x20000;


protected:

    void setChipID(int Chip);
    void setPixelID(int Pixel);
    //    void setPixelMode(int Mode);

   
};

}

#endif
