#include "lcio.h"

#include "IO/LCWriter.h"
#include "EVENT/LCIO.h"

#include "IMPL/LCEventImpl.h" 
#include "IMPL/LCRunHeaderImpl.h" 
#include "IMPL/LCCollectionVec.h"
#include "IMPL/LCTOOLS.h"
#include "IMPL/TrackerRawDataImpl.h"
#include "IMPL/LCFlagImpl.h"

#include "gearimpl/RectangularPadRowLayout.h"
#include "gearimpl/VersatileDiskRowLayout.h"
#include "gearimpl/Util.h"
#include "gearxml/GearXML.h"
#include "gear/GearMgr.h"
#include "gear/GEAR.h"
#include "gear/TPCModule.h"
#include "ADCChannelMapping.h"
#include "Pedestal.h"

#include <iostream>
#include <sstream>
#include <string>
#include <cstdio>
#include <typeinfo>
#include <CLHEP/Random/RandGauss.h>
#include <CLHEP/Random/RandFlat.h>

using namespace std ;
using namespace lcio ;
using namespace gear;
using namespace tpcconddata;

int lastHitRow;
int lastHitMod;

void printUsage()
{
  cout << "generateRawData gearfile.xml [outfile.slcio]" <<endl;
    cout <<" outfile name is optional, "
	 << "default is TPCPulses.slcio"<< endl;
}

// return true if hit is on pad plane, false otherwise
bool placeHit(float x, float y, float z, bool checkRow, TPCParameters const &tpcParams, LCCollectionVec *trackerRawDataVec)
{
          float padPitch;

          if(!tpcParams.isInsideModule(x,y))
          { 
             //cout << "Point " << x << " " << y << " is not in any module!" << endl; 
             return false;
          }

          const TPCModule &module=tpcParams.getNearestModule(x,y);

          int padNr;    // the pseudo hardware channel number, the same with padIndex:trival
	  int padIndex; // the gear pad index

          static const PadRowLayout2D& tpcPadLayout = module.getLocalPadLayout();
          static const VersatileDiskRowLayout* vLayout = dynamic_cast< const VersatileDiskRowLayout*> (&tpcPadLayout);

	  try
	  {
	    padIndex=module.getNearestPad(x,y);
	    padNr=padIndex;
          }
	  catch(exception& e)
	  {
	    return false;
	  }
	  
          if(checkRow)
          {
             if(lastHitRow !=module.getRowNumber(padIndex))
             {
                lastHitRow=module.getRowNumber(padIndex);
                lastHitMod=module.getModuleID();
                padPitch = vLayout->rows().at(lastHitRow).getPadPitch_mm();
             }
             else
                return false;
          }
          else
          {
             padPitch = vLayout->rows().at(lastHitRow).getPadPitch_mm();
          }

	  Vector2D padCenter1= module.getPadCenter(padIndex);

	  TrackerRawDataImpl* rawData = new TrackerRawDataImpl; 
	  rawData->setCellID0(padNr);
          rawData->setCellID1(module.getModuleID());

	  float charge1=fabs( (255. *(1.- fabs(y-padCenter1[1])/padPitch)) );
	  EVENT::ShortVec ADCValues;
	  ADCValues.push_back(static_cast<short>(charge1*0.1));
	  ADCValues.push_back(static_cast<short>(charge1*0.7));
	  ADCValues.push_back(static_cast<short>(charge1*0.2));
	  rawData->setADCValues(ADCValues);
	  rawData->setTime(static_cast<int>(z*module.getReadoutFrequency()/tpcParams.getDriftVelocity()));

	  // debug:
	  cout << "padIndex: " << padIndex 
	       << "\t x/y: "<< x<<"/"<<y
	       << "\t center: "<< padCenter1[0]<<"/"
	       << padCenter1[1] 
	       << "\t charge: "<< charge1
	       << endl;


	  // add the  rawData to the collection
	  trackerRawDataVec->push_back(rawData);

	  // in case of a central hit these is almost  no charge 
	  // on the neighbouring pad, skip it
	  //if (fabs(padCenter1[1]-y)<0.1)
	  //  return true;

	  // try to find neighbour which also gets charge
	  int neighbouringPadNr;
	  int neighbouringPadIndex;
	  try
	  {
	    if (y-padCenter1[1] < 0)
	    {
	      neighbouringPadIndex = module.getRightNeighbour(padIndex);
	    }
	    else
	    {
	      neighbouringPadIndex = module.getLeftNeighbour(padIndex);
	    }
	    neighbouringPadNr = neighbouringPadIndex;
	  }
	  catch(exception& e)
	  {
	    //no neighbour found, must be edge pad
	    return true;
	  }

	  // if there is a neighbour calculate charge and add it to collection
	  Vector2D padCenter2= module.getPadCenter(neighbouringPadIndex);
	  
	  rawData = new TrackerRawDataImpl; 
	  rawData->setCellID0(neighbouringPadNr);
          rawData->setCellID1(module.getModuleID());

	  float charge2=fabs((255. *(1.- fabs(y-padCenter2[1])/padPitch)));

	  EVENT::ShortVec ADCValues2;
	  ADCValues2.push_back(static_cast<short>(charge2*0.1));
	  ADCValues2.push_back(static_cast<short>(charge2*0.7));
	  ADCValues2.push_back(static_cast<short>(charge2*0.2));
	  rawData->setADCValues(ADCValues2);
	  rawData->setTime(static_cast<int>(z*module.getReadoutFrequency()/tpcParams.getDriftVelocity()));

	  // debug:
	  cout << "   nb: " << neighbouringPadNr
	       << "\t x/y: "<< x<<"/"<<y
	       << "\t center:" 
               << padCenter2[0] <<"/"
	       << padCenter2[1] 
               << "\t charge: " << charge2
               << "   pos="
	       << (charge1*padCenter1[1]+charge2*padCenter2[1])/(charge1+charge2)
	       << endl;

	  // add the  rawData to the collection
	  trackerRawDataVec->push_back(rawData);

	  return true;
}

int main(int argc, char **argv)
{

  if (argc!=2 && argc!=3 )
  {
    printUsage();
    return(1);
  }

  // read pad layout from gear
  GearXML gearXML( argv[1] ) ;
  GearMgr* gearMgr = gearXML.createGearMgr() ;
  TPCParameters const &tpcParams=gearMgr->getTPCParameters();

  // get outfile name
  string outFileName;
  if (argc==3)
    outFileName=argv[2];
  else
    outFileName="TPCRawData.slcio";

  // open outfile			     
  try{
    // create sio writer
    LCWriter* lcWrt = LCFactory::getInstance()->createLCWriter() ;
    
    try{  lcWrt->open( outFileName, LCIO::WRITE_NEW ) ;
    }
    catch( IOException& e ){    
      cout << e.what() << endl ;
      return 0 ;
    }
      
      // create runheader and write it
      // the information should come from a file
      // or be given interactively in a future version
      LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ; 
      int runNr=0;
      runHdr->setRunNumber( runNr) ;
      
      string detName("demo")  ;
      runHdr->setDetectorName( detName ) ;
      
      stringstream description ; 
      description << "demo tpc rawData along straight line";
      runHdr->setDescription( description.str()  ) ;
      
      string tpcName("demotpc") ;
      runHdr->addActiveSubdetector( tpcName ) ;
      
      lcWrt->writeRunHeader( runHdr ) ;
  
      // loop the events
      int maxEvents=1000;
      for (int eventNr=0; eventNr<maxEvents; eventNr++)
      {
        cout << "Event=" << eventNr << endl;
	LCEventImpl*  evt = new LCEventImpl();
	
	evt->setRunNumber(  runNr   ) ;
	evt->setEventNumber( eventNr ) ;
	evt->setDetectorName( detName ) ;
      
	LCCollectionVec* trackerRawDataVec = new LCCollectionVec( LCIO::TRACKERRAWDATA ) ;
   
        LCFlagImpl flag(0);
        flag.setBit( LCIO:: TRAWBIT_ID1);
        trackerRawDataVec->setFlag( flag.getFlag() );
       
        lastHitRow=-1;
        lastHitMod=-1;   

        float x,y,z; //a hit
        float k=0.05;//slope
        float a=-30.;//intercept

        const float X_MIN =1250.;
        const float X_MAX =2100.;
        const float X_STEP=2.;
        const float SIGMA =0.05;
        const float FAKE_HIT_REGION=150;

        x=X_MIN;
        y=a-k*x;
        z=300.;//a fixed value to z

        while(x<X_MAX)
        {
          //place a hit
          if( placeHit(x,y,z,true,tpcParams,trackerRawDataVec) )
          {
            float fakeHitY=CLHEP::RandFlat::shoot(-FAKE_HIT_REGION,FAKE_HIT_REGION);
            placeHit( x, y+fakeHitY, z, false, tpcParams, trackerRawDataVec);
          }

          x+=X_STEP;
          y=a-k*x+CLHEP::RandGauss::shoot(0.,SIGMA);
        }


	// add the rawData collection to the event
	evt->addCollection( trackerRawDataVec , "TPCRawData" ) ;

	// in the first event: write the mapping and the pedestals
	if (eventNr==0)
	{
	  // create a collections of generic objects
	  LCCollectionVec* mappingCol  = new LCCollectionVec( LCIO::LCGENERICOBJECT );
	  LCCollectionVec* pedestalCol = new LCCollectionVec( LCIO::LCGENERICOBJECT );
	  
	  //loop all pads and create a conditions objects for it

	  //outer loop: the rows
	  int nModule=tpcParams.getNModules();

          for(int m=0; m<nModule; m++)
          {
             const TPCModule *module=tpcParams.getModules().at(m);
             
	     for (int row = 0; row < module->getNRows(); row++)
	     {
	       // vector of the pad indices in this row
	       std::vector<int> padsInRowVector = module->getPadsInRow(row);
	    
	       // nnow loop all pads in row
	       for (vector<int>::iterator padIndex=padsInRowVector.begin(); padIndex < padsInRowVector.end(); padIndex++  )
	       {
	          int padNr=*padIndex;

                  cout << "PadIndex=" << padNr << endl;
	          mappingCol->push_back( new ADCChannelMapping(padNr, *padIndex, -1, module->getModuleID() , module->getModuleID() ) );

	          // create pedestal with 0 as value and width, data are already zero-suppressed
                  pedestalCol->push_back( new Pedestal(padNr,0)) ;
	       }
	     }
          }

	     // add the pe collection to the event
	     evt->addCollection( mappingCol , "TPCChannelMapping" ) ;
	     evt->addCollection( pedestalCol , "TPCPedestal" ) ;
        }

	// write the event to the file
	lcWrt->writeEvent( evt ) ;
	
	// dump the event to the screen 
	LCTOOLS::dumpEvent( evt ) ;

	// ------------ IMPORTANT ------------- !
	// we created the event so we need to delete it ...
	delete evt ;
	// -------------------------------------

      } // for eventNr

      delete runHdr ;
      
      lcWrt->close() ;
      
  } catch( std::exception& ex){
    
    cout << " an excpetion occured: " << endl ;
    cout << "  " << ex.what() << endl ;
    return 1 ;
  }
return 0;

}
