#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"

// lcio
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackImpl.h>
#include <lcio.h>

// clhep
#include <CLHEP/Vector/ThreeVector.h>

// default libraries
#include <iostream>
#include <cmath>

using namespace lcio;
using namespace marlintpc;
using namespace CLHEP;

int main()
{
  std::cout << "This programme creates a straight line and a helix and hits on the tracks." << std::endl;
  std::cout << "It passes the tracks to TrackFitterBase to calculate the residuals and prints them."<< std::endl;
  std::cout << "Up to a track length of 100 the residuals should be 0, -2.5 for track length between 100 and 135 and +2.5 above 135." << std::endl;
  std::cout << "To test all possible situations change the track parameters in the source code and rerun." << std::endl;
  std::cout << std::endl;

  TrackImpl *straightTrack = new TrackImpl;

  straightTrack->setTanLambda( -0.1 );
  straightTrack->setZ0( 375. );
  straightTrack->setPhi( M_PI*0.1 );
  straightTrack->setD0( -68. );

  // now the same for the curved track
  
  TrackImpl *curvedTrack = new TrackImpl;

  curvedTrack->setTanLambda( straightTrack->getTanLambda()  );
  curvedTrack->setZ0( straightTrack->getZ0() );
  curvedTrack->setPhi( straightTrack->getPhi() );
  curvedTrack->setD0( straightTrack->getD0() );
  curvedTrack->setOmega( 0.001 ); // 1 m radius
  

  // The point on the helix which is clostest to the origin (in the projection)
  Hep3Vector PCA( -sin( straightTrack->getPhi() )*straightTrack->getD0(),
		  cos( straightTrack->getPhi() )*straightTrack->getD0(),
		  straightTrack->getZ0() );

  Hep3Vector alongStraightLine( cos( straightTrack->getPhi() ),
				sin( straightTrack->getPhi() ),
				straightTrack->getTanLambda() );

  Hep3Vector PCAProjection = PCA; PCAProjection[2] = 0 ; // The PCA in the xy-projection

  std::cout << "PCAProjection " << PCAProjection << std::endl;

  Hep3Vector helixCentreProjection = PCAProjection.unit() * ( fabs(curvedTrack->getD0()) -  (  curvedTrack->getD0() < 0 ? -1. : 1. ) /curvedTrack->getOmega() );
  Hep3Vector radiusVector = PCAProjection.unit() * (  curvedTrack->getD0() < 0 ? -1. : 1. ) / curvedTrack->getOmega() ;
   std::cout << "radiusVector " << radiusVector << std::endl;

   double stepOnTrack=17;
   double nSteps=10;
 
  // add a few hits on the track (all residuals should be 0 )
  for ( double s = 0; s <= nSteps*stepOnTrack; s+=stepOnTrack )
  {
    Hep3Vector hitOnTrack = PCA + s* alongStraightLine;

    TrackerHitImpl  * newLcioHit = new TrackerHitImpl;
    
    double pos[3];
    pos[0]= hitOnTrack[0];
    pos[1]= hitOnTrack[1];
    pos[2]= hitOnTrack[2] + ( s > 100. ? (s>135 ? +1.2 : -1.2 ) : 0 );

    newLcioHit->setPosition( pos );
    
    straightTrack->addHit( newLcioHit );
  }

  std::cout << "Helix centre " << helixCentreProjection << std::endl;

  // ok, now the same for the helix
  for ( double s = 0; s <= nSteps*stepOnTrack; s+=stepOnTrack )
  {
    

    Hep3Vector hitOnTrack = helixCentreProjection + radiusVector // the xy component
      +  Hep3Vector(0.,0.,1.)*( curvedTrack->getZ0() + s*curvedTrack->getTanLambda() ) ; // the z component

    TrackerHitImpl  * newLcioHit = new TrackerHitImpl;
    
    double pos[3];
    pos[0]= hitOnTrack[0];
    pos[1]= hitOnTrack[1];
    pos[2]= hitOnTrack[2] + ( s > 100. ? (s>135 ? +1.2 : -1.2 ) : 0 );

    newLcioHit->setPosition( pos );
    
    curvedTrack->addHit( newLcioHit );

    radiusVector.rotateZ( -stepOnTrack*curvedTrack->getOmega() ) ;
  }

  // test the residuals
  TrackFitterBase *trackFitter = TrackFitterFactory::getTrackFitter( TrackFitterBase::FitterTypes::SIMPLEMINIMIZER, 0 );

  std::cout << "#Straight line"<< std::endl;
   for (TrackerHitVec::const_iterator hitIter = straightTrack->getTrackerHits().begin();
		 hitIter <  straightTrack->getTrackerHits().end(); hitIter++)
   {
     EVENT::DoubleVec residuals = trackFitter->calculateResiduals(*hitIter, straightTrack);
     
     double const * pos = (*hitIter)->getPosition();
     std::cout << "Hit : "<< pos[0] << " ; "
	       << pos[1] << " ; "
	       << pos[2] << " ; Residuals: "
	       << residuals[0] << " ; " << residuals[1]  << std::endl;
   }

   std::cout << "#Curved track"<< std::endl;
   for (TrackerHitVec::const_iterator hitIter = curvedTrack->getTrackerHits().begin();
		 hitIter <  curvedTrack->getTrackerHits().end(); hitIter++)
   {
     EVENT::DoubleVec residuals = trackFitter->calculateResiduals(*hitIter, curvedTrack);
     
     double const * pos = (*hitIter)->getPosition();
     std::cout << "Hit : "<< pos[0] << " ; "
	       << pos[1] << " ; "
	       << pos[2] << " ; Residuals: "
	       << residuals[0] << " ; " << residuals[1]  << std::endl;
   }

  // memory leak: clean up the lcio stuff. Currently I don't care...

  return 0;
}
